class dbpedia  {

    /**
     * 
     * @param {*} $CIRCUIT_MOD 
     * @param {tracker_map} $trackerMap 
     */
    constructor($CIRCUIT_MOD=null, $trackerMap=null) {
        this.aec = new anne_external_connector("https://dbpedia.org/data/", "dbpedia", $CIRCUIT_MOD);
        this.aec.setTracker($trackerMap);

        this.person=0;
        this.company=0;
        this.location=0;
        this.date=0;
        this.other = 0;

        this._props = [];
        
        //will be populated by the specific search.
        this.resolvedNOBJ = null;
        this.page_name = null;              //needs to be set the last slash uri of this page.
        this.wiki_type_nObj = null;         //needs to be set to the neuron we will make in ANNE for this new search result thing  
        this.original_search_text = null;
        this.url = "";
        this.new_relons = [];

        this.wiki_type_nObj = null;
        
        this.buildProps();

        this._ignoredValues=[];//["deep state operative", "deep state actor", "deep state functionary", "deep state politician"]
    
        this.json = null;

        this._processing_complete_callback = null;
        this.ambiguous_resolutions = [];

        this._busted = [];
        this.location_support = [];
        this.primary_inst_support = [];

        //do we want to chase group and group event for dbpedia?  Not sure what those look like yet..
        this.known_classes_to_chase = [MEM.person.temp_nid, MEM.company.temp_nid,
                                       MEM.location.temp_nid];

        this.classForcer = null;
        this.corp_type = null;


        this._rez_part = MEM.dbpedia_respart;
        this.savedParams=null;

    }


    addAmbiguous=function($jsonobj, $matches) {
        var ambig = {jsonobj:$jsonobj, matches:$matches};
        this.ambiguous_resolutions.push(ambig);
        console.log("%c AMBIGUIOUS RESULTS ", TALK.bgred);
        console.log(ambig);
    }


    /**
     * incriment our likely neuron parent classifiction.
     * We generally have a flat heirerarchies with instantiation classes...
     * and we generate dynamic heirerarchies with lateral relons.
     * @param {*} $prop 
     */
    incrLikely=function($prop) {
        if (this.wiki_type_nObj!=null) return;
        if ($prop.incr==undefined) return;
        
        var tags = $prop.incr.split("/");
        if ($prop.classForcer!=undefined) {
            this.classForcer = $prop.classForcer;
        }
        for (var t=0; t < tags.length; t++) {
            if (tags[t]=="per") {
                this.person+=1;
            }
            else if (tags[t]=="cor") {
                this.company+=1;
            }
            else if (tags[t]=="date") {
                this.date+=1;
            }
            else if (tags[t]=="loc") {
                this.location+=1;
            }
            else this.other+=1;
        }
    }

    //sometimes we don't have the data spelled out for us on this..
    setLikelyANNEClass=function() {
        var last_high = 0;
        if (this.wiki_type_nObj!=null) return;

        //we hit a classForcer and that is the end of it.
        if (this.classForcer!=null) {
            this.wiki_type_nObj = this.classForcer;
            return;
        }

        if (this.person>0) {
            this.wiki_type_nObj = MEM.person;
            last_high = this.person;
        }
        if (this.company && this.company > last_high) {
            last_high = this.company;
            this.wiki_type_nObj = MEM.company;
        }
        if (this.location && this.location > last_high) {
            last_high = this.location;
            this.wiki_type_nObj = MEM.location;
        }
        if (this.date && this.date > last_high) {
            last_high = this.date;
            this.wiki_type_nObj = MEM.gregorian_date;
        }

        if (this.wiki_type_nObj==null) return;
        
        if (this.wiki_type_nObj!=null && this.wiki_type_nObj.nid==MEM.location.nid) {
            if (this.location_support.length) {
                console.log(this.location_support);
                //we have critical relons to figure out here!!
                //ie, "queens" is a bourough and a district in New York City.
                //New York City is a city in New York.
                //New York is in the USA.
                //we must get these right in order to be able to properly make a new location!
            }
        }
    }

    formatSearchUri=function($searchObj) {
        return this.aec.baseUri + $searchObj.search_ready+".json";
    }

    /**
     * props we want to map
     */
    buildProps=function(){   
        //person specific.
        this.makeFM(null, "id", "dbpedia id", TYPE_BE, MEM.dbpedia_id, TO_TYPE_NUMBER); 
        var ppbb = this.makeFM(this.processBirthname, "birthName", "birth name", TYPE_BE, MEM.birthname.nid, TO_TYPE_STRING, "per"); 
        ppbb.classForcer = MEM.person;
        
        var pp0 = this.makeFM(this.processSimpleText, "imdbId", "imdb id", TYPE_HAS, MEM.IMDB_id.nid, TO_TYPE_STRING, "per"); 
        pp0.prop_override  = "http://dbpedia.org/ontology/imdbId";
        pp0.classForcer = MEM.person;

        var pp88 = this.makeFM(this.processDate, "birthDate", "birth date", TYPE_BE, MEM.birthday.nid, TO_TYPE_ANNE, "per"); 
        pp88.bINSTSUPPORT=true;
        pp88.classForcer = MEM.person;

        var pp90 = this.makeFM(this.processDate, "deathDate", "death date", TYPE_BE, MEM.deathday.nid, TO_TYPE_ANNE, "per"); 
        pp90.prop_override = "http://dbpedia.org/ontology/deathDate";//we want the uri for their page name.
        pp90.classForcer = MEM.person;

        var pp2 = this.makeFM(this.processDate, "apprehended", "apprehended", TYPE_BE, MEM.apprehended.nid, TO_TYPE_ANNE, "per"); 
        pp2.prop_override = "http://dbpedia.org/ontology/apprehended";

        var pp1 = this.makeFM(this.processLocation, "birthPlace", "birth place", TYPE_BE, MEM.birthplace.nid, TO_TYPE_ANNE, "per"); 
        pp1.prop_override = "http://dbpedia.org/ontology/birthPlace";//we want the uri for their page name.
        pp1.process_array_as_one=true;
        pp1.classForcer = MEM.person;
        pp1.onlyOne=true;

        var pp2 = this.makeFM(this.processLocation, "deathPlace", "death place", TYPE_BE, MEM.deathplace.nid, TO_TYPE_ANNE, "per"); 
        pp2.prop_override = "http://dbpedia.org/ontology/deathPlace";//we want the uri for their page name.
        pp2.bONLYONE=true;
        pp2.process_array_as_one=true;
        pp2.classForcer = MEM.person;
        pp2.onlyOne = true;

        //this.makeFM(this.processLivingThing, "occupation", "occupation", TYPE_HAS, MEM.occupation.nid, TO_TYPE_ANNE, "per"); 
        
        this.makeFM(this.processReligion, "religion", "religion", TYPE_HAS, MEM.religion.nid, TO_TYPE_ANNE, "per"); 

        //people be alumnis of.
        this.makeFM(this.processEducation, "education", "education", TYPE_BE, MEM.alumni_of.nid, TO_TYPE_ANNE, "per"); 
        
        this.makeFM(this.processParty, "party", "party", TYPE_HAS, MEM.political_party.nid, TO_TYPE_ANNE, "per"); 

        var prop1 = this.makeFM(this.processPerson, "parent", "parent", TYPE_HAS, MEM.kinfolk_parent.nid, TO_TYPE_ANNE, "per"); 
        prop1.make_be_dual = MEM.kinfolk_child;
        prop1.dual_type = TYPE_HAS;
        prop1.must_be_person = true;
        
        var prop2 = this.makeFM(this.processPerson, "child", "child", TYPE_HAS, MEM.kinfolk_child.nid, TO_TYPE_ANNE, "per"); 
        prop2.make_be_dual = MEM.kinfolk_parent;
        prop2.dual_type = TYPE_HAS;
        prop2.prop_override = "http://dbpedia.org/ontology/child";
        prop1.must_be_person = true;
        
        
        var propd = this.makeFM(this.processCompany, "divisions", "divisions", TYPE_HAS, MEM.division.nid, TO_TYPE_ANNE, "cor"); 
        propd.make_be_dual = MEM.division_of;
        propd.dual_type = TYPE_BE;
        propd.must_be_corp = true;

        var prop3 = this.makeFM(this.processPerson, "spouse", "spouse", TYPE_HAS, MEM.spouse.nid, TO_TYPE_ANNE, "per"); 
        prop3.make_be_dual = MEM.spouse;
        prop3.dual_type = TYPE_HAS;
        prop3.prop_override = "http://dbpedia.org/ontology/spouse";

        var prop3 = this.makeFM(this.processPerson, "namedAfter", "named after", TYPE_HAS, MEM.named_after.nid, TO_TYPE_ANNE, "per/cor/loc"); 
        
        //we're not actually going to use this.  Just helps identify location
        var id1 = this.makeFM(null, "populationTotal", "population total", TYPE_HAS, "", TO_TYPE_ANNE, "loc"); 
        id1.calcOnly=true;
        id1.classForcer = MEM.location;
        
        var id2 = this.makeFM(null, "areaLandSqMi", "area land sq miles", TYPE_HAS, "", TO_TYPE_ANNE, "loc"); 
        id2.calcOnly = true;
        id2.classForcer = MEM.location;
        
        var pdmn = this.makeFM(this.processNonLivingThing, "populationDemonym", "population demonym", TYPE_HAS, MEM.pop_demonym.nid, TO_TYPE_STRING, "loc"); 
        pdmn.classForcer = MEM.location;
        
        var propHonors = this.makeFM(this.processNonLivingThing, "honours", "honours", TYPE_HAS, MEM.military_honours.nid, TO_TYPE_ANNE, "per"); 
        propHonors.prop_override = "http://dbpedia.org/ontology/honours";


        //this really should be in a TERM neuron.
        //documenting the term and office and rank.
        //hmmm.
        
        //NOT WORTH THE EFFORT RIGHT NOW.
        //this.makeFM(this.processNonLivingThing, "rank", "military rank", TYPE_HAS, MEM.military_rank.nid, TO_TYPE_ANNE, "per"); 
        
        //this is going to have to be compatible with our existing service term..
        //how we build them for wikipedia.
        //how we need to build them for wikispooks..
        //and this.  THEY ALL MUST CREATE THE SAME SERVICE TERM.
        this.makeFM(this.processOffice, "office", "service term", TYPE_HAS, MEM.service_term.nid, TO_TYPE_ANNE, "per"); 
        
        //no fields other than education for potential companies??????
        //anywhere??? in the data set????

        //corp specific
        var prop5 = this.makeFM(this.processPersonOrCorp, "foundedBy", "founded by", TYPE_HAS, MEM.founder.nid, TO_TYPE_ANNE, MEM.person, "cor/loc"); 
        prop5.make_be_dual = MEM.founder_of;
        prop5.dual_type = TYPE_BE;

        var prop55 = this.makeFM(this.processPersonOrCorp, "founders", "founded by", TYPE_HAS, MEM.founder.nid, TO_TYPE_ANNE, MEM.person, "cor/loc"); 
        prop55.make_be_dual = MEM.founder_of;
        prop55.dual_type = TYPE_BE;

        this.makeFM(this.processInteger, "foundedYear", "established", TYPE_HAS, MEM.established.nid, TO_TYPE_ANNE, "cor"); 
        this.makeFM(this.processNonLivingThing, "industry", "industry", TYPE_HAS, MEM.industry.nid, TO_TYPE_ANNE, "cor"); 
        this.makeFM(this.processPerson, "keyPeople", "key people", TYPE_HAS, MEM.key_person.nid, TO_TYPE_ANNE, "cor"); 
        
        //we should make a be dual with this relon..  (sub/sub of)
        var prop6 = this.makeFM(this.processCompany, "subsidiary", "subsidary", TYPE_HAS, MEM.subsidary.nid, TO_TYPE_ANNE, "cor"); 
        prop6.make_be_dual = MEM.subsidary_of;
        prop6.dual_type = TYPE_BE;

        //TO DO.
        //Fate ?  (should have a fate for corps that were absorbed, closed, defunct etc)

        var prop7 = this.makeFM(this.processEducation, "almaMater", "alma mater", TYPE_BE, MEM.alumni_of.nid, TO_TYPE_ANNE, "per"); 
        prop7.prop_override = "http://dbpedia.org/ontology/almaMater";
        prop7.make_be_dual = MEM.alumni;
        prop7.dual_type = TYPE_HAS;

        //country

        //these are used to help us get the proper data needed for location parent.
        var mainPropA = this.makeFM(this.supportingSelfLocationRelons, "subdivision", "subdivision", TYPE_HAS, null, TO_TYPE_ANNE, "loc");
        mainPropA.onlyTags = "loc";
        mainPropA.prop_override = "http://dbpedia.org/ontology/subdivision";
        var mainPropB = this.makeFM(this.supportingSelfLocationRelons, "synonym", "synonym", TYPE_HAS, null, TO_TYPE_ANNE, "loc");
        mainPropB.onlyTags = "loc";
        mainPropB.prop_override = "http://dbpedia.org/ontology/synonym";
        var mainPropC = this.makeFM(this.supportingSelfLocationRelons, "type", "type", TYPE_HAS, null, TO_TYPE_ANNE, "loc/per/cor/date");
        mainPropC.onlyTags = "loc";
        mainPropC.prop_override = "http://dbpedia.org/ontology/type";

        this.makeFM(null, "homepage", "website", TYPE_HAS, MEM.website.nid, TO_TYPE_ANNE, "cor"); 

        this.makeFM(this.processLocation, "location", "location", TYPE_HAS, MEM.location.nid, TO_TYPE_ANNE, "cor"); 
        this.makeFM(this.processLocation, "locationCity", "location", TYPE_HAS, MEM.location.nid, TO_TYPE_ANNE, "cor"); 

        this.makeFM(this.processImage, "logo", "logo", TYPE_HAS, MEM.logo.nid, TO_TYPE_ANNE, "cor"); 
        this.makeFM(null, "setCik", "setCik", TYPE_HAS, MEM.setCik, TO_TYPE_NUMBER, "cor"); 
        this.makeFM(this.processNonLivingThing, "products", "product", TYPE_HAS, MEM.product.nid, TO_TYPE_ANNE, "cor");

        //wikicomons uri parts.
        //var uri1 = this.makeFM(this.processUriPart, "imageFlag", "imageFlag", TYPE_HAS, MEM.wcImgUriPart_Flag.nid, TO_TYPE_URL, "loc");
        //uri1.bDONTCLEAN=true;
        //var uri2 = this.makeFM(this.processUriPart, "imageMap", "imageMap", TYPE_HAS, MEM.wcImgUriPart_Map.nid, TO_TYPE_URL, "loc");
        //uri2.bDONTCLEAN=true;
        //var uri3 = this.makeFM(this.processUriPart, "imageSeal", "imageSeal", TYPE_HAS, MEM.wcImgUriPart_Seal.nid, TO_TYPE_URL, "loc");
        //uri3.bDONTCLEAN = true;
        //var uri4 = this.makeFM(this.processUriPart, "imageSkyline", "imageSkyline", TYPE_HAS, MEM.wcImgUriPart_Skyline.nid, TO_TYPE_URL, "loc");        
        //uri4.bDONTCLEAN = true;

        //var uri6 = this.makeFM(this.processUriPart, "thumbnail", "thumbnail", TYPE_HAS, MEM.wcImgUriPart_thumbnail.nid, TO_TYPE_URL, "loc");        
        //uri6.bDONTCLEAN = true;
        //uri6.prop_override = "http://dbpedia.org/ontology/thumbnail";
        //http://commons.wikimedia.org/wiki/Special:FilePath/Donald_Trump_official_portrait.jpg?width=300

    }


    makeFM=function($cb_resolve, $src_prop_name, $anne_prop_name, $type, $reln, $to_type, $incr="per"){
        var prop = this.aec.makeFM($cb_resolve, $src_prop_name, $anne_prop_name, $type, $reln, $to_type, $incr);
        this._props.push(prop);
        prop.onlyTags = [];
        return prop;
    }

    

    /**
     * Don't chase classes we don't want.
     * @returns 
     */
    chaseableClass = function() {
        if (this.resolvedNOBJ==null) return true;
        return this.aec.CIRCUIT_MOD.ACE.isDbpediaChaseable(this.resolvedNOBJ);
    }
    

    /**
     * @param {*} $resolvedNOBJ 
     * @param {*} $txt 
     * @param {*} $original_search_val 
     * @returns 
     */
    getData=async($resolvedNOBJ=null, $txt = "Barack Obama", $external_search_val="", $completeCallback=null)=> {
        console.log("%cgetDATA DBPEDIA %c"+$txt, TALK.bgredbig, TALK.bggreen);
        this._processing_complete_callback = $completeCallback;
        this.resolvedNOBJ = $resolvedNOBJ;
        
        //we don't want to get this data from dbpedia afterall
        //(if we have a legit uri from dbpedia bi-pass this)
        if (this.resolvedNOBJ!=null && $external_search_val=="") {
            if (this.chaseableClass()==false) {
                this.aec.addToConsole("<b>"+this.resolvedNOBJ.text+"</b> text "+$resolvedNOBJ.text+" is not a dbpedia chaseable class");
                return this._processing_complete_callback(this);
            }
        }
        
        var txt1= $txt;
        //we don't want existing ANNE names that are not the proper external data file name to fail.
        //we populate external search val if we have it from our resolvedNOBJ
        if (txt1=="") {
            if ($resolvedNOBJ.text!="") txt1=$resolvedNOBJ.text;
        }
        var searchObj = this.prepSearch(txt1, $external_search_val);

        this.name =  searchObj.search;
  
        var url = this.formatSearchUri(searchObj, $external_search_val);
        console.log(url);
        this.url = url;

        this.name =  searchObj.search;
        this.external_ready_search = searchObj.search_ready;
        this.url = url;

        if (this.url.indexOf(" ")!=-1) {
            console.log("THIS IS RETARDED");  
            this.url = this.url.replaceAll(" ", "_");
        }
        
        var responseObj = await ANNE.API.avoidCorsDBPEDIA(searchObj.search_ready);
        console.log(responseObj);

        this.status = responseObj.status;
        if (responseObj.status=="bad") {
            this.aec.addToConsole("dbpedia query problem for "+this.url);
            return this._processing_complete_callback(this);
        }

        var json2 = responseObj.json2;
        console.log(json2);
        
        if (json2==undefined){
            this.aec.addToConsole("dbpedia query problem for "+this.url);
            return this._processing_complete_callback(this);
        }

        if (json2["http://dbpedia.org/ontology/wikiPageRedirects"]!=undefined) {
            var redir = json2["http://dbpedia.org/ontology/wikiPageRedirects"];
            
            var slash = redir[0].value.split("/");
            console.log("need to redirect to "+slash);
            
            //fix the url with the new value
            var newSearch =  slash[slash.length - 1];

            //replace the old string with the new string.
            this.url = this.url.replace(searchObj.search_ready, newSearch);
            
            searchObj.search_ready = newSearch;
            this.external_ready_search = newSearch;

            //fix our name for this thing.
            this.name = this.external_ready_search.replaceAll("_", " ");

            console.log(this.url);
            
            responseObj = await ANNE.API.avoidCorsDBPEDIA(searchObj.search_ready);
            json2 = responseObj.json2;

            console.log(json2);
        }

        if (json2==undefined) {
            this.aec.addToConsole("dbpedia query problem for "+this.url);
            return this._processing_complete_callback(this);
        }

        
        if (json2["http://dbpedia.org/property/type"]!=undefined) {
            var typeHelper = json2["http://dbpedia.org/property/type"];
            for (var t=0; t <typeHelper.length; t++) {
                var jobj = typeHelper[t];

                //these can both be true.. (worry about that later)
                if (jobj.value.indexOf("Conglomerate_(company)")!=-1) {
                    this.wiki_type_nObj = MEM.company;
                    this.corp_type = MEM.conglomerate; //conglomerate
                    break;
                }
                else if(jobj.value.indexOf("Division_(business)")!=-1) {
                    this.wiki_type_nObj = MEM.company;
                    this.corp_type = MEM.division; //conglomerate
                }
            }

        }
        
        var pre = "http://dbpedia.org/property/";
        for (var p = 0; p < this._props.length; p++) {
            var prop = this._props[p];
            if (prop=="founders" || prop=="subsid") {
                console.log("test here");
            }
            if (this.wiki_type_nObj!=null) {
                if (this.wiki_type_nObj.nid==ANNE.MEM.company.nid && prop.must_be_person!=undefined && prop.must_be_person==true) {
                    continue;
                }
                else if (this.wiki_type_nObj.nid==ANNE.MEM.person.nid && prop.must_be_corp!=undefined 
                    && prop.must_be_corp==true) {
                    continue;
                }
            }

            prop.json_objs = [];

            //we don't know what the class
            if (prop.onlyTags.length>0) {
                var incrs = prop.incr.split("/");
                for (var t=0; t < incrs.length; t++) {
                    if (prop.onlyTags.indexOf(incrs[t])==-1) continue;
                }
            }
            
            var propf = pre+prop.src_prop_name;
            
            if (prop.prop_override!=undefined) propf=prop.prop_override;

            //just a helper to identify ANNE class.
            if (prop.calcOnly!=undefined && json2[propf] != undefined) {
                this.incrLikely(prop);
                continue;
            }

            if (json2[propf] != undefined) {
                //for any given prop in our props we have an array of results (if any at all)
                var arr = json2[propf];

                //array of json results for THIS prop.
                for (var a=0; a < arr.length; a++) {
                    var jsonobj = arr[a];
                    
                    console.log(jsonobj);
                    
                    //take only one even if more are here.
                    if (a>0 && prop.bONLYONE!=undefined && prop.bONLYONE==true) continue;
                    
                    //untested for dbpedia
                    if (prop.process_array_as_one!=undefined && prop.process_array_as_one==true) {
                        if (prop.resolve_obj != undefined && prop.resolve_obj!=null && typeof prop.resolve_obj === "function") {
                            prop.json_objs.push(arr[0]); //push only the first one.
                            await prop.resolve_obj(arr, prop);
                            break;
                        }
                        break;
                    }

                    //if a literal, load the to_s already.
                    //we are ready to go.
                    if (jsonobj.type=="literal") {
                        if (jsonobj.value=="") continue;
                        
                        if (typeof jsonobj.value=="String") {
                            if (prop.bDONTCLEAN!=undefined) jsonobj.to_s = jsonobj.value;
                            else jsonobj.to_s = jsonobj.value.replaceAll("_", " ");
                        
                            jsonobj.to_s = jsonobj.to_s.replace("]]");

                            if (prop.bDONTCLEAN==undefined) {
                                var fixedObj = ANNE.UTILS.stripParens(jsonobj.value);
                                jsonobj.to_s = fixedObj.str;
                                jsonobj.context = fixedObj.context;
                            }
                        }

                       
                        prop.json_objs.push(jsonobj);
                    }

                    else if (jsonobj.type=="uri") {
                        var slash = jsonobj.value.split("/");
                        jsonobj.last_slash_uri = slash[slash.length - 1].replaceAll("-2D","-");

                        if (prop.bDONTCLEAN!=undefined) jsonobj.text_to_map = jsonobj.last_slash_uri;
                        else jsonobj.text_to_map = jsonobj.last_slash_uri.replaceAll("_", " ");

                        //will strip out "(xxx)" but put the xxx into .context.
                        if (prop.bDONTCLEAN==undefined) {
                            var fixedObj = ANNE.UTILS.stripParens(jsonobj.text_to_map);
                            jsonobj.text_to_map = fixedObj.str;
                            jsonobj.context = fixedObj.context;
                        }

                        //just hacky willy wonka stuff here, that's all.
                        jsonobj.text_to_map = jsonobj.text_to_map.replace("]]");

                        //we are not creating any new neurons to ourselves here.  This gets willywonky very
                        //quickly.
                        if (jsonobj.text_to_map==this.name) {
                            continue;
                        }

                        prop.json_objs.push(jsonobj);  //keep the uri
                    }

                    //RESOLVE CALLED FOR EACH PROP.
                    //resolve to a known neuron if we have resolve set.
                    //we have prepped the data in a string or an array of strings.
                    //we will map ALL possible wikispook values to neurons here.
                    if (prop.resolve_obj != undefined && prop.resolve_obj!=null && typeof prop.resolve_obj === "function") {
                        console.log("handling prop ",prop);
                        if (jsonobj!=null && jsonobj.text_top_map!=undefined 
                            && jsonobj.text_to_map.indexOf("Family of")==0) {
                            console.log("we need a resolveGroup here, not person");
                            continue;
                        }
                        await prop.resolve_obj(jsonobj, prop);
                    }
                    this.incrLikely(prop);
                }
            }
        }

        //will set wiki_type_nObj
        this.setLikelyANNEClass();
        
        //resolve ambigs.
        //iterate through the ambigs..
        if (this.ambiguous_resolutions.length) {
            this.aec.CIRCUIT_MOD.ACE.resolveAmbigs( this.ambiguous_resolutions, this._processing_complete_callback, this );
            return;
        }

        else this._processing_complete_callback(this);
    }
    



    /**
     * we need all first letters capitalized.
     * we need underscores for spaces.             
     * we need to put dots after suffix and middle initials!
     * ie, John_D._Rockefeller_III
     * John_D._Rockefeller_Jr.
     * @param {*} $search 
     * @returns 
     */
    prepSearch = function($search, $external_override="") {
        var searchObj = {search_ready:"", search:""};
        if ($external_override.length) {
            searchObj.search_ready = $external_override.trim();
            searchObj.search = $external_override.trim();
            this.external_ready_search = searchObj.search_ready;
            return searchObj;
        }
        if (typeof $search != "string") {
            console.log("what in sam hill's name....");
        }

        else if (UTILS.isMixedCase($search)==true) {
            console.log("leave as camel case");
        }
        else $search = $search.trim().toLowerCase();

        var wds = $search.split(" ");
        var nomiddleinitial =  "";
        for (var p=0; p < wds.length; p++) {
            if (wds[p]=="") continue;
            wds[p] = wds[p][0].toUpperCase() + wds[p].substr(1);
            if (wds[p].length==1) wds[p]=wds[p]+".";
            else {
                var wd = wds[p];
                if(wd=="Sr" || wd=="Jr") wd = wd+".";
            }
            if (p>0 && wds[p-1]=="'") wds[p] = wds[p].toUpperCase();    //D'Aniello
        }
        return {search_ready:wds.join("_"), search:wds.join(" ")};
    }

    supportingSelfLocationRelons = async($jsonobj, $prop)=> {
        this.location_support.push({jsonobj:$jsonobj, $anne_prop_name:$prop.$anne_prop_name});
    }


    processSimpleText = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processSimpleText";
        
        console.log($jsonobj);
        $jsonobj.to_s=$jsonobj.value;
    }
    processBirthname = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processBirthname";
        
        console.log($jsonobj);
        //we aleady have this as the name.
        if (this.resolvedNOBJ!=null) {
            var lower = $jsonobj.value.toLowerCase();
            if (lower==this.resolvedNOBJ.text.toLowerCase() 
                || (this.resolvedNOBJ.aka_lower!=undefined && this.resolvedNOBJ.aka_lower.indexOf(lower)!=-1)) {
                return;
            }
        }
        $jsonobj.to_s=$jsonobj.value;
    }
    processDate = async($jsonobj, $prop)=> {
        $jsonobj.processedBy="dbpedia.processDate";
        
        console.log($jsonobj);
        var ad = new anne_date($jsonobj.value);
        if (ad==null) return;
        var bMAKEGHOSTIFNOTFOUND=true;
        await ad.testDate(bMAKEGHOSTIFNOTFOUND);
        if (ad.nObj!=null) {
            $jsonobj.to_s = ad.nObj.nid;

            //this is be-defining and could help us find a non-found neuron we couldn't
            //otherwise yet get.  (ie,  D TRUMP born June 14, 1946)
            this.primary_inst_support.push({reln:$prop.reln, $to:$jsonobj.to_s});
        }
        else {
            console.log("something wrong with this birth date "+$jsonobj);
            this._busted.push($jsonobj);
        }
    }

    processLogo = async($jsonobj)=>{
        $jsonobj.processedBy="dbpedia.processLogo";
        
        console.log($jsonobj);
        return;
    }
    
    /**
     * Just keep the reference as it.  This is for wikicommons links.
     * @param {*} $jsonobj 
     * @returns 
     */
    processUriPart = async($jsonobj)=>{
        $jsonobj.processedBy="dbpedia.processUriPart";
        
        console.log($jsonobj);
        var uripart ="";
        if ($jsonobj.text_to_map!=undefined) uripart = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) uripart = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) uripart = $jsonobj.value;
        else return;

        var to_s = uripart.replaceAll(" ", "_");
    
        if (to_s.indexOf("?")!=-1) {
            //"Donald_Trump_official_portrait.jpg?width=300"
            var qi = to_s.indexOf("?");
           to_s =to_s.substr(0, qi);
        }

        $jsonobj.to_s = to_s;
    }


    processLocation = async($arr, $prop)=> {
        //$jsonobj.processedBy="dbpedia.processUriPart";
        
        var lookupval ="";
        console.log($arr);

        if (Array.isArray($arr)==true) {
            var parts = [];

            //if dbpedia returns more than one value and that value has a comma in it, 
            ///only use slot 0.
            if ($arr.length>1 && $arr[0].value.indexOf(",")!=-1) {
                $arr = $arr.splice(0,1)
            }

            for(var a=0; a < $arr.length; a++) {
                var jsonobj = $arr[a];
                console.log(jsonobj);
                var slash = jsonobj.value.split("/");
                var fix = slash[slash.length - 1].replaceAll("-2D","-");
                fix = fix.replaceAll("_", " ");
                fix = fix.trim();

                var fixedObj = ANNE.UTILS.stripParens(fix);
                fix = fixedObj.str;
                if (fixedObj.context) jsonobj.context = fixedObj.context;//(magazine) ??? lol (eppy death)
                

                //if this entry has a comma in it, split those out as their own spots in the parts array.
                if (fix.indexOf(",")!=-1) {
                    var fparts = fix.split(",");
                    for (var f=0; f<fparts.length; f++) {
                        parts.push( fparts[f].trim());
                    }
                }
                else parts.push(fix);
            }
            
            parts = parts.reverse();//why?
            
            lookupval = parts.join(",");
            var jsonobj = $arr[0];
            jsonobj.text_to_map = lookupval;
        }
        else {
            console.log("this function expects an array of location values to build a location.");
            return;
        }
        
        var anne_loc = new anne_location(lookupval);
        var bMAKEGHOSTIFNOTFOUND = true;
        await anne_loc.processFull([], false, bMAKEGHOSTIFNOTFOUND);
        anne_loc.print();
        anne_loc.printFullDebugString();
        
        if (anne_loc.nObj!=null) {
            console.log(anne_loc.nObj);
            
            if (this.resolvedNOBJ!=null && $propObj.onlyOne!=undefined && $propObj.onlyOne==true) {
                var bISBEST = MEMZ.isBestLocation(this.resolvedNOBJ, $propObj.reln, $obj.to_s, anne_loc);
                if (bISBEST==false) {
                    var jsonobj = $arr[0];
                    jsonobj.to_s  = "";
                    return;
                }
            }
            
            jsonobj.to_s = anne_loc.nObj.nid;
            if (anne_loc.nObj.bNEW==undefined) {
                if (anne_loc.nObj.bNEW==false) {
                    this.aec.mappedToExisting(lookupval, anne_loc.nObj, ANNE.MEM.location);
                }
                else if (anne_loc.nObj.bNEW==true) {
                    this.aec.trackerMap.tracker.locations[anne_loc.nObj.text] = anne_loc.nObj;
                    anne_loc.nObj.createdBy="DBPEDIA";
                }
            }
            return jsonobj;
        }
        else if (anne_loc.bAMBIGUOUS==true) {
            this.addAmbiguous(jsonobj, anne_loc.matches);
            return jsonobj;
        }
        else {
            console.log("location????? %c"+jsonobj, TALK.bgred);
            return $arr;
        }
    }


    processLivingThing = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processLivingThing";
        console.log($jsonobj);
        return await this.findOrAddNew($jsonobj, MEM.ec_living);
    }
    processNonLivingThing = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processNonLivingThing";
        console.log($jsonobj);
        return await this.findOrAddNew($jsonobj, MEM.ec_nonliving);
    }

    //we need an office -- as a new nonLivingThing.
    //we also, however, want a service term for this particular thing.
    processOffice = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processOffice";
        console.log($jsonobj);
        var matches = await this.findOrAddNew($jsonobj, MEM.ec_nonliving);
        if (matches.length==0) return matches;
        
        //the office text we resolve is actually the officeTheme.
        //like senator, govenor, president, or whatever title the thing is.
        
        //see: "Benjamin Disraeli"
        //offices held:
        //Chancellor of the Exchequer
        //Leader of the Opposition

        var officeTheme = matches[0];

        //this will be the term neuron (we make or resolve)
        $jsonobj.to_s = "";

        //SERVICE TERM
        //ROLE_START
        //ROLE_END
        //ROLE_ACTOR
        //ROLE_THEME

        //country
        //political party
        //state

        var roles =[{role:MEM.role_theme_thing, nObj:officeTheme}];
        //process roles..

        //findClassifiedClump(MEM.service_term, roles);
        return;
    }


    /**
     * Need to make sure we don't conflict with wikispooks entries here and create new religions
     * ....
     * @param {*} $jsonobj 
     * @returns 
     */
    processReligion = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processReligion";
        return;
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;
    
        var GMOBJ = await MEMG.getOrMakeIfNewReligion(this.aec, lookupval);
        var nObj = null;
        if (GMOBJ.status=="good") nObj = GMOBJ.match;
        else if (GMOBJ.status=="ambiguous") {
            this.addAmbiguous($jsonobj, GMOBJ.matches, lookupval);
            return;
        }
        $jsonobj.to_s = nObj.nid;
        return;
    }
    

    processParty = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processParty";
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;

        console.log("party:"+lookupval);
        if (lookupval=="Republican Party") {
            $jsonobj.to_s = MEM.us_republican.nid;
            return;
        }
        else if (lookupval=="Democratic Party") {
            $jsonobj.to_s = MEM.us_democrat.nid;
            return;
        }
        return await this.findOrAddNew($jsonobj, MEM.political_party, NTYPE_INST);
    }

    processEducation = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processEducation";
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;
        if (lookupval==undefined || lookupval==null || lookupval=="") return;
        var highest_nObj = await this.aec.estimateClass(lookupval, MEM.higher_education);
        var GMOBJ = await MEMG.getOrMakeIfNewCorp(this.aec, lookupval, highest_nObj);
        var nObj = null;
        if (GMOBJ.status=="good") nObj = GMOBJ.match;
        else if (GMOBJ.status=="ambiguous") {
            this.addAmbiguous($jsonobj, GMOBJ.matches);
            return;
        }
        $jsonobj.to_s = nObj.nid;
        return;
    }    



    //this will be a peson.
    //the challenge here is the dupes w/ same names.
    //with the uri we can differentiate but we'd have to then look that up too.
    processPerson = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processPerson";
        
        var lookupval ="";
        if ($jsonobj.value.indexOf("[[#")===0) {
            console.log("skipping this.")
            return;
        }
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;

        console.log("(procesPerson) %c"+lookupval, TALK.bgyellow);

        //if spouse, and gender=female,
        console.log($jsonobj);

        //we probably want a last name, first name search..
        //so we can find people even if they don't have a supplied alias ?
        var nObj = null;
        var GMOBJ = await MEMG.getOrMakeIfNewPerson(this.aec, lookupval);
        if (GMOBJ.status=="good") {
            nObj = GMOBJ.match;
            if (nObj.bNEW==true) {
                this.aec.addLocation(nObj);
            }
        }
        else if (GMOBJ.status=="ambiguous") {
            this.addAmbiguous($jsonobj, GMOBJ.matches);
            return;
        }
        $jsonobj.to_s = nObj.nid;
        return nObj;
    }


    processPersonOrCorp = async($jsonobj)=>{
        $jsonobj.processedBy="dbpedia.processPersonOrCorp";
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;
        if (lookupval==undefined || lookupval==null || lookupval=="") return;

        if (lookupval.indexOf("[[#")===0) {
            console.log("skipping this.")
            return;
        }

        //NEED TO DISAMBUATE HERE!!! TO DO:
        console.warn("DISMBIGUATE PERSON OR COMPANY WITH DBPEDIA HERE.");
        console.log($jsonobj);
    }


    /**
     * TEST THIS HIGHEST_SIM GUESS!!!!.
     * @param {*} $jsonobj 
     * @returns 
     */
    processCompany = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processCompany";
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;
        if (lookupval==undefined || lookupval==null || lookupval=="") return;

        if (lookupval.indexOf("[[#")===0) {
            console.log("skipping this.")
            return;
        }
        
        var highest_nObj = await this.aec.estimateClass(lookupval, MEM.corp_entity_thetype);
        var GMOBJ = await MEMG.getOrMakeIfNewCorp(this.aec,lookupval, highest_nObj);
        var nObj = null;
        if (GMOBJ.status=="good") nObj = GMOBJ.match;
        else if (GMOBJ.status=="ambiguous") {
            this.addAmbiguous($jsonobj, GMOBJ.matches);
            return;
        }
        $jsonobj.to_s = nObj.nid;
        return;
    }   

    /**
     * 
     * @param {*} $jsonobj 
     */
    processImage = async($jsonobj)=>{
        $jsonobj.processedBy="dbpedia.processImage";
        
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else if ($jsonobj.to_s!=undefined) lookupval = $jsonobj.to_s;
        else if ($jsonobj.value!=undefined) lookupval = $jsonobj.value;
        else return;

        //hmmm.
        if (lookupval==undefined || lookupval==null || lookupval=="") return;

        if (typeof lookupval!="string") {
            alert("what in sam hill.. fix this");
            return;
        }

        var base = "http://commons.wikimedia.org/wiki/Special:FilePath/";
        
        //not a valid pattern we handle for images w/ fetch.
        if (lookupval.indexOf("Image:")==0) return;
        
        var anneimg = await IMAGES.getImgMetaFromUrl(base+lookupval+"?width=300");
        if (anneimg==undefined) {
            console.log("bad url for us");
            return;
        }
        if (anneimg.finger==undefined || anneimg.finger==null) return;

        //check that this checks local space too (ghost imgs not on ANNE)
        var result = await CHASE.getGlobal(MEM.neuronImg, anneimg.finger);
        console.log(result);
        if (result.status=="good" && result.message=="available") {
            var imgNOBJ = await anneimg.makeGhost(null, this.nObj, TYPE_HAS, MEM.neuronImg.nid, this.original_search_text);
            $jsonobj.to_s = imgNOBJ.nid;
            this.aec.addToConsole("added new img <b>"+lookupval+"</b>");
        }
        //dupe img
        else {
            console.log("%c SEE WHERE THE NEURON IS AND SET THE TO_S");
            console.log(result);
            this.aec.addToConsole("found existing img of <b>"+lookupval+"</b>");
        }
    }

    /**
     * we want a real integer here.
     * @param {*} $jsonobj 
     */
    processInteger = async($jsonobj)=> {
        $jsonobj.processedBy="dbpedia.processInteger";
        
        return await this.findOrAddNew($jsonobj, MEM.integer, NTYPE_INST);
    }


    /**
     * Standard search and if not found, create new.
     * We will process all pvals in the json_objs array as relons according to the prop definition.
     * We will also create another relon for MEM.dbpedia_respart / last_slash_uri if this to_s was a uri value in dbpedia. 
     * @param {*} $prop 
     * @param {*} $classNOBJ 
     */
    findOrAddNew=async($jsonobj, $classNOBJ, $ntype=NTYPE_CLASS)=> {
        console.log("findOrAddNew %c" + $classNOBJ.text, TALK.bgyellow);
        var lookupval ="";
        if ($jsonobj.text_to_map!=undefined) lookupval = $jsonobj.text_to_map;
        else lookupval =$jsonobj.to_s;
        if (lookupval==undefined || lookupval==null || lookupval=="") return [];
        
        var matches = await ANNE.CIRCUIT.searchToResolve(lookupval, $classNOBJ);

        if (matches.length==1) {
            $jsonobj.to_s = matches[0].nid;
            this.aec.addToConsole("found existing "+$classNOBJ.text+" of <b>"+matches[0].text+"</b>");
        }
        else if (matches.length>1) {
            this.aec.addToConsole("ambiguous "+$classNOBJ.text+" for <b>"+lookupval+"</b>");
            $jsonobj.to_s = matches[0].nid;
            return matches;
        }
        else {
            var ghostNOBJ = await GHOSTDATA.makeGhostNeuron($classNOBJ, lookupval, $ntype);
            $jsonobj.to_s = ghostNOBJ.nid;
            this.aec.addToConsole("added new "+$classNOBJ.text+" <b>"+ghostNOBJ.text+"</b>");
        }
        return matches;
    }
}