const PATHTYPE_ANNE = 1;
const PATHTYPE_UDF = 2;
const RENDERTYPE_BASE = 1;
const RENDERTYPE_STANDARD = 2; 
const RENDERTYPE_FLOATER = 3;
const RENDERTYPE_TAB = 4;
const RENDERTYPE_MODAL = 5; 
const RENDERTYPE_INJECT = 6;

const TYPE_NAME = -1;

const TYPE_SYS = 0;             //Sytem info for neuron (sponsored_by, annother, owned_by).  Builds extra indexes.
const TYPE_BE = 1;              //semantic shothand for BE DIMENSION.  CLASS/INST RELNS create a DAG.
const TYPE_PARAM = 2;           //semantic shorthand for param (class attribute/property)
const TYPE_HAS = 3;             //semantic shorthand for param fulfilling has value (attribute/property)
const TYPE_AWARENESS = 4;       //for awareness space (emotions/feelz)

const TYPE_EXPERIENCE = 5;      //for experience (temporal) neuron space

//CLUMPS can function as a TYPE_BE, TYPE_HAS, OR TYPE_SYS.
//those Types are actually short-hand for clumps but they can get away with
//only using one TX to specify the triplet info.
//A full clump is for specifying more than the standard 3 pieces of semantic info
//in a single ANNE TX.
//** For ease of use we can put a TYPE_XXX in the clump
//to clue our interpreter to what dimension the clump lives in.
//The other CLUMPS of episodic, BRAIN and EXPERIENCE are already their own dimension
//so this is only relevant for semantic clumps that are handled in semantic memory space.
const TYPE_CLUMP = 6;

const TYPE_SYS_MSG = 7;        //only relevant to coms, not the neuron
const TYPE_BRAIN = 8;          //for neural-net space wirings a->wire->b
const TYPE_ROLE = 9;           //for specifiying  semantic roles inside of a clump

const TYPE_METAN = 10;         //for specifying meta information about roles used in a clump

const TYPE_EPISODIC = 11;      //for documenting a self-sensory/vitals data at run-time in a clump.
const TYPE_INTERNAL_PAY = 17;  //keyless neurons paying out.
const TYPE_BROADCAST_PAY = 18; //standard broadcast payment.

const TYPE_BRAIN0 = 20;        //20+ are for sequence in the brain dimension

//100+ are available for brain type/custom use.. assumed to be order if more than one in a neuron's outs.
const TYPE_BRAIN120 = 120;     //used in ASDF-> annode relons to store trigger pats. (can use how u like in other neurons)
const TYPE_BRAIN121 = 121;     //myelination measure (a->b->[to_type_number]#c)

const TO_TYPE_SEED = 0;        //relon to is an encrypted seed (all seed are encrypted)
const TO_TYPE_ANNE = 1;        //relon to is a .nid, the idenfier of another neuron.  This is a legit graph triplet.
const TO_TYPE_TXID = 2;        //relon to is a txid.
const TO_TYPE_URL = 3;         //relon to is a url.

//relon to is a number.  Note: our ANNODE stores numerics in a LONG field
//if a non-long/int is needed use the TO_TYPE_STRING

const TO_TYPE_NUMBER = 4;       
const TO_TYPE_DATE = 5;         //relon to is a date in the ANNE date format.    
const TO_TYPE_DATETIME = 6;     //relon to is a datetime in the ANNE datetime format.
const TO_TYPE_STRING = 7;       //relon to is a string

const TO_TYPE_TEXT = 7;         //alias in case u conflate

const TO_TYPE_BLOB = 8;         //relon to is a blob (bad donkey, not welcome in ANNE)
const TO_TYPE_TIME = 9;         //relon to is a time in the ANNE time format.
const TO_TYPE_BOOLEAN = 10;     //relon to is a boolean string which will be converted to true/false at runtime
const TO_TYPE_FILEHASH = 11;    //relon to is a filehash
const TO_TYPE_IPV4 = 12;    //relon to is a IPV4 address
const TO_TYPE_IPV6 = 13;    //relon to is a IPV6 address

//encrypted versions of the above + 30 in value from its respective type.
//expansion to 30 base types possible (don't foresee this as needed
//in a perfect-brain world, all are references to either neurons or other TXIDs

const TO_TYPE_ENCR_ANNE = 31;                
const TO_TYPE_ENCR_TXID = 32;
const TO_TYPE_ENCR_URL = 33;        
const TO_TYPE_ENCR_NUMBER = 34;
const TO_TYPE_ENCR_DATE = 35;
const TO_TYPE_ENCR_DATETIME = 36;
const TO_TYPE_ENCR_STRING = 37; 
const TO_TYPE_ENCR_BLOB = 38;
const TO_TYPE_ENCR_TIME = 39;
const TO_TYPE_ENCR_BOOLEAN = 40;
const TO_TYPE_ENCR_FILEHASH = 41;
const TO_TYPE_ENCR_IPV4 = 42;
const TO_TYPE_ENCR_IPV6 = 43;

const NTYPE_CLASS = 0;
const NTYPE_INST = 1;
const NTYPE_UNKNOWN = 2;
const NTYPE_CLUMP = 3;

const ORIGIN_CHAIN = 0;         //a real TX on-chain
const ORIGIN_NODE = 1;          //an emulated neuron created by the node
const ORIGIN_NODE_PRELOAD = 2;  //an emulated neuron preloaded by the node (with a real seed)
const ORIGIN_OWNERPROOF = 3;    
const ORIGIN_CONTRACTS = 4;     //emulated from contracts
const ORIGIN_KEYLESS = 5;
const ORIGIN_GLOBAL = 6;
const ORIGIN_CUSTODIAL  = 8;
const ORIGIN_EXTERNAL_ML = 12;  //if a ML is joined w/ the DB, that data is ML.

const RORIGIN_ANNODE = 0;         //origination made on the ANNODE in ANNE hypergraph only
const RORIGIN_BROADCAST = 1;    //origination broadcast as a TX
const RORIGIN_COLLAB_BROADCAST = 2;        //collab broadcast (waiting approval)
const RORIGIN_GHOST = 3;        //origination user local

const ANNE_ACCTID = "7777770163980908667";    
const ANNE_ADMIN_ACCTID = "7777771543345157118";
const ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";

const TARGET_NID = 1;
const TARGET_TXID = 2;

const TXMODE_GHOST = 0;             //store to local DB only
const TXMODE_ANNE = 1;              //instant send

const CTYPE_COLLAB_ONLY = 0;
const CTYPE_COLLAB_HANAI = 1;
const CTYPE_PENDING_ANJ = 2;
const CTYPE_HANAI_ONLY = 3;
const CTYPE_MERGE_REQUEST = 4;

//Keyless neurons allow for easy creation of child relons w/o the management of another key
//They can be collaborated w/ by request or easy delegation rights set to other approved authors.
//Note: ANNE globals (date, numbers, locations, keywords) are keyless neurons and are created on-demand.
const NEWNEURONMODE_KEYLESS = 0;  

//Key'd neurons provide real a real/key that is only known by the identity that created it.
//Key'd neurons can be collaborated w/ by request but they cannot be delegated since only the
//Key holder can broadcast transactions from the key'd neuron.
//Note: ANNE preloaded neurons are key'd as they need pre-known neuron IDs to reference them in the base-protocol.
const NEWNEURONMODE_KEYED = 1;        

const NEURON_NOT_FROZE = 0;
const NEURON_FROZE = 1;

const LINKS_NEURON_MODAL = 1;            //quick neuron summary page
const LINKS_NEURON_EMBED = 2;            //full neuron summary page
const LINKS_NEURON_DS = 3;               //data-source view of neuron's BE dimension
const LINKS_NEURON_TBL_VIEW = 4;         //table view of neuron's relons
const LINKS_NEURON_TREE = 5;             //tree view of a neuron's BE dimension
const LINKS_NEURON_ANNET = 6;            //graph explore of neurons
const LINKS_NEURON_ANDY = 7;             //ADMIN AREA
const LINKS_NEURON_ALEXANNDRIA = 8;      //library of your neurons  (can also find others)

const LINKS_NEURON_USER = 10;            //user profile page.
const LINKS_NEURON_INJECT_CB = 11;       //injection link w/ callback
const LINKS_NEURON_BLOCKCHAIN = 12;      //a link to the 3d viewer of blockchain data.
const LINKS_NEURON_DOSSIER = 14;         //detailed dossier view
const LINKS_NEURON_BRAIN = 16;           //decide if rememberz/forgetz is on/off
const LINKS_NEURON_ANNULET_OF_TRUST = 17;

const LINKS_RELON_ANNE = 20;             //an ANNE relon of somekind.
const LINKS_TX_BLOCKEXPLORER = 21;       //not an ANNE relon, a standard blockchain TX

//these mirror the value sin the node L1 code.
const BLOCK_HEADER_LENGTH = 232; //TODO: this may change if we add sm in the block...
const MAX_BALANCE_ANNE = 2200000000000; 
const ONE_ANNE = 100;
const MIN_ANNIES_SEND = 1000;                   //not enforced.               
const DELETE_WHEN_MINER_FIXED = ONE_ANNE;
const MAX_BALANCE_NQT = MAX_BALANCE_ANNE * ONE_ANNE;
const INITIAL_BASE_TARGET = 38650387592;        //irrelevant
const BLOCK_REWARD_START = 1000_000;
const BLOCK_REWARD_CYCLE = 21240;
const BLOCK_REWARD_CYCLE_PERCENTAGE = 99;
const BLOCK_REWARD_18500_HEIGHT = 8453521;    //mining reward sunset mode 
const BLOCK_REWARD_SUNSET_AMOUNT = 18500;
const BLOCK_REWARD_LIMIT_HEIGHT = 14655601;   //mining reward end block 
const BLOCK_REWARD_LIMIT_AMOUNT = 0;            //0 reward after that
const SUPER_CYCLE = BLOCK_REWARD_CYCLE * 13;    //every 13 cycles is a super block.
const BLOCK_REWARD_SOLO_CUT = 51;
const BLOCK_REWARD_SHARE_CUT = 0.07;             //7% of the block reward.
const BLOCK_REWARD_SHARE_BROADCAST_FEE = 0.02;   //2% of the 7% cut.

const MAINNET = "mainnet";
const TESTNET = "testnet";

const MEDIA_AUDIT_NOTRUN = 0;
const MEDIA_AUDIT_PASSED = 1;
const MEDIA_AUDIT_PASSED_W_NEWDATA = 2;
const MEDIA_AUDIT_MISSING_DATA = -2; //missing data
const MEDIA_AUDIT_CONFLICTING_TYPE = -3; 
const MEDIA_AUDIT_CONFLICTING_SIZE = -4;
const MEDIA_AUDIT_FINGER_DOES_NOT_MATCH = -5;

const FILEDATASTATUS_NOT_SET = -1;
const FILEDATASTATUS_REQUEST_ERROR = 6;
const FILEDATASTATUS_ACQUIRED = 7;
const FILEDATASTATUS_PENDING_ANTOR = 277;
const FILEDATASTATUS_CANNOT_INIT_ANTOR = 287;
const FILEDATASTATUS_ANTOR_OFF = 297;

const QT_PERCENT = 0;
const QT_SPECIFIC = 1;
const SRC_FUND = 0;
const SRC_OWNER = 1;
const SP_MIN_PAYMENT = 1_000_000;  //new min, 1m annies (10k annecoin)


/************** (from file ./js/base/consts.js)**/ 

class anne_singleton {
    constructor(){
        this._bINIT = false;
    }
    set_init = function ($REF) {
        this._bINIT=true;
        Object.freeze($REF);
    }
    is_init = function () {
        if (this._bINIT==true) return true;
        return false;
    }
}

/************** (from file ./js/base/anne_singleton.js)**/ 

/**
 * Basically, an array of objects, and a corresponding map to help manage their
 * addition/tracking/grouping
 */
class arraymap {
    constructor($mapKey="") {
        this.arr = [];
        this.map = new Object();
        this.mapKey = $mapKey;
        this.arr_other = [];

        this.secondary_mapKey = "";
        this.secondary_map = null;
        this.secondary_mapKey_forceLowerCase = false;
    }

    //good enough...
    copy = function($arraymap) {
        this.mapKey = $arraymap.mapKey;
        this.secondary_mapKey = $arraymap.secondary_mapKey;

        var keys = Object.keys($arraymap.arr);
        for (var k=0; k < keys.length; k++) {
            this.arr[keys[k]] = $arraymap.arr[ keys[k]]
        }
        if ($arraymap.secondary_map!=null) {
            var keys2 = Object.getKeys($arraymap.arr_other);
            for (var k=0; k < keys2.length; k++) {
                this.arr_other[keys2[k]] = $arraymap.arr_other[ keys2[k]]
            }
        }
    }


    setMapKey = function($mk) { this.mapKey = $mk;}
    setSecondaryIndex = function($2ndkey, $bMAKELOWER=false) {
        this.secondary_mapKey = $2ndkey;
        this.secondary_mapKey_forceLowerCase = $bMAKELOWER;
        this.secondary_map = new Object();
    }

    //aliases
    setPrimaryKey = function($mk) { this.setMapKey($mk); }
    setSecondaryKey = function($2ndkey, $bMAKELOWER=false) { this.setSecondaryIndex($2ndkey, $bMAKELOWER);}


    addArray = function($inputArr) {
        for(var i=0; i<$inputArr.length; i++) {
            this.addU($inputArr[i]);
        }
    }

    contains = function($mapKey) {
        if (this.map[$mapKey]==undefined) return false;
        return true;
    }


    //keep obj so we have ptr in mem still in other arrays (like tbls.)
    updateU_oldkey = function($oldObj, $newObj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $oldObj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];
        delete this.map[key];
        var newkey = $newObj[this.mapKey];
        
        var keys = Object.keys($newObj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if (keys[k]==this.mapKey) continue;
            var key = keys[k];  
            objToUpdate[key] = $newObj[key];
        }
        this.map[newkey]=objToUpdate;
    }


    getSlice=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get(f);
            if (obj==null) {
                console.log("no f:"+f+" obj in arr map");
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    getSlice2=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get2(f);
            if (obj==null) {
                //console.log("no f:"+f+" obj in arr map");
                //console.warn("big problem");
                continue;
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    updateU = function($obj, $skip_keys=[]) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $obj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];

        var keys = Object.keys($obj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;

            var key = keys[k];
            if ($obj[key]!=undefined) objToUpdate[key] = $obj[key];
        }

        return objToUpdate;
    }

    
    updateU2 = function($obj, $skip_keys=[]) {
        if ($obj==undefined) {
            console.warn("dude, what are you doing?");
            return;
        }
        if (this.secondary_mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }
        
        var key2 = $obj[this.secondary_mapKey];
        if (this.secondary_mapKey_forceLowerCase==true) {
            key2 = key2.toLowerCase();
        }

        if (this.secondary_map[key2]==undefined) return;
        
        var objToUpdate = this.secondary_map[key2];

        var keys = Object.keys($obj)
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;
            var keyx = keys[k];
            objToUpdate[keyx]=$obj[keyx];
        }
    }


    getKeys = function() {
        return Object.getKeys(this.map);
    }

    getU = function($obj) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        var key = $obj[this.mapKey];
        return this.get(key);
    }

    get = function($mapKey) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        if (this.map[$mapKey]!=undefined) return this.map[$mapKey];
        return null;
    }

    get2 = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
            return null;
        }
        if (this.secondary_map[$mapKey2]==undefined) return null;
        return this.secondary_map[$mapKey2];
    }

    get2_partial = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
        }
        var txt =$mapKey2;
        if (this.secondary_mapKey_forceLowerCase==true) txt = $mapKey2.toLowerCase();
        var matches = [];
        var keys = Object.keys(this.secondary_map);
        for (var k=0; k < keys.length; k++) {
            if (keys[k].indexOf(txt)==0 && keys[k].length>txt.length){ 
                matches.push( this.secondary_map[ keys[k]]);
            }
        }
        return matches;
    }


    addIndex2 = function($pkOrObj, $key2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use addIndex2 without a secondary index set!");
        }

        //if first param string, use it as a primary keyvalue of the .map
        //to get the obj. Use that object's key2 to set the secondary map.
        if (typeof $pkOrObj==="string") {
            var obj = this.get($pkOrObj);
            if (obj==null) return null;
            this.secondary_map[$key2]=obj;
        }

        //standard, the pkOrObj is an object, and the $key2 is the value of the secondary mapKey
        //for that object.
        else {

            //audit.
            if ($key2!=$pkOrObj[this.secondary_mapKey]) {
                console.log("houston we have a problem!");
                return;
            }
            this.secondary_map[$key2]=$pkOrObj;
        }
    }

    quickAddMK = function($obj) {
        if (this.mapKey=="") {
            console.warn("U can't use quick add without a quickAddMK set.");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
        }
    }

    quickAdd = function($obj) {
        this.arr.push($obj);
    }
    
    addUnique = function($obj, $mapKeyVal) {
        if ($mapKeyVal==undefined) {
            console.warn("fix this!")
            return;
        }
        if (this.map[$mapKeyVal]==undefined) {
            this.map[$mapKeyVal]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }
        else this.addU2($obj);
    
        return false;
    }

    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    addU = function($obj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addU");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }

        //still add secondary key if relevant.
        else this.addU2($obj);
    
        return false;
    }


    addU2=function($obj) {
        if (this.secondary_map!=null) {
            var key2 = $obj[this.secondary_mapKey];
            if (key2==undefined) return true;
            if (this.secondary_mapKey_forceLowerCase==true) key2 = key2.toLowerCase();  
            var val2 = $obj[key2];
            if (val2==undefined) this.addIndex2($obj, key2);
        }
    }

    
    removeNoKey = function($index) {
        this.arr.splice($index, 1); 
    }

    remove = function($mapKey, $prop="") {
        if ($prop!="") this.prop = $prop;
        if (this.map[$mapKey]!=undefined) {
            delete this.map[$mapKey];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.prop]==$mapKey) {
                    this.arr.splice(i, 1);
                    return;
                }
            }
        }
    }

    getArrayFromAltProp = function($altprop, $altpropvalue) {
        var matches = [];
        for (var i=0; i < this.arr.length; i++) {
            if (this.arr[i][$altprop]==$altpropvalue) {
                matches.push(this.arr[i]);
            }
        }
        return matches;
    }

    removeU = function($mapKeyValue) {
        if (this.map[$mapKeyValue]!=undefined) {
            delete this.map[$mapKeyValue];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.mapKey]==$mapKeyValue) {
                    this.arr.splice(i, 1);
                    return true;
                }
            }
        }
        return false;
    }


    /**
     * 
     * @param {*} $obj 
     * @param {*} $objkey 
     * @returns 
     */
    addArrAM = function($obj, $objkey="txid") {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addArrAM");
            return;
        }
        var mk = $obj[this.mapKey];
        if (this.map[mk]==undefined) {
            this.map[mk]=new arraymap();
            if (this.map[mk].addUnique($obj, $objkey)==true) {
                this.arr.push($obj);
                return true;
            }
            return false;
        }
        if (this.map[mk].addUnique($obj, $objkey)==true) {
            this.arr.push($obj);
            return true;
        }
        return false;
    }


    addArrObj = function($obj, $mapKey) {
        if (this.map[$mapKey]==undefined) {
            this.map[$mapKey]=[$obj];
            this.arr.push($obj);
            return true;
        }
        this.map[$mapKey].push($obj);
        this.arr.push($obj);
        return false;
    }

    sortArrDesc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return 1;
            else if ($a[$prop] > $b[$prop]) return -1;
            return 0;
        })
    }

    sortArrAsc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return -11;
            else if ($a[$prop] > $b[$prop]) return 1;
            return 0;
        })
    }

}

/************** (from file ./js/base/utils/arraymap.js)**/ 

class anne_alert{
    constructor($title, $body, $cb_confirm=null, $cb_no=null, $cb_close=null){
        var el_alert_header = document.getElementById("alerta_header");
        el_alert_header.innerHTML = $title;

        var el_alert_text = document.getElementById("alerta_text");
        el_alert_text.innerHTML = $body;

        var el_alert_left_btns =  ANNE.A.empty("alerta_left_btns")
        var el_alert_right_btns =  ANNE.A.empty("alerta_right_btns")
        
        var el_alert_close = document.getElementById("alerta_header_close");
        el_alert_close.addEventListener("click", function($e){
            if ($cb_close!=null) $cb_close();
            ANNE.A.hide("alerta");
        })

        var btns = [];

        //confirm callback.  (if exists make confirm/yes button)
        if ($cb_confirm!=null) {
            var btnconf = document.createElement("button");
            btns.push(btnconf);
            btnconf.classList = "alerta-btn-confirm";
            btnconf.innerHTML = "yes";
            el_alert_left_btns.append(btnconf);
            el_alert_left_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_confirm();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        //no callback (if exists make no button)
        if ($cb_no!=null) {
            var btnno = document.createElement("button");
            btns.push(btnno);
            btnno.classList = "alerta-btn-no";
            btnno.innerHTML = "no";
            el_alert_right_btns.append(btnno);
            el_alert_right_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_no();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        ANNE.A.fadeIn("alerta");
    }
}


class anne_toggle_obj {
    constructor($divObj) {
        var divObj = $divObj.div;
        var btnObj = $divObj.btn;

        this.div_el = null; //will be set to the el that is made for this anne_toggle_obj

        if (divObj.id!=undefined) this.id = divObj.id;
        else this.id="";

        if (divObj.title!=undefined) this.title = divObj.title;
        else this.title="";
        
        if (divObj.style!=undefined) this.style = divObj.style;
        else this.style="";

        if (divObj.text!=undefined) this.text = divObj.text;
        else this.text="";
        
        if (divObj.div_class!=undefined) this.div_class = divObj.div_class;
        else this.divclass="";
        
        if (divObj.cb!=undefined && typeof divObj.cb==="function") this.cb = divObj.cb;
        else this.cb=null;

        //btn for the toggling.
        this.btn_el = null;
        this.btn_toggle_id = "";
        this.btn_text = "";
        this.btn_class = "";
        this.btn_tnid = "";

        if (btnObj!=undefined) {
            if (btnObj.id!=undefined) this.btn_toggle_id = btnObj.id;
            if (btnObj.text!=undefined) this.btn_text = btnObj.text;
            if (btnObj.btn_class!=undefined) this.btn_class = btnObj.btn_class;
            if (btnObj.btn_tnid!=undefined) this.btn_tnid = btnObj.btn_tnid;   
        }
    }
}


/**
 * 
 */
class anne_toggle_set {
    constructor($objs) {
        this.atos = [];
        if ($objs.length) {
            for (var a=0; a < $objs.length; a++) {
                var ato = new anne_toggle_obj($objs[a]);
                this.atos.push(ato);
            }
        }
    }

    /**
     * 
     */
    makeToggleSetFromAwarePol=function(){
    }

    /**
     * 
     */
    makeToggleSetFromAwareBucket=function(){
    }

}


/**
 * 
 */
class anne_select_option {
    constructor($selVal="", $selDisplayVal="", $titleHTML="", $bodyHTML="", $val2=""){
        this.selval = $selVal;
        this.selDispVal = $selDisplayVal;
        this.titleHTML = $titleHTML;
        this.bodyHTML = $bodyHTML;
        this.val2 = $val2;
        this.gui_id = UTILS.randHex(8);
        this.el_opt = null;
    }
}

class anne_select {
    constructor($w="400px", $el, $options, $misc) {
        this.width = $w;
        this.el = ANNE.A.__getEl($el);
        
        if (this.el == null) return null;
        
        this.options = $options;

        if ($misc.defaultLbl!=undefined) this.defaultLbl = $misc.defaultLbl;
        else this.deafultLbl = "";

        if ($misc.addNewCb!=undefined) this.addNewCb=$misc.addNewCb;
        else this.addNewCb = null;

        if ($misc.cb!=undefined) this.cb = $misc.cb;
        else this.cb=null;

        if ($misc.cb_ddopen!=undefined) this.cb_ddopen = $misc.cb_ddopen;
        else this.cb_ddopen = null;

        if ($misc.cb_ddclose!=undefined) this.cb_ddclose = $misc.cb_ddclose;
        else this.cb_ddclose = null;

        if ($misc.ddclass!=undefined) this.ddclass = $misc.ddclass;
        else this.ddclass = "";

        if ($misc.max_height!=undefined) this.max_height=$misc.max_height;
        //default 660px
    }


}


/******************* anne_dom - a vanilla GUI/DOM stuff *********************
*           lighter, faster, and simpler for my style of coding.            *
****************************************************************************/
class anne_dom extends anne_singleton {
    constructor() {
        super();
    }

    /**
     * Create a button that launches a dialogue w/ various details and callbacks.
     * @param {*} $el 
     * @param {*} $btnClass 
     * @param {*} $title 
     * @param {*} $msg 
     * @param {*} $cbYes 
     * @param {*} $cbNo 
     * @param {*} $cbOk 
     */
    btnToLaunchConfirm=async($el, $btnClass="anne-btn2", $title, $msg, $cbYes=null, $cbNo=null, $cbOk=null)=>{
        this.ezButton($el, "", $title, $btnClass, async($e)=>{
            new anne_alert($title, $msg, $cbYes, $cbNo, $cbOk);
        });
    }


    /**
     * 
     * @param {*} $wPx 
     * @param {function} $clickCB
     * @returns 
     */
    memeframer = function($wPx="900px", $clickCB=null, $name="testjpg") {
        const el_meme = document.createElement("div");
        el_meme.classList="meme-container";
        el_meme.style['max-width']=$wPx;
        el_meme.style['width']=$wPx;
        
        var row = document.createElement("div");
        row.style['min-width']='100%';
        row.classList = "meme-row";
        el_meme.appendChild(row);

        el_meme.setAttribute("file_name", $name);

        var img = document.createElement("img");
        img.src="./imgs/anne_logo.png";
        img.setAttribute("alt", "logo");
        img.classList="meme-logo";
        row.appendChild(img);

        const btn = document.createElement("button");
        btn.classList="meme-btn-make";
        btn.innerHTML="make meme!"
        row.append(btn);
        
        var row2 = document.createElement("div");
        row2.style['min-width']='100%';
        row2.classList = "meme-body";
        el_meme.appendChild(row2);


        if ($clickCB!=null && typeof $clickCB==="function") {
            btn.addEventListener("click", (e)=>{
                ANNE.A.hide(btn);
                var name = el_meme.getAttribute("file_name");
				$clickCB(el_meme, name);
			});
        }

		return {el:el_meme, body:row2};
    }


    buildGridNeuron = function($el, $test, $class, $afile=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-neuron-holder");
        var neuron = document.createElement("div");
        neuron.classList.add("grid-neuron");
        
        if ($afile!=null) {
            ANNE.A.doImageB64(neuron, "", $afile, 0, 99, "", 99);
        }
        else {
			neuron.innerHTML = $test;
			neuron.classList.add($class);
		}
        el.appendChild(neuron);
    }


    buildGridArrow=function($el, $arrowClass, $lbl="test") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-"+$arrowClass); 
        var lbl = document.createElement("div");
        lbl.innerHTML = $lbl;
        lbl.classList.add("grid-arrow-lbl");
        lbl.classList.add($arrowClass+"-text");
        el.appendChild(lbl);
        return;
    }


    /**
     * [x][->][x][<-][x]
     * @param {*} $el 
     */
    build1con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        var grid = this.buildDOMGrid(el, 5, 1);
        console.log(grid);
        
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow", $m0[0].afile);
        this.buildGridArrow(grid["i1"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m0[1].text, "ggreen", $m0[1].afile);
        this.buildGridArrow(grid["i3"], "left", $m1[0].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue", $m1[0].afile);
        grid.parent_el = el;
        return grid;
    }
    

    build1con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "ne", "");
        this.buildGridArrow(grid["i6"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i7"], $m0[1].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");

        this.buildGridArrow(grid["i13"], "horizontal", $m1[1].reln_text);
        this.buildGridArrow(grid["i12"], "left-up", "");
        
        grid.parent_el = el;
        return grid;
    }



    build2con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gblue");
        this.buildGridArrow(grid["i11"], "horizontal", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-up", "");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "nw", $m1[0].reln_text);
        this.buildGridArrow(grid["i8"], "left", "");
        this.buildGridNeuron(grid["i7"], $m0[2].text, "ggreen");
        
        grid.parent_el = el;
        return grid;
    }

    /**
     * [a]     [b]
     *  |       |
     * [c]     [d]
     *  |       |   
     *   ->[e]<-
     * @param {*} $el 
     * @param {*} $w 
     * @returns 
     */
    build2con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "ne", "");
        this.buildGridArrow(grid["i16"], "right", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "nw", "");
        this.buildGridArrow(grid["i18"], "left", $m0[1].reln_text);
        

        grid.parent_el = el;
        return grid;
    }

    build3con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "horizontal", $m0[2].reln_text);
        this.buildGridArrow(grid["i22"], "right-up", "");
        this.buildGridNeuron(grid["i17"], $m0[3].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i13"], "left", "");
        this.buildGridArrow(grid["i12"], "left-down", "");
        
        
        grid.parent_el = el;
        return grid;
    }


    build2con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i11"], "right", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
        this.buildGridArrow(grid["i22"], "left-up", "");
    
        grid.parent_el = el;
        return grid;
    }


    build3con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
    
        grid.parent_el = el;
        return grid;
    }


    /**
     * [x0]   [y1]-->[y3]
     *  |      |      |
     * [x1]   [y0]   [y4]
     *  |             |
     * [x2]-->[x3]->[match]  
     * @param {*} $el 
     * @param {*} $m0 
     * @param {*} $m1 
     * @param {*} $w 
     * @returns 
     */
    build4con4con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "gyellow");
        this.buildGridArrow(grid["i23"], "right", $m0[3].reln_text);
        this.buildGridNeuron(grid["i24"], $m0[4].text, "ggreen");

        
        this.buildGridNeuron(grid["i12"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i7"], "up", $m1[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i3"], "right", $m1[1].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[2].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[3].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[3].reln_text);

        grid.parent_el = el;
        return grid;
    }



    buildDOMGrid=function($el, $flexWidth = 5, $rows = 5, $pre="gridx") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var grid = new Object();
        el.classList.add("grid-container");
        for (let i = 0; i < $flexWidth * $rows; i++) {
            const gridSquare = document.createElement('div');
            gridSquare.classList.add('gridSquare');
            gridSquare.style.setProperty('--flexWidth', $flexWidth);
            el.appendChild(gridSquare);
            var r = Math.floor((i+$flexWidth)/$flexWidth)-1;
            var row = grid["r"+r];
            if (row==undefined) {
                row = new Object();
                grid["r"+r]=row;
            }
            var c = (i+$flexWidth) % $flexWidth;
            el.setAttribute("id",$pre+"_"+r+"_"+c)
            row["c"+c] = gridSquare;
            grid["i"+i]=gridSquare;
        }

        return grid;
    }


    
    /**
     * Make a nifty little HTML DD for the users DD.
     * //works with .text/.nickname and .nid/.public_nid
     * @param {*} $el 
     * @param {*} $users 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromUserObjs = function($w="400px", $el, $users, $configs) {
        var options = [];
        for (var n=0; n < $users.length; n++) {
            var user = $users[n];
            var nid = user.public_nid;
            if (nid==undefined) nid = user.nid;
            if (nid==undefined || nid==null) {
                continue;
            }
            var userNOBJ = MEMG.getOrMakeIfNew(nid);
            var lastKnownAnnies = user.last_known_annies;
            if (lastKnownAnnies==undefined || lastKnownAnnies==null) lastKnownAnnies = "0";
            //else 
            var titleHTML = "";
            var bodyHTML = "";
            var sel_disp_val = "";
            var text = user.text;
            if (text==undefined || text==null || text == "") {
                text = user.nickname;
                if (text==undefined || text==null) text="";
            }
            if (text!="") {
                titleHTML ='<span class="acct-name">'+text+'</span>';
                sel_disp_val = text;
            }
            else {
                titleHTML ="<span class='acct-name'>"+nid+"</span>";
                sel_disp_val = nid;
            }

            bodyHTML = "<span class='bal-small bal-"+userNOBJ.temp_nid+"'>"+UTILS.displayAnneValue(lastKnownAnnies)+"<span style='font-size:1.6rem;'>&#x1d110;</span></span>";
            
            var opt = new anne_select_option(nid, sel_disp_val, titleHTML, bodyHTML);
            options.push(opt);
            user.gui_id = opt.gui_id;
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
    }


    
    /**
     * make a dropdown of annodes
     */
    makeANNODEDD = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:ANNE._annodesLS.last_uri,
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item")
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
                if ($annode_nid!=undefined && $annode_nid!=null && $uri!=undefined && $uri!=null) {
                    if (typeof ANNE.setNewAnnodeURI==="function") {
                        ANNE.setNewAnnodeURI($annode_nid, $uri);
                    }
                }
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };
        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);
    }

    removeAllButTopOption = function($el) {
        var el = this.__getEl($el);
        if (el==null) return;
        const optionsCount = el.options.length;
        for (let i = optionsCount - 1; i > 0; i--) {
            el.remove(i);
        }
    }

    getDDSelectedValue=function($el) {
        var el = this.__getEl($el);
        if (el==null) return null;
        if (el.options==undefined) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.value;
    }
    getDDSelectedAttr=function($el, $attr) {
        var el = this.__getEl($el);
        if (el==null) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.getAttribute($attr);
    }

    setDDToIndex=function($el, $index=0) {
        var el = this.__getEl($el);
        if (el==null) return;
        el.selectedIndex = $index;
    }
    setDDToValue=function($el, $value) {
        var el = this.__getEl($el);
        if (el == null) return;
        for (var i = 0; i < el.options.length; i++) {
            if (el.options[i].value == $value) {
                el.selectedIndex = i;
                return;
            }
        }    
    }


    spinnerInject = function($el, $bWIPE=false) {
        var el = this.__getEl($el);
        if (el==null) return;
        if (el.id==null || el.id=="") el.id = ANNE.UTILS.randHex(4);
        var el_spinner = document.createElement("span");
        el_spinner.id = el.id+"_s";
        el_spinner.classList = "inlinespinner";
        if ($bWIPE==true) ANNE.A.emptySelf(el);
        el.appendChild(el_spinner);
        ANNE.A.show(el);
    }
    
    spinnerRemove = function($el, $bREMOVEPARENT=false) {
		var el = this.__getEl($el);
        if (el==null) return;
        var el_spinner = document.getElementById(el.id+"_s");
        if(el_spinner==null) {
            var inlinespinner = el.querySelectorAll(".inlinespinner");
            if (inlinespinner!=null) ANNE.A.removeSelf(inlinespinner);
        }
        else {
            ANNE.A.removeSelf(el_spinner);
        }
        if ($bREMOVEPARENT==true) {
            ANNE.A.removeSelf(el);
        }
    }

    makeANNODEDD_first_time = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:"select annode",
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item");
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };

        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);

    }



    /**
     * 
     * @param {*} $bDEFOCUS 
     */
    smartDefocus = async($bDEFOCUS=true)=> {
        if ($bDEFOCUS==true) {
            var el = ANNE.A.addClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.setAttribute("disabled",true);
        }
        else {
            var el = ANNE.A.removeClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.removeAttribute("disabled",false);
        }
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $annodes 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromAnnodeObjs = function($w="400px", $el, $annodes, $configs) { 
        var options = [];
        for (var n=0; n < $annodes.length; n++) {
            var annode = $annodes[n];
            var titleHTML = annode.nickname+' - '+annode.nid;
            var bodyHTML = annode.uri;
            var sel_disp_val = annode.nickname+" "+annode.uri;
            var opt = new anne_select_option(annode.nid, sel_disp_val, titleHTML, bodyHTML, annode.uri);
            options.push(opt);
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
        return ASELECT;
    }



    /**
     * 
     * @param {*} $el 
     * @param {*} $options
     * @param {*} config $defaultLbl, $cb
     * 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     * @returns 
     */
    makeHtmlDD=function($ASELECT) {
        if ($ASELECT.el==null) return;
        var sb = null;  //select box

        if ($ASELECT.el.classList.contains("selectbox")==false) {
            sb = document.createElement("div");
            sb.classList = "selectbox";
            $ASELECT.el.appendChild(sb);
        }
        else sb = $ASELECT.el;

        sb.style.width = $ASELECT.width;
    
        if ($ASELECT.addNewCb!=null) {
            //var plusSpan = document.createElement("span");
            //$ASELECT.el.parentNode.appendChild(plusSpan);
            //plusSpan.innerHTML = "[+]";
            //plusSpan.classList = "btn-select-add-option"; 
        }
        
        const optionsContainer = document.createElement("div");
        optionsContainer.classList = "options-containersb";
        sb.appendChild(optionsContainer);

        if ($ASELECT.max_height!=undefined) {
            sb.style['max-height']=$ASELECT.max_height;
            optionsContainer.style['max-height']=$ASELECT.max_height;
            optionsContainer.setAttribute("mh_backup", $ASELECT.max_height);
        }
		else {
			sb.style['max-height']='50vh';
            optionsContainer.style['max-height']='50vh';
            optionsContainer.setAttribute("mh_backup", '50vh');
		}

        const selectedsb = document.createElement("div");
        selectedsb.classList = "selectedsb";

        if ($ASELECT.ddclass!="") {
            selectedsb.classList.add($ASELECT.ddclass);
        }
        selectedsb.style.width = $ASELECT.width;

        sb.appendChild(selectedsb);
        if ($ASELECT.defaultLbl!="") {
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3 pr30";
            lbl_sp3.innerHTML = $ASELECT.defaultLbl;
            selectedsb.append(lbl_sp3);
        }

        //create options
        for (var i=0; i  < $ASELECT.options.length; i++) {
            var opt = $ASELECT.options[i];
            var selval = opt.selval;
            var val2 = opt.val2;
            const divopt = document.createElement("div");
            optionsContainer.appendChild(divopt);
            divopt.classList = "optionsb";
            if (selval!=null) divopt.setAttribute("selval", selval);
            if (val2!=null) divopt.setAttribute("val2", val2);
            if (opt.selDispVal!=null) divopt.setAttribute("dispval", opt.selDispVal);
            
            var divopt_rad = document.createElement("input");
            divopt_rad.setAttribute("type", "radio");
            divopt_rad.classList = "radiosb";
            divopt_rad.id = opt.gui_id;
            divopt.appendChild(divopt_rad);
            
            var divopt_lbl = document.createElement("label");
            divopt_lbl.setAttribute("for", divopt_rad.id);
            divopt_lbl.classList = "opt-lbl";
            divopt.appendChild(divopt_lbl);
            
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3";
            lbl_sp3.innerHTML = opt.titleHTML;
            divopt_lbl.append(lbl_sp3);
            
            var lbl_sp5 = document.createElement("span");
            lbl_sp5.classList = "sbh5";
            lbl_sp5.innerHTML = opt.bodyHTML;
            divopt_lbl.append(lbl_sp5);
            opt.el_opt = divopt;

            opt.el_opt.addEventListener("click", ($e) => {
				console.log("(option clicked!)");
                $e.stopPropagation();
                $e.preventDefault();
				var selval = divopt.getAttribute("selval");
                var val2 = divopt.getAttribute("val2");
                if (val2==undefined) val2 = null;
                var dispval =  divopt.getAttribute("dispval");
				//selectedsb.innerHTML = divopt.querySelector("label").innerHTML;
				//selectedsb.innerHTML = selval;
                ANNE.A.empty(selectedsb);
                var lbl_sp3 = document.createElement("span");
                lbl_sp3.classList = "sbh3 pr30";
                if (dispval) lbl_sp3.innerHTML = dispval;
                else lbl_sp3.innerHTML = selval;
                selectedsb.append(lbl_sp3);

                document.removeEventListener("click", clickout);

                optionsContainer.classList.add("hide");
                optionsContainer.style['z-index']=-1;
				optionsContainer.classList.remove("active");
                optionsContainer.style['pointer-events']='none';

                if ($ASELECT.cb!=null) {
                    if ($ASELECT.cb_ddclose!=null) {
                        $ASELECT.cb_ddclose();
                    }
                    $ASELECT.cb(selval, val2);
                }
			});
        }

        const clickout = function(){
            console.log("clickout");
            optionsContainer.style['z-index']=-1;
            optionsContainer.style['max-height']=0;
            optionsContainer.style['pointer-events']='none';
            optionsContainer.classList.add("hide");
            document.removeEventListener("click", clickout);
            optionsContainer.classList.toggle("active");
            if ($ASELECT.cb_ddclose!=null) {
				$ASELECT.cb_ddclose();
	        }
		}

        optionsContainer.addEventListener("mouseout", $e=> {
            if (optionsContainer.classList.contains("active")==true) {
                const fromElement = $e.relatedTarget;
                if (!optionsContainer.contains(fromElement)) {
                    clickout();
                }
            }
        });

        selectedsb.addEventListener("click", ($e) => {
            $e.stopPropagation();
            $e.preventDefault();

            optionsContainer.classList.remove("hide");
            optionsContainer.style['pointer-events']='all';

            //this will add hide.
			if (optionsContainer.classList.contains("active")==true) {
                console.log("oc active, make inactive");
                clickout();
            }

            //this will have no hide.
            else {
                console.log("oc inactive, make active");
                document.addEventListener("click", clickout);
                optionsContainer.style['z-index']=9999;
                optionsContainer.style['opacity']=1;
                var mh = optionsContainer.getAttribute("mh_backup");
                optionsContainer.style['max-height']=mh;
                optionsContainer.classList.toggle("active");
                if ($ASELECT.cb_ddopen!=null) {
					$ASELECT.cb_ddopen();
				}
            }   
           
		});
    }

    

    /**
     * Standardized link color maker based on the neuron (not the relon)
     * This will be used to generated links for neurons.
     */
    getLinkColorClass = function($nObj) {
        if ($nObj==undefined || $nObj==null) return "link-black";
        if ($nObj.ghost==true) return "link-blue";

        //presume ANNE ownership if not set (in the js)
        if (ANNE.isANNE($nObj.owned_by)==true 
            || $nObj.owned_by==ANNE.MEM.ANNE_ACCTID) return "link-yellow";  //anne yellow
        
        else {
            if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
                //possible hanai could be here.  Look for pending hanai.
                if ($nObj.bPENDING_HANAI==undefined) {
                    var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
                    if (pending_hanai!=null) {
                        $nObj.bPENDING_HANAI=true;
                        return "link-pink";  //hanai pink
                    }
                }
                return "link-red";   //red for self owned
            }
            else {
                return "link-grey";  //black for someone else's neuron (not ANNE's).
            }
        }
    }


    /**
     */
    getNeuronColorClass = function($nObj, $pre="title-", $post, $default="other") {
        if ($pre!="" && $pre[ $pre.length-1]!="-") $pre = $pre+"-";
        if ($post!="" && $post[ $post.length-1]!="-") $post = "-"+$post;

        if (ANNE.isANNE($nObj.owned_by)==false && $nObj.ghost==false && $nObj.bPENDING_HANAI==undefined) {
            var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
            if (pending_hanai!=null) {
                $nObj.bPENDING_HANAI=true;
                return $pre+"hanai"+$post;  
            }
        }
        else if ($nObj.bPENDING_HANAI==true) return $pre+"hanai"+$post; 

        else if ($nObj.ghost==true) {
            return $pre+"ghost"+$post;  
        }
        else if (ANNE.isANNE($nObj.owned_by)==true) {
            return $pre+"anne"+$post;  
        }
        else if ($nObj.owned_by==USER._userNObj.nid) {
            return $pre+"self"+$post; 
        }
        return $pre+$default+$post;  //hanai pink
    }


    /**
     */
    addClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($classname) == false) el.classList.add($classname);
        return el;
    }

    
    /**
     * Quickly add an ai-ico class image (see anne-icos.css).
     * 
     * @param {*} $el 
     * @param {*} $icoClass 
     * @returns 
     */
    addIco = function ($el, $icoClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $icoClass;
        el.appendChild(span);
    }


    addRemoveClass = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }

    addRemoveClasses = function ($el, $addclass, $removeClasses) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var r = 0; r < $removeClasses.length; r++) {
            if (el.classList.contains($removeClasses[r]) == true) el.classList.remove($removeClasses[r]);
        }
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        return el;
    }

    /**
     * Same as addRemoveClass but checks if the removeclass is on.
     */
    addRemoveClassIf = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) {
            if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        }
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }


    animateProp = function ($el, $prop, $propVal, $timestring = "0.5s ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        el.style[$prop] = $propVal;
        el.style.transition = $prop + " " + $timestring;
    }
    animateProps = function ($el, $jsonprops, $dur = ".5s", $delay = "0", $ease = "ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        var keys = Object.keys($jsonprops);
        var propstrs = [];
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            propstrs.push(prop);
        }
        el.style['transition-property'] = propstrs.join(",");
        el.style['transition-duration'] = $dur;
        el.style['transition-timing-function'] = $ease;
        if ($delay != "") el.style['animation-delay'] = $delay;

        //set the props after the transition styling.
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            el.style[prop] = val;
        }
    }





    /**
     * Disambig chases of information from external lookups.
     * @param {*} $neurons 
     * @param {*} $val 
     * @param {*} $ambigObj 
     * @param {*} $el_console 
     * @param {*} $resolveAmbigs 
     * @returns 
     */
    disambigInner=async($neurons, $val, $ambigObj, $el_console, $resolveAmbigs)=> {
        if ($neurons==undefined || $neurons==null || $neurons.length==0) {
            return $resolveAmbigs(); //call the next one.
        }
        var el_dchoice = document.createElement("div");
        var sp = document.createElement("div");
        sp.innerHTML="original text: <b>"+$val+"</b>";
        sp.style.display="block";
        el_dchoice.appendChild(sp);

        //the choice (and all options)
        $el_console.appendChild(el_dchoice);

        for (var n=0; n < $neurons.length; n++) {
            var el_choice = document.createElement("div");
            el_choice.classList = "neuron-choice";
            el_dchoice.appendChild(el_choice);
            
            var el_choice_info = document.createElement("div");
            el_choice_info.classList = "neuron-choice-info";
            
            var nObj = $neurons[n];
            await ANNE.GUI_STUFF.disambigHTMLStr(el_choice_info, nObj, ["sp"]);
            el_choice.appendChild(el_choice_info);
            
            var el_choice_btn = document.createElement("div");
            el_choice_btn.classList = "neuron-choice-btn-holder";  
            el_choice.appendChild(el_choice_btn);
            
            var el_btn = ANNE.A.ezButton(el_choice_btn, "", "choose match "+(n+1), "anne-tiny-btn btn-purple-sq btn-maxH", async(e)=>{
                var nid_x = e.target.getAttribute("nid");
                var nObjx = MEMG.getNeuronByNid(nid_x);
                
                ANNE.A.empty(e.target.parentElement, true);//empty and hide choice div.
                
                //key to set this.
                $ambigObj.jsonobj.to_s = nObjx.nid;
                return await $resolveAmbigs(); //call the next one.
            });
            el_btn.setAttribute("nid", nObj.nid);
            el_btn.setAttribute("val", $val);
            $el_console.appendChild(el_dchoice);
        }

        //skip option.
        var el_skip = document.createElement("div");
        el_skip.classList = "neuron-choice";
        el_dchoice.appendChild(el_skip);

        var el_skip_info = document.createElement("div");
        el_skip_info.classList = "neuron-choice-skip";
        el_skip.appendChild(el_skip_info);     
        var el_skip_btn = document.createElement("div");
        el_skip_btn.classList = "neuron-choice-btn-holder";  
        el_skip.appendChild(el_skip_btn);
        var el_btn = ANNE.A.ezButton(el_skip_btn, "", "SKIP THIS", "anne-tiny-btn btn-red-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true);//empty and hide choice div.
            $ambigObj.jsonobj.to_s = null;
            return await $resolveAmbigs(); //call the next one.
        });
        

        //make new
        var el_new = document.createElement("div");
        el_new.classList = "neuron-choice";
        el_dchoice.appendChild(el_new);
        var el_new_info = document.createElement("div");
        el_new_info.classList = "neuron-choice-new";
        el_new.appendChild(el_new_info);     
        var el_new_btn = document.createElement("div");
        el_new_btn.classList = "neuron-choice-btn-holder";  
        el_new.appendChild(el_choice_btn);
        var el_btn = ANNE.A.ezButton(el_choice_btn, "", "NEW", "anne-tiny-btn btn-blue-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true); //empty and hide choice div.
            console.log($ambigObj);
            $ambigObj.jsonobj.to_s = null;
            //this is going to need a callback to make a "NEW" correctly!!!!
            return await $resolveAmbigs(); //call the next one.
        });
        $el_console.appendChild(el_dchoice);
    }


    /**
     * Generalized ez annies entry.
     * i_min_val, i_max_val and i_placeholder attrs can be added to the outter button.
     * They will passthrough to the input field.
     * 
     * @param {*} $el 
     * @param {*} $cb_submit - submit callback
     * @param {*} $btnTitle - title of outter button
     * @param {*} $cb_entry - callback on entry of data (if more than just formatting of annies is desired)
     * @param {*} $idpre - Used to create a unique ezAnnecoin_entry if there is more than one in the DOM at once.
     * @returns 
     */
    ezAnnecoin_entry = function($el, $cb_submit=null, $btnTitle="accept available risk", $btn2Title="accept risk", 
                                $cb_entry=null, $idpre="unique", $btnClassStr = "noround btn-someanniesaction") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_holder = ANNE.A.ezEl(el, "div", "flexrow");
        const idpre = $idpre;
        const btn_aar = ANNE.A.ezButton(el_holder, "btn_main_"+$idpre, $btnTitle, $btnClassStr, async(e)=>{
            
            //snag passthrough attrs we put on button AFTER we created here.
            //these will be ready for this event since we created them outside of this funct on the return
            //of the gui creation. (optional attrs)
            var min_val = btn_aar.getAttribute("i_min_val");
            var max_val = btn_aar.getAttribute("i_max_val");
            var placeholder = btn_aar.getAttribute("i_placeholder");

            //create the btn clone so we get the button input button functionality
            var btnClone = btn_aar.cloneNode(true);
            const disp = document.createElement("div");
            btnClone.id = "btn_"+idpre+"_final";
            btnClone.setAttribute("tied_input", "input_"+idpre);
            ANNE.A.hide(btn_aar);
    
            const input_field = document.createElement("input");
            input_field.setAttribute("size", 24);
            input_field.setAttribute("type", "text");
            input_field.id="input_"+idpre;
            input_field.setAttribute("el_display", "annies_"+idpre+"_disp");
            input_field.setAttribute("el_submit", "btn_"+idpre+"_final");
            if (min_val!=undefined) input_field.setAttribute("min_val", min_val);
            if (max_val!=undefined) input_field.setAttribute("max_val", max_val);
            if (placeholder!=undefined) input_field.setAttribute("placeholder", placeholder);
            
            //add th input field after the btn_aar
            btn_aar.after(input_field);

            var max_btn = null;
            var min_btn = null;


            //add max btn
            if (max_val!=undefined) {
                max_btn = document.createElement("button");
                max_btn.innerHTML="max";
                max_btn.classList = 'btn-minmax';
                max_btn.addEventListener("click", function($e){
                    input_field.value = max_val;
                    ANNE.A.hide(max_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(max_btn)
            }

            //add max btn
            if (min_val!=undefined) {
                min_btn = document.createElement("button");
                min_btn.classList = 'btn-minmax';
                min_btn.innerHTML="min";
                min_btn.addEventListener("click", function($e){
                    input_field.value = min_val;
                    ANNE.A.hide(min_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(min_btn)
            }


            //add input input event
            input_field.addEventListener("input", async($e)=>{
                await ANNE.A.anne_entered($e);
                if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
            });


            //a display for the formatted/summary version of what the user typed.
            disp.classList="flexrow";
            disp.id="annies_"+idpre+"_disp";

            //if we have max, add disp after it.  else add disp after input field.
            if (max_btn!=null) max_btn.after(disp);
            else if (min_btn!=null) min_btn.after(disp);
            else input_field.after(disp);

            disp.after(btnClone);
            btnClone.innerHTML=$btn2Title;
            ANNE.A.hide(btnClone);

            //add submit callback.
            if (typeof $cb_submit=="function") {
                btnClone.addEventListener("click", async($e)=>{
                    var val = input_field.value;
                    $cb_submit(val)
                });
            }
        });
        return el_holder;
    }


    /**
     * annies entered
     * This can handle el-display (and display value entered)
     * This can handle el-submit (and show a submit btn after valid entry)
     * @param {*} $e 
     */
    anne_entered = async($e, $target=null)=>{
        var el_input = $target;
        if (el_input==null && $e!=null && $e.target!=null) {
            el_input = $e.target;
        }
        if (el_input==null) return;
        
        var val = el_input.value+"";
        var lastChar = val[val.length-1];

        var el_display = el_input.getAttribute("el_display");
        if (el_display!=null) el_display = document.getElementById(el_display);

        var el_submit = el_input.getAttribute("el_submit");
        if (el_submit!=null) el_submit = document.getElementById(el_submit);

        var minval = el_input.getAttribute("min_val");
        var maxval = el_input.getAttribute("max_val");
        
        var final_val = 0;

        if (lastChar=="k" || lastChar=="K") {         //shorthand for thousands
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000;
            el_input.value = newval;
        }
        else if (lastChar=="m" || lastChar=="M") {    //shorthand for millions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000;
            el_input.value = newval;
        }
        else if (lastChar=="b" || lastChar=="B") {    //shorthand for billions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000000;
            el_input.value = newval;
        }
        else if ("0123456789.,".indexOf(lastChar)==-1) {
            var newval = "";
            var amt = el_input.value;
            for (var a=0; a < amt.length; a++) {
                if ((a!=(amt.length-1)) && ("kKmMbB".indexOf(amt[a])!=-1)) continue;
                else if ("0123456789.,".indexOf(amt[a])!=-1) continue;
                newval = newval+amt[a];
            }
            el_input.value = newval;
        }

        var val  = el_input.value;
        if (val=="") {
            return;
        }
        final_val = parseInt(val);

        if (minval!=undefined && final_val < minval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='min-msg'>minimum "+minval+" annecoin</span>"); 
        }

        else if (maxval!=undefined && final_val > maxval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='max-msg'>max "+maxval+" annecoin</span>"); 
        }

        else if (final_val>0) {
            if (el_submit!=null) ANNE.A.show(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, final_val, true); 
        }
        else {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, 0, true); 
        }
    }

    appendHTML = function($el, $html) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = el.innerHTML+$html;
    }

    append = function ($el, $misc, $bSHOW = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.append($misc);
        if ($bSHOW == true) this.show(el);
        return el;
    }

    appendStringList=function($el, $strArr, $size="10px", $class="") {
        var el = this.__getEl($el);
        if (el==null) {
            console.log("missing el");
            return null;
        }
        if ($strArr==undefined || $strArr==null || Array.isArray($strArr)==false || $strArr.length==0) return;
        var spanlist=document.createElement("span");
        spanlist.style['display']="inline-block";
        spanlist.style['font-size']=$size;
        if ($class!="") spanlist.classList.add($class);
        if ($class!="") {
            spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        }
        else spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        el.append(spanlist);
    }

    addEmojisFromRelTos=function($el, $rels, $title) {
        var el = this.__getEl($el);
        if (el == null) return null;

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var sp = document.createElement("span");
                sp.innerHTML = toNOBJ._uico;
                sp.setAttribute("title", $title+" "+toNOBJ.bestNameHTML());
                sp.setAttribute("txid", rel.txid);
                el.appendChild(sp);
            }
        }
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    getAnum=function($str) {
        let annum = 0;
        const normalizedText = $str.normalize("NFD").replace(/[\u0300-\u036f]/g, "").toLowerCase().trim();
        for (let i = 0; i < normalizedText.length; i++) {
            const c = normalizedText.charAt(i);
            const asciiValue = c.charCodeAt(0);
            if (asciiValue < 49) continue;
            if (asciiValue <= 57) {
                annum += asciiValue - 48;
                continue;
            }
            else if (asciiValue < 97 || asciiValue > 123) {
                continue;
            }
            else annum += asciiValue - 96;
        }
        return annum;
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $text 
     * @param {*} $classpost 
     */
    doAnumWithCalc = function($el, $text, $classpost="") {
        var num = this.getAnum($text+"");
        this.doAnum($el, num+"", $classpost);
        return num;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $num 
     * @param {*} $classpost 
     * @returns 
     */
    doAnum=function($el, $num, $classpost="") {
        var el = this.__getEl($el);
        if (el == null) return null;

        if(el.classList.contains("anumbox"+$classpost)==false) el.classList.add("anumbox"+$classpost);
        el.innerHTML="";

        $num=$num+"";
        
        for (var n=0; n <$num.length; n++) {
            var sp = document.createElement("span");
            sp.classList = "anum"+$classpost;
            sp.innerHTML = $num[n];
            el.appendChild(sp);
        }
        return;
    }


    /**
     * 
     * @param {*} $rels 
     * @param {*} $title 
     * @param {*} $skips 
     * @returns 
     */
    getEmojisFromRelTos_html=function($rels, $title, $skips=[]) {
        if (ANNE.SETTINGS && ANNE.SETTINGS.bUICOS==false) return "";
        var html = "";
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if ($skips.indexOf(rel.txid)!==-1) continue;
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var span = "<span title='"+$title+" "+toNOBJ.bestNameHTML()+"' txid='"+rel.txid+"'>"+toNOBJ._uico+"</span>";
                html+=span;
                $skips.push(rel.txid);
            }
        }
        return html;
    }


    click = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.click();
        return el;
    }

    //this clones the DOM node.
    //DOM node doesn't have to be in the DOM.. it can just be a temp made
    //el that is not yet appended.
    cloneAndAppend = function ($getel, $putel, $bCLONE = true, $bSHOWPUTEL = true) {
        if ($bCLONE == true) {
            var getel = this.__getEl($getel);
            if (getel == null) return null;
            var clone = getel.cloneNode(true);
        }
        else clone = this.__getEl($getel);
        if (clone == null) return null;
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }

    //this clones the .content of the template
    cloneAndAppendTemplate = function ($getel, $putel, $bSHOWPUTEL = true) {
        var getel = this.__getEl($getel);
        if (getel == null) return null;
        var clone = getel.content.cloneNode(true);
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }



    createElement = function ($el, $tag, $class = "", $html = "", $value = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newtag = document.createElement($tag);
        if ($class != "") newtag.classList = $class;
        if ($html != "") newtag.innerHTML = $html;
        if ($value != "") newtag.value = $value;
        el.append(newtag);
        return newtag;
    }

    /**
     */
    doImageDyno = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile - ANNE FILE
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDynoB64 = function ($div, $imgElId = "", $afile=null, $imgClass = "left_dyno_w") {
        
        if ($afile == null) return null;
        if ($afile.b64data==null) return null;

        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2 = function ($div, $imgElId = "", $src = "", $biggest = 200, $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }
            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $afile 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2_B64 = function ($div, $imgElId = "", $afile = null, $biggest = 200, $imgClass = "left_dyno_w") {
        if ($afile == null) return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }

            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     */
    doBadImg = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        var newImgEl = this.ezEl($div, "span", $imgClass + " img-bad-audit", "FAILED<BR>AUDIT");
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        newImgEl.style.top = "0px";
        if ($src != undefined && $src != null && $src != "") {
            newImgEl.setAttribute('bad-audit-url', $src);
        }
        var attr_fix = newImgEl.getAttribute("ssfixpostload");
        if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        return newImgEl;
    }



    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $w 
     * @param {*} $h 
     * @param {*} $imgClassAdd 
     * @param {*} $max 
     * @returns 
     */
    doImage = function ($div, $imgElId, $src, $w = 600, $h = 900, $imgClassAdd = "", $max = 0) {
        console.log("src "+$src);
        
        const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);
        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
        newImgEl.onerror = function(image) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($max && ($w / ratio) > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", $w);
                }
                else {
                    if ($max!=0 && realH > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", realW);
                }
            }
            else if ($h) {
                if ($h < realH) {
                    if ($max && (ratio * $h) > $max) {
                        newImgEl.setAttribute("width", $max);
                    }
                    newImgEl.setAttribute("height", $h);
                }
                else newImgEl.setAttribute("height", realH);
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile 
     * @param {int} $w 
     * @param {int} $h 
     * @param {string} $imgClassAdd 
     * @param {int} $maxOtherD 
     * @returns 
     */
    doImageB64 = function ($div, $imgElId, $afile=null, $w = 600, $h = 900, $imgClassAdd = "", $maxOtherD = 0) {      
		const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);

		if ($afile==null) {
			console.log("must have an afile here");
			return newImgEl;
		}

        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
		
        newImgEl.onerror = function($img) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($maxOtherD && ($w / ratio) > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                    } 
                    else {
                        newImgEl.setAttribute("width", $w);
                        newImgEl.setAttribute("height", $w / ratio); 
                    }
                } 
                else {
                    if ($maxOtherD != 0 && realH > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                        newImgEl.setAttribute("width", $maxOtherD * ratio); 
                    } 
                    else {
                        newImgEl.setAttribute("width", realW);
                        newImgEl.setAttribute("height", realH);
                    }
                }
            } 

            else if ($h) {
                if ($h < realH) {
                    if ($maxOtherD && (ratio * $h) > $maxOtherD) {
                        newImgEl.setAttribute("width", $maxOtherD);
                        newImgEl.setAttribute("height", $maxOtherD / ratio); // Adjust height to maintain aspect ratio
                    } 
                    else {
                        newImgEl.setAttribute("height", $h);
                        newImgEl.setAttribute("width", ratio * $h); // Adjust width to maintain aspect ratio
                    }
                } 
                else {
                    // If the specified height is greater than or equal to the original height
                    newImgEl.setAttribute("height", realH);
                    newImgEl.setAttribute("width", realW);
                }
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;

		if ($afile.b64data==undefined || $afile.b64data==null) {
			console.log("problem with b64 data");
			return newImgEl;
		}
		
        if ($afile.b64data.indexOf("data:")!=-1) {
			newImgEl.src = $afile.b64data;
		}
        else {
			newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
		}
        return newImgEl;
    }

    emptySelfAfterChild=function($el, $childIndex=0, $bHIDE=false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var c=0; c < el.children.length; c++) {
            if (c<=$childIndex) continue;
            el.removeChild( el.children[c]);
            c=c-1;
        }
        if ($bHIDE == true) this.hide($el);
        return el;
    }

    emptySelf = function($el, $bHIDE=false){ return this.empty($el, $bHIDE);}   //i keep thinking it is named this.
    empty = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        while (el.firstChild) el.removeChild(el.firstChild);
        el.innerHTML = "";
        if ($bHIDE == true) this.hide($el);
        return el;
    }
    emptyParent = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        if (el_parent) this.empty(el_parent);
        else return el;
        if ($bHIDE == true) this.hide(el_parent);
        return el_parent;
    }

    emptyAllBut=function($el, $skipEl, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        for (var c=0; c < el_parent.children.length; c++) {
            var child = el_parent.children[c];
            if (child==$skipEl) continue;
            el_parent.removeChild(child);
        }
        if ($bHIDE == true) this.hide(el);
        return;
    }


    ezAlert = function ($el, $class, $id_pre, $id, $msg = "", $insertAfter = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        const alert = document.createElement("div");
        var classList = "alert fade show";
        if ($class) classList = classList + " " + $class; //could be multiple ones.
        alert.classList = classList;
        alert.setAttribute("role", "alert");
        if ($id_pre != "" && $id != "") alert.setAttribute("id", $id_pre + "_" + $id);
        else if ($id != "") alert.setAttribute("id", $id);
        if ($msg != "") alert.innerHTML = $msg;
        if ($insertAfter != null) $insertAfter.after(alert);
        else el.append(alert);
        return alert;
    }


    /**
     * HTML is silly.
     * @param {*} $el - the select el or string of the id
     * @param {String} $attr - property to extract 
     * @returns 
     */
    getSelectedProp=function($el, $attr) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectedOption = el.options[el.selectedIndex];
        if (selectedOption==null) return null;
        return selectedOption.getAttribute($attr);
    }

    //var el_lvl = ANNE.A.setHTML("annex_lvl", aplex.price_lvl);
    ezAnniesToAnneDisplay = function ($el, $val = 0, $bdclass="balance-display", $ico="fermatta1rem", $valClass="annecoin-balance-value", $bRESET=true) {
        var el = this.__getEl($el);
        if (el == null) return null;

		if ($bRESET==true) el.innerHTML="";
		
        if ($val == 0) {
            el.classList.add("text-danger-lbl");
            el.innerHTML = "0";
            return;
        }
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var span = ANNE.A.ezSpan(el, "", $bdclass);
        var val = UTILS.formatNumberCommas(annecoin);
        //console.log(el.id + " val:" + val);
        ANNE.A.ezSpan(span, "", $valClass, val + "");
        ANNE.A.ezSpan(span, "", $ico);
    }

    /**
    * annies are never decimal, only whole
    */
    validateAndFormatAnnecoin = function ($el_input, $displayEl, $annecoinstring, $bRESET = true) {
        var str = "";
        var lastchar = $annecoinstring[$annecoinstring.length - 1];
        var multiplier = 0;
        if (lastchar == "k") {
            multiplier = 1000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "m") {
            multiplier = 1000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "b") {
            multiplier = 1000000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        for (var s = 0; s < $annecoinstring.length; s++) {
            if ("0123456789".indexOf($annecoinstring[s]) == -1) continue;
            str = str + $annecoinstring[s];
        }
        if (multiplier != 0) str = (parseInt(str) * multiplier) + "";
        $el_input.value = str;

        if (str.length > 0) {
            this.ezAnnecoinDisplay($displayEl, str, $bRESET);
        }
    }

    
    /**
     * 
     */
    ezAddNote = function($el, $noteText, $class="small-note") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.innerHTML = $noteText;
        span.classList = $class;
        el.append(span);
    }

    displayAnnecoinWithCommas = function ($el, $annecoin = 0, $bRESET=false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = ANNE.UTILS.commas($annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + annecoin + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    ezAnnecoinDisplay = function ($el, $annecoin = 0, $bRESET = false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = Math.round(parseInt($annecoin)*100)/100;
        var acval = UTILS.formatNumber(annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    /**
     * will round for better display
     */
    ezAnneDisplayApprox = function ($el, $val = 0, $bRESET = false, $addon=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bRESET == true) ANNE.A.empty(el);
        if ($val == 0) {
            var span = ANNE.A.ezSpan(el, "", "balance-display");
            ANNE.A.ezSpan(span, "", "annecoin-balance-value", 0 + "");
            ANNE.A.ezSpan(span, "", "annecoin-small-ico")
            return;
        }
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var acval = UTILS.formatNumber(annecoin, true);
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico")
        if ($addon!=null) ANNE.A.ezSpan(span, "", "", $addon);
    }
    


    /**
     * 
     * Get and display our annies balance.
     * 
     * Get and display the annies balance in both ANNE coin and annies.
     */
    ezDisplayAnnecoinWithNeuron = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj!=null) {
            if (balanceObj.annies != undefined) {
                var annies = parseInt(balanceObj.annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    ezDisplayPendingAnnies = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj) {
            if (balanceObj.avail_soon_annies != undefined) {
                var annies = parseInt(balanceObj.avail_soon_annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {*} $value 
     * @param {*} $lbl 
     * @param {*} $checked 
     * @param {*} $disabled 
     * @param {*} $truncateClass 
     * @param {*} $cb 
     * @returns 
     */
    ezCheckMark = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, 
                            $truncateClass = "", $cb=null, $chkclass="chk-parent") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($id=="" || $id==null) $id = ANNE.UTILS.randHex(3);
        const div = document.createElement("div");
        if ($chkclass!="") div.classList.add($chkclass);
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "checkbox");
        if ($id!="") chk.setAttribute("id", $id);
        if ($id!="") chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.classList="chk-label";
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);

        //simple callback
        if ($cb) {
            chk.addEventListener("change", $e=>$cb($e));
        }
        return chk;
    }


    /**
     */
    ezInputText = function ($el, $id = "", $classList = "", $placeholder = "", $inputEventCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_input = document.createElement("input");
        el_input.setAttribute("type", "text");
        if ($classList != "") this.__classListParse(el_input, $classList);
        if ($id != "") el_input.setAttribute("id", $id);
        if ($placeholder) el_input.setAttribute("placeholder", $placeholder);
        el.append(el_input);
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            el_input.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                $inputEventCallback(val);
            });
        }
        return el_input;
    }


    /** 
     * @param {element or string} - parent element to append to
     * @param {anne_neuron} - class to apply uom to.
     */
    ezInput = function ($el, $classNOBJ = null, $id = "", $class = "",
        $placeholder = "", $inputEventCallback = null, $inputFilterCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($class != "") {
            var div = document.createElement("div");
            div.classList = $class;
            el.append(div);
        }
        else div = el;

        const input = document.createElement("input");
        input.setAttribute("type", "text");
        if ($id != "") input.setAttribute("id", $id);
        if ($placeholder) input.setAttribute("placeholder", $placeholder);
        div.append(input);
        var uom = null;
        if ($classNOBJ != null) uom = MEM.unitOfMeasure($classNOBJ);
        if (uom != null) {
            var span = document.createElement("span");
            span.style.display = 'inline-block';
            span.classList = 'uom-symbol';
            div.appendChild(span);
        }
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            div.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                if ($inputFilterCallback != null && typeof $inputFilterCallback === "function") {
                    val = await $inputFilterCallback(val);
                    $e.target.value = val;
                }
                if (uom != null) {
                    var thisconv = null;
                    if (uom.conversions != undefined && uom.conversions["_" + val] != undefined) {
                        thisconv = uom.conversions["_" + val];
                    }
                    else thisconv = selector_number.convert(val, uom);
                    span.innerHTML = thisconv.total + "<span>&nbsp;" + thisconv.post + "</span>";
                }
                $inputEventCallback(val);
            });
        }
        return div;
    }


    //should do radios.
    ezRadio = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, $truncateClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const div = document.createElement("div");
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "radio");
        chk.setAttribute("id", $id);
        chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);
        return chk;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @param {*} $cb 
     */
    injectInputAfter=function($el, $wrapperclass, $inputclass, $cb, $bDOCANCEL=true){
        var el = this.__getEl($el);
        if (el == null) return null;
        var wrapperdiv=document.createElement("div");
        if ($wrapperclass!="") wrapperdiv.classList = $wrapperclass;
        var newel = document.createElement("input");
        if ($inputclass!="") newel.classList = $inputclass;
        newel.style['flex-grow']=1;
        newel.style.cursor='auto';

        wrapperdiv.append(newel);
        this.appendAfter(el, wrapperdiv);
        ANNE.A.makeReturnPressEvent(newel, $cb);
        ANNE.A.injectEnterAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        })
        if ($bDOCANCEL==true) ANNE.A.injectCancelAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        });
        return newel;
    }

    injectEnterAfter = function($el, $cb, $class="") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var btnEnter = document.createElement("button");
        btnEnter.innerHTML = "enter";
        if ($class!="") btnEnter.classList = $class;
        el.append(btnEnter);
        btnEnter.addEventListener("click", $cb);
    }

    injectCancelAfter = function($el, $cb, $class="btn-red-sq") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const btnCancel = document.createElement("button");  
        if ($class!="") btnCancel.classList=$class;         
        btnCancel.innerHTML = "[x]";
        el.appendChild(btnCancel);
        btnCancel.addEventListener("click", $cb);
    }


    /**
     * Make a return press event to callback the $cb.
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeReturnPressEvent = async($el, $cb)=>{
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($cb!=null) {
            el.addEventListener("keyup", async($e)=>{
                if ($e.keyCode === 13) {
                    $cb(el);
                }
            });    
        }
        return;
    }

    
    /**
     * Make a new element just like the $el and inject it after the $el as a sibbling.
     * @param {*} $el 
     * @returns 
     */
    makeSibAfter=function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var clone = $el.cloneNode(true);
        clone.innerHTML="";
        clone.id="newidclon_"+Math.floor(Math.random()*1000);
        this.appendAfter(el, clone);
        return clone;
    }


    /**
     * 
     * @param {anne_gui_role} $agr - the gui role object 
     */
    ezRoleEdit = function ($el, $clump_nid, $rel, $guid, $bEDIT = false, $bMOD = false, $el_inject = "", $newcb = null, $cancelcb = null, $deletecb = null, $span2 = null) {
        var span2 = $span2;
        if (span2 == null) span2 = this.ezRoleDisplay($el, $rel, $guid);
        else span2 = this.__getEl($span2);

        var el_inject = this.__getEl($el_inject);
        if (el_inject) {
            span2.setAttribute("el_inject", $el_inject.id);
        }

        span2.setAttribute("clump_nid", $clump_nid);
        span2.setAttribute("guid", $guid);

        if ($bEDIT == true) {
            var btnAdd = document.createElement("button");
            btnAdd.innerHTML = "+add";
            btnAdd.id = "add_" + $rel.rid + "_" + $guid;
            span2.append(btnAdd);
            btnAdd.addEventListener("click", e => ANNE.A.editRoleLaunchSelector(span2, $newcb, $cancelcb));
            span2.setAttribute("delete_ok", 1);
            btnAdd.style['margin-left'] = '4px';

            if ($rel.type == TYPE_ROLE) {
                var btndelete = document.createElement("button");
                btndelete.innerHTML = "[remove]";
                btndelete.id = "delete_" + $rel.rid + "_" + $guid;
                span2.append(btndelete);
                btndelete.setAttribute("txid", $rel.txid);
                btndelete.addEventListener("click", e => $deletecb(span2, btndelete));
                btndelete.style['margin-left'] = '4px';
            }

        }

        if ($bMOD == true) {
            var btnmod = document.createElement("button");
            btnmod.innerHTML = "+mod";
            btnmod.id = "mod_" + $rel.rid + "_" + $guid;
            btnmod.style['margin-left'] = '4px';
            btnAdd.addEventListener("click", e => function (e) { alert("todo"); });
            div.append(btnmod);
        }

        var ptxid = "";
        var to_s = "";
        if ($rel != null) {
            ptxid = $rel.txid;
            to_s = $rel.to_s;
        }

        var default_class_nid = null;
        var clumpNOBJ = MEMG.getNeuronByNid($clump_nid);

        if (ptxid != "") default_class_nid = clumpNOBJ.getDefaultClassForParam(ptxid);
        if (default_class_nid != null) span2.setAttribute("classRestrictNid", default_class_nid);
        else if (to_s != "") span2.setAttribute("classRestrictNid", to_s);
        else if ($default_class_nid != null) span2.setAttribute("classRestrictNid", $default_class_nid);
    }


    /**
     * An editable role has been clicked on, launch the selector
     * @param {*} $span2 
     */
    editRoleLaunchSelector = async ($span2, $cb_new = null, $cb_cancel = null, $cb_defocus = null) => {
        var tnid = $span2.getAttribute("role_tnid");
        var el_inject = $span2.getAttribute("el_inject");
        var roleNOBJ = MEMG.getNeuronByTempNid(tnid);
        var misc = {};
        misc.smartDefocus = $cb_defocus;
        misc.bCALLBACKONLY = true;
        var pObj = new page_params($cb_new, $cb_cancel, null, roleNOBJ.nid);
        pObj.misc = misc;
        var contextNOBJ = null;
        var classRestrictNid = $span2.getAttribute("classRestrictNid");
        if (classRestrictNid != undefined) {
            pObj.misc.classRestrictNid = classRestrictNid;
            contextNOBJ = MEMG.getNeuronByNid(classRestrictNid);
        }
        var page = ANNE.MEM_UDF.getDefaultSelectorPageForClass(roleNOBJ, null, contextNOBJ);
        console.log("%cPAGE "+page, TALK.bggreen);
        var clump_nid = $span2.getAttribute("clump_nid");
        pObj.misc.from_nid = clump_nid;   //all TYPE_ROLE relons will be put on the clump.
        pObj.misc.type = TYPE_ROLE;
        pObj.misc.reln_nid = roleNOBJ.nid;
        ANNE.A.empty(el_inject);
        pObj.injection_div = el_inject;
        var guid = $span2.getAttribute("guid");
        if (guid == null) return;
        
        var pgui = ANNE.MEM_UDF.getPguiByGuid(guid);
        if (pgui.INST && typeof pgui.INST.smartDefocus === "function") {
            pgui.INST.smartDefocus(true);
        }
        if (pgui == null) return;
        await pgui.loadPart(page, pObj);
    }


    /**
     * @param {*} $rel 
     */
    ezRoleDisplay = function ($el, $rel, $guid) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var roleNOBJ = MEMG.getNeuronByTempNid($rel.r);
        var _roletname = roleNOBJ.text.replaceAll("_", "-");

        //frienly replacement of role name from affected_thing to who
        //when dealing with persons in the that slot.
        if ($rel.to_type == TO_TYPE_ANNE) {
            if (_roletname == "role-affected-thing") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-affected-who";
            }
            else if (_roletname == "role-actor") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-actor-who";
            }
        }

        var div = document.createElement("div");
        div.classList = "role2 " + _roletname + "-outter";
        el.append(div);
        var span1 = document.createElement("span");
        span1.classList = _roletname + "-" + "lbl";
        div.append(span1);
        var span2 = document.createElement("span");
        span2.classList = "r-inner";
        span2.id = "val_" + roleNOBJ.temp_nid + "_" + $guid;
        span2.setAttribute("role_tnid", roleNOBJ.temp_nid);
        div.append(span2);

        //display only if a role relon
        if ($rel.type == TYPE_ROLE) {
            if ($rel.to_type == TO_TYPE_ANNE) {
                ANNE.MEM_UDF.makeNeuronLink(div, $rel.to_s);
            }
            else if ($rel.to_type == TO_TYPE_TXID) {
                ANNE.MEM_UDF.makeRelonViewLink(div, $rel.to_s);
            }
            else {
                var totxt = MEMG.__MASTER_DISPLAY_RELON_TO($rel);
                ANNE.A.setHTML(div, totxt);
            }
        }

        return span2;
    }


    ezSQDWrapper = function ($el, $class, $display = "", $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var termdelete = document.createElement("div");
        el.append(termdelete);
        if ($class != "") termdelete.classList = $class;
        if ($display != "") termdelete.style.display = $display;
        ANNE.A.ezSquareDelete(termdelete, $callback, $callbackParams, $size, $title);
        return termdelete;
    }

    ezSquareDelete = function ($el, $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var dsquare = document.createElement("div");
        dsquare.classList = "delete-square";
        el.append(dsquare);
        var boxX = ANNE.A.ezEl(dsquare, "span", "boxX", "<img src='../imgs/boxX.png' height='" + $size + "' width='" + $size + "'>", "", $title);
        var sl1c = this.sl1c;
        if (typeof $callback !== "function") {
            console.log("callback jacked up in ezSquareDelete");
            return;
        }
        boxX.addEventListener("click", e => $callback(...$callbackParams));
    }
    
    /**
     * 
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearestObjByProp=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return $el;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }


    /**
     * Extract the prop of a nearest-ancestor
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearest=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return attr_val;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }



    /**
     * @param {Element} $el
     */
    ez_getNearestParentWithProp = function ($el, $prop="nid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var nid = $el.getAttribute($prop);
            if (nid != null) return nid;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }

    /**
     */
    ez_floatActionIco = async ($nid, $ainame, $titleOverride = "")=> {
        var title = ""
        var actionz = "";

        if ($ainame == "mine") {
            title = "This is your data.  Click on it to delete it.";
            actionz = "delete";
        }

        else if ($ainame == "collab") {
            title = "This is data you can collab on.  Click on it to delete it.";
            actionz = "delete";
        }

        //this is locally ghosted data..
        else if ($ainame == "ghost") {

            if (ANNE._txMode == TXMODE_GHOST) {
                title = "This is your ghost.  Only you see it.";
            }

            //if in ANNE mode...
            else if (ANNE._txMode == TXMODE_ANNE) {

                //if it is upgradable...
                var nObj = MEMG.getNeuronByTempNid($nid);
                if (nObj._govenor == null) await nObj.setGovRights(USER._userNObj, false);
                if (nObj.collab_ACTION >= govenor.COLLAB_REQ_OK) {
                    title = "Save your ghost to ANNE.";
                    $ainame = "ghost-to-anne";
                    actionz = "upgrade-ghost-to-anne";
                }
                else {
                    title = "This class of local ghost data cannot be upgraded to ANNE Ghost data.";
                }
            }
        }

        var newspan = document.createElement("span");
        newspan.classList = "hoverable";
        newspan.setAttribute("title", title);
        var itag = document.createElement("i");
        itag.classList = "ai-" + $ainame + " actions-ico";

        //this will trigger DOM events on class clicks to optimize (minimize) events.
        //the widget calling this helper will have to be listening.
        if (actionz) itag.setAttribute("actionz", actionz);

        newspan.append(itag);
        return newspan;
    }

    /**
     */
    ez_auto_inject_close_btn = function ($el, $class, $hrefid, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var betterEl = el.querySelector(".inject-close-here");

        //always prepend into the 
        if (betterEl) el = betterEl;

        var bPREPEND = true;    //if we didn't find a better el, PREPEND (to injection div)
        return ANNE.A.ezHrefBtn(el, $class, $hrefid, $spanClass, $icoclass, bPREPEND);
    }


    /**
     * EXT - NOT IN USE IN EXT...
     * 
     * This will render a gui ico for deleting or updgrading ur own records.
     */
    ez_floatActionIcos = async ($nObj, $el, $ownStatus = "collab")=> {
        var el = this.__getEl($el);
        if (el == null) return null;
        var bSHOW = false;

        if ($nObj.mine == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "mine");
            floatico.style = "cursor:default";
            el.append(floatico);
            bSHOW = true;
        }

        //only one of these...
        else if ($nObj.ghost == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "ghost");
            el.append(floatico);
            bSHOW = true;
        }
        else if ($ownStatus == "collab") {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "collab");
            el.append(floatico);
            bSHOW = true;
        }

        if (bSHOW == true) ANNE.A.show(el);
    }


    /**
     * text dd
     */
    ezDD_basic = function ($el, $items, $prompt = "select", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        ANNE.A.empty(selectel);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(selectel.value));
        }
        var opt = document.createElement("option");
        opt.setAttribute("value", "");
        opt.innerHTML = $prompt;
        selectel.append(opt);
        for (var i = 0; i < $items.length; i++) {
            var item = $items[i];
            var opt = document.createElement("option");
            opt.setAttribute("value", item);
            opt.innerHTML = item;
            selectel.append(opt);
        }
        ANNE.A.show(selectel);
        return selectel;
    }


    /** 
     */
    ezDD = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false, $skiptnids = []) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            if ($skiptnids.length && $skiptnids.indexOf(nObj.temp_nid) != -1) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);

            opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }


    /**
     * used in pref dds 
     */
    ezDD_masterDDObj = function ($el, $masterOBJS, $callback = null, $msg = "easy access acct nids", $ddClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else selectel = this.ezEl(selectec, "select");

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);

        for (var n = 0; n < $masterOBJS.length; n++) {
            var masterOBJ = $masterOBJS[n];
            var nObj = masterOBJ.nObj;
            var type = masterOBJ.type;
            var nick = masterOBJ.nick;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            var text = "";
            if (nick == nObj.nid) opt.innerHTML = "(" + type + ") " + nObj.nid;
            else opt.innerHTML = "(" + type + ") " + nick + " - " + nObj.nid;
            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDD_nObjs = function ($el, $nObjs, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false) {
        var el = this.__getEl($el);
        if (el == null) {
            console.log("(ezDD_nObjs) el "+$el+" doesnt exist");
            return null;
        }
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => {
				$callback(e);
			});
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        //var aNObjs=MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) $nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            if (nObj == null) continue;
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            if (nObj.local_nickname != undefined) opt.innerHTML = nObj.local_nickname;
            else opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    doGoogleMap = function ($el, $lat = -34.397, $long = 150.644) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var map = new google.maps.Map(el, {
            center: { lat: $lat, lng: $long },
            zoom: 8,
        });
        return map;
    }

    /**
     */
    ezDDWithGens = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];
                //doHeading(nObj);
                var indent = "&#9494;";
                var space = '&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
                //items.push({id: nObj.temp_nid,text: '<div>'+space+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
            }
            //results.push({text: $heading,children: items});
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDDWithGens_ACTIONS = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $V1ONLY = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;

            //we only want V1 vbs.
            if ($V1ONLY == true) {
                var rel_vbForm = MEMG.getBeRelon(nObj.outs, MEM.vb_form.temp_nid);
                if (rel_vbForm != null) {
                    if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                        console.log("skipping non v1");
                        continue;
                    }
                }
            }

            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];

                //we only want v1 vbs.
                if ($V1ONLY == true) {
                    var rel_vbForm = MEMG.getBeRelon(childNOBJ.outs, MEM.vb_form.temp_nid);
                    if (rel_vbForm != null) {
                        if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                            console.log("skipping non v1");
                            continue;
                        }
                    }
                }

                var indent = "&#9494;";
                var space = '&#160;&#160;&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;&#160;&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
            }
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $val 
     * @returns 
     */
    setSelectedIndexByValue=function($el, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var i=0; i<el.options.length; i++) {
            if (el.options[i].value==$val) {
                selectel.setAttribute('selectedIndex', i);
                break;
            }
        }
    }


    /**
     * ez row with 2 col
     */
    ez2Col = function ($el, $v1 = "", $v2 = "", $v1extraclass = "", $v2extraclass = "") {
        var v1class = "simple-col";
        if ($v1extraclass != "") v1class = v1class + " " + $v1extraclass;
        var v2class = "simple-col";
        if ($v2extraclass != "") v2class = v2class + " " + $v2extraclass;
        var row = ANNE.A.ezEl($el, "div", "simple-row");
        ANNE.A.ezEl(row, "div", v1class, $v1);
        ANNE.A.ezEl(row, "div", v2class, $v2);
        return row;
    }


    //intended for new divs, spans, and other elements that
    //you want to also givev a class, id and style.
    ezEl = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.append(newEl);
        return newEl;
    }
    ezElPrepend = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.prepend(newEl);
        return newEl;
    }


    doImgB64 = function($el, $nObj, $icostyle="dossier", $w=84, $h=0) {     
        var el = this.__getEl($el);
        if (el == null) return null;
        var url = ANNE.A.getThumbImgUrl($nObj, $icostyle);
        if (url.length) {
            ANNE.A.doImageB64(el, "", $afile, $w, $h);
        }
    }



    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    getThumbImgUrl=function($nObj) {
        var parentNOBJ = $nObj.getParent();
        var neuronImgNOBJ = MEM.__GET_DEFAULT_IMG_NEURON(parentNOBJ, $nObj);
        if (neuronImgNOBJ!=null) {
            var url = IMAGES.getUrl(neuronImgNOBJ);
            if (url!="") return url;
        }

        //we're going to nixx all this drama.
        return ANNE.A.getThumbImgUrlFromRelons($nObj);
    }


    /**
     * 
     * @returns 
     */
    getThumbImgUrlFromRelons = function($nObj, $icostyle=null) {
        var test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_thumbnail.temp_nid);
        if (test==null) {
            test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Skyline.temp_nid);
            if (test==null) {
                test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Map.temp_nid);
                if (test==null) {
                    test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Flag.temp_nid); 
                    if (test==null) test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Seal.temp_nid); 
                }
            }
        }
        if (test!=null) { 
            return "http://commons.wikimedia.org/wiki/Special:FilePath/"+test.to_s+"?width=200";
        }

        //console.log("not enough info on "+$nObj.text+" for img or ico");
        return "";
    }


    focusTextFix = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.addEventListener("focus", function () {
            el.setSelectionRange(0, 0);
        });
    }

    
    __classListParse = function ($el, $cl) {
        if ($cl == "") return;

        if ($cl != "") {
            if ($cl.indexOf(" ") != -1) {
                var cl = $cl.split(" ");
                for (var c = 0; c < cl.length; c++) {
                    var clstr = cl[c].trim();
                    if (clstr == "") return;
                    $el.classList.add(clstr);
                }
            }
            else $el.classList = $cl;
        }
    }

    //we can inject a close button into a model with this..
    //and then wire it to its pgui.close function, automatically. :)
    ezHrefBtn = function ($el, $hrefClass, $hrefid=null, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var href = document.createElement("a");
        if ($hrefid!=null) href.setAttribute("id", $hrefid);
        if ($hrefClass) href.classList = $hrefClass;
        if ($bPREPEND == false) $el.append(href);
        else $el.prepend(href);
        ANNE.A.ezIcoSpan(href, $spanClass, $icoclass);
        return href;
    }

    /**
     * ez ico img using the css ico classes
     */
    ezIcoImgSpan = function ($el, $img_class, $msg, $bEMPTY = true, $bSHOW = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bEMPTY == true) ANNE.A.empty(el);
        var span = document.createElement("span");
        span.classList = $img_class;
        el.append(span);
        if ($msg != "") {
            var msg_span = document.createElement("span");
            msg_span.innerHTML = $msg;
            el.append(msg_span);
        }
        if ($bSHOW == true) ANNE.A.show(el);
    }


    ezIcoLink = function ($el, $href = "#", $ico_class = "ai-anne") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        el.append(link);
        var link_i = document.createElement("i");
        link_i.classList = $ico_class;
        link.append(link_i);
    }


    ezHref = function ($el, $href = "#", $target = "_blank", $txt = "", $ico_class = "") {
        var el = this.__getEl($el);
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        link.setAttribute("target", $target);
        if ($txt == "") link.innerHTML = $href;
        else link.innerHTML = $txt;
        el.append(link);
        if ($ico_class != "") {
            var link_i = document.createElement("i");
            link_i.classList = $ico_class;
            link.append(link_i);
        }
    }


    ezIcoSpan = function ($el, $spanClass, $icoClass, $spanMsg = "", $stylejson = null) {
        var el = this.__getEl($el);
        //if (el==null) return null;
        var newel = document.createElement("span");
        if ($spanClass) newel.classList = $spanClass;
        if (el != null) el.append(newel); //append if we have it, else we just return the new one
        if ($stylejson != null) el.style = $stylejson;
        var ico = document.createElement("i");
        ico.classList = $icoClass;
        newel.append(ico);
        if ($spanMsg != "") newel.innerHTML = newel.innerHTML + $spanMsg;
        return newel;
    }

    ezlblSpan = function ($el, $lbl, $value = "", $spanclass = "", $valclass = "", $sep = "&nbsp;") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newel = document.createElement("span");
        if ($spanclass != "") newel.classList = $spanclass;
        if ($valclass == "") newel.innerHTML = $lbl + $sep + $value;
        else newel.innerHTML = $lbl + $sep + '<span class="' + $valclass + '">' + $value + '</span>';
        el.append(newel);
        return newel;
    }

    ezTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $class = "simple-row", $exclude = [TYPE_PARAM], $sep = ", ") {
        var list = TALK.printListByReln($outs, $relnNOBJ.temp_nid, $exclude, $sep);
        if (list.length > 0) {
            var eltextlist = document.createElement("div");
            if ($class != "") eltextlist.classList = $class;
            if ($lbl != "") eltextlist.innerHTML = $lbl + list;   //lbl should have trailing space if it is needed.
            $el.append(eltextlist);
        }
    }

    
    ezTwoColTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $exclude = [TYPE_PARAM], $lblClass = "span-lbl", $dataClass = "span-data") {
        var relons = MEMG.getRelonsByTempR($outs, $relnNOBJ.temp_nid, $exclude);
        if (relons.length == 0) return "";

        var elrow = this.ezEl($el, "div", "simple-row");
        var lblcol = this.ezEl(elrow, "div", "simple-col tal");
        this.ezSpan(lblcol, "", $lblClass, $lbl);

        var datacol = this.ezEl(elrow, "div", "simple-col tal");

        for (var r = 0; r < relons.length; r++) {
            var rel = relons[r];
            var totxt = MEMG.__MASTER_DISPLAY_RELON_TO(rel);
            var el_link = EXT_BTNS.makeRelonViewLink(datacol, rel.txid, false, 8, true, "", totxt);
            el_link.style.display = "block";
        }
    }

    ezSortCol = function ($el, $colObj, $tbldivid, $sort) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var classList = ["sortico", "hoverable-green"];

        if ($colObj.ttype == "number") classList.push("ai-sort-amount-down");
        else if ($colObj.ttype == "string" || $colObj.ttype == "anne") classList.push("ai-sort-alpha-down");
        else if ($colObj.ttype == "date") classList.push("ai-sort-down");

        var span = document.createElement("span");
        span.innerHTML = $colObj.text + "&nbsp;";
        el.append(span);
        var ico = document.createElement("i");
        ico.classList = classList.join(" ");
        ico.setAttribute("prop", $colObj.ttype);
        ico.setAttribute("sort", $sort);
        ico.setAttribute("tbl", $tbldivid);
        span.append(ico);

        return ico; //return the sortico (so we can easily put events on it)
    }


    ezSpan = function ($el, $id, $class, $innermsg, $styleJson = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, $styleJson);
        el.append(span);
        return span;
    }

    //make it even if empty.
    ezSpan2 = function ($el, $id, $class, $innermsg, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, null);
        el.append(span);
        if ($callback) span.addEventListener("click", $e=>$callback($e));
        return span;
    }

    /**
     * does not append, just creates
     */
    ezSpanSolo = function ($id = "", $class = "", $innermsg = "", $styleJson = null) {
        var newel = document.createElement("span");
        if ($class) newel.classList = $class;
        if ($id != "") newel.setAttribute("id", $id);
        newel.innerHTML = $innermsg;
        if ($styleJson != null) newel.style = $styleJson;   //not working as i thought
        return newel;
    }

    ezButton = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        btn.innerHTML = $text;
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        return btn;
    }

    ezButtonWithIcoClass = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null, $icoClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        var icoEl = document.createElement("span");
        icoEl.classList = $icoClass;
        btn.appendChild(icoEl);
        icoEl.style['pointer-events']= 'none';
        var textEl = document.createElement("span");
        textEl.innerHTML = $text;
        textEl.style['pointer-events']= 'none';
        btn.appendChild(textEl);
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        
        return btn;
    }

    ezFbwBtn = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.appendChild(btn);
        return btn;
    }

    ezFbwBtn_prepend = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.prepend(btn);
        return btn;
    }

    /**
     */
    getPrintNameIcoSpan = function ($nObj) {
        if ($nObj.temp == true) return "<span class='temp-neuron'>" + $nObj.bestName() + "</span>";
        else if ($nObj.ghost == true) return "<span class='ghost-neuron'><span class='ghost-small-blue-ico'></span>" + $nObj.bestName() + "</span>";
        else if (ANNE.isANNE($nObj.owned_by)==true) return "<span class='real-neuron'><span class='anne-small-black-ico'></span>" + $nObj.bestName() + "</span>";
        else if ($nObj.owned_by == USER._userNObj.nid) return "<span class='real-neuron'><span class='mine-small-ico'></span>" + $nObj.bestName() + "</span>";
        return $nObj.text;
    }


    imgBadAudit = function ($el, $id = "", $imgref = "", $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var spanEl = this.doBadImg(el, $id, $imgref, $imgclass)
        if ($callback != null && typeof $callback === "function") {
            spanEl.addEventListener("click", function (e) { $callback(e); });
        }
        return spanEl;
    }


    imgButton = function ($el, $id = "", $imgref, $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var imgEl = this.doImageDyno(el, $id, $imgref, $imgclass)
        if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {anne_file} $afile 
     * @param {*} $imgclass 
     * @param {function} $callback 
     * @returns {element}
     */
    imgButtonB64 = function ($el, $id = "", $afile=null, $imgclass = "search-img-small", $callback = null) {
        if ($afile==null || $afile==undefined) {
			return null;
		}
        var el = this.__getEl($el);
        if (el == null) {
			return null;
		}
        var imgEl = this.doImageDynoB64(el, $id, $afile, $imgclass)
        if (imgEl==undefined || imgEl==null) {
			return null;
		}
		if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }

    
    /**
     * Insert an element into the DOM after another element.
     * lolz, as this is the same as the native .after, so we should just use that.
     * @param {Element or String} $existingNode 
     * @param {Element} $newNode 
     * @returns 
     */
    insertAfter = function($existingNode, $newNode) {
        var el = this.__getEl($existingNode);
        if (el == null) return;
        //if (el.parentNode==null) return;
        //el.parentNode.insertBefore($newNode, el.nextSibling);
        $existingNode.after($newNode);
        return;
    }

    /**
     * Move an el to be after another el.
     */
    moveElAfter=function($el1, $insertAfterThisEl) {
        var el1 = this.__getEl($el1);
        if (el1 == null) return;
        var insertAfterThisEl = this.__getEl($insertAfterThisEl);
        if (insertAfterThisEl == null) return;
        insertAfterThisEl.after($el1);
    }

    appendAfter = function($existingNode, $newNode) {
        this.insertAfter($existingNode, $newNode);
    }

    //triple toggle,  asc, desc, none
    sortToggle = function ($el, $colNumb, $sort = "desc", 
                            $callback = null, 
                            $sortype="numeric") {

                            
                            //$descending = "bi bi-sort-alpha-down", 
                            //$ascending = "ai-sort-asc") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("span");
        btn.style.cursor = "pointer";
        btn.classList = "ez-sort-btn";
        btn.setAttribute("col_num", $colNumb);
        if ($sort == "desc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
            btn.classList.add("active");
        }
        else if ($sort == "asc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
            }
            btn.classList.add("active");
        }

        //none is parked on open desc (not selected);
        else {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
        }

        btn.setAttribute("sort_val", $sort);

        el.appendChild(btn);

        if ($callback != null) {
            //CLICK ORDER CYYCLE::
            //NONE->DESC->ASC
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                target = ANNE.A.getFirstParentOf(target, "ez-sort-btn");
                var colNumb = target.getAttribute("col_num");
                var sort_val = target.getAttribute("sort_val");
                if (sort_val == "desc") {
                    target.setAttribute("sort_val", "asc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
                    }
                    $callback(colNumb, "asc");
                }

                //asc becomes NONE.. not active.
                //show non-active desc sort.  if clicked, it becomes active to desc sort.
                else if (sort_val == "asc") {
                    target.classList.remove("active");
                    target.setAttribute("sort_val", "none");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "none");
                }

                else if (sort_val == "none") {
                    target.setAttribute("sort_val", "desc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "desc");
                }


            });

            //init sort.
            //if ($sort=="desc" || $sort=="asc") $callback( $colNumb, $sort);
        }
    }

    /**
     * simple toggle of a bool attribute with get/set
     */
    toggleElAttributeBool=function($el, $prop) {
        if ($el==null) return;
        if ($el.getAttribute($prop)==true) $el.setAttribute($prop, false);
        else $el.setAttribute($prop, true);
    }

    toggleBtn = function ($el, $nObj, $class = "", $callback = null, $on = true, $lbl = "", $ton = "toggled-on", $toff = "toggled-off") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = "toggle-btn " + $class;
        if ($on == true) btn.classList.add($ton);
        else btn.classList.add($toff);
        btn.setAttribute("tnid", $nObj.temp_nid);
        if ($lbl.length) btn.innerHTML = $lbl;
        else btn.innerHTML = nObj.bestNameHTML();
        el.appendChild(btn);

        btn.addEventListener("click", async (e)=> {
            var target = e.target;

            //if this is off we are turning it on.
            //if it is on.. it stays on even when clicked again.
            if (target.classList.contains($toff)) {
                target.classList.remove($toff);
                target.classList.add($ton);
            }
            else {
                target.classList.remove($ton);
                target.classList.add($toff);

                //callback.
                if ($callback != null && typeof $callback === "function") $callback(e);
            }
        });

        return btn;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $masterObj 
     * @param {*} $eltype 
     */
    makeToggleSet = function($el, $masterObj, $eltype="div") {
        var ats = new anne_toggle_set($masterObj);
        this.makeToggleSetGUI($el, ats, $eltype);
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @returns 
     */
    geAllSibs = function($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent_el = el.parentNode;
        if (parent_el==null) return null;
        return parent_el.querySelectorAll("."+$class);
    }


    /**
     * Creates divs from array of anne_toggle_objs.
     * Will allow for toggle of each div via a toggle btn.
     * @param {*} $el 
     * @param {anne_toggle_set} ats 
     * @param {*} $eltype 
     * @returns {anne_toggle_set}
     */
    makeToggleSetGUI = function($el, $ats, $eltype="div") {
        var el = this.__getEl($el);
        if (el == null) return;

        var el_btn_row = ANNE.A.ezEl($el, "div", "tg-btns", "", "");
        var el_toggle_divs = ANNE.A.ezEl($el, "div", "tg-toggle-divs", "", "");

        for (var i=0; i < $ats.atos.length; i++) {
            const ato = $ats.atos[i];
            console.log("btn tnid "+ato.btn_tnid);
            
            ato.el_toggle_divs = el_toggle_divs;

            //put new button in btn row
            var btn = ANNE.A.ezButton(el_btn_row, "", ato.btn_text, ato.btn_class);
            ato.btn_el = btn;
            if (ato.btn_tnid!="") btn.setAttribute("tnid", ato.btn_tnid);

            //put new div in the .tg-set div.
            var newEl = ANNE.A.ezEl(el_toggle_divs, $eltype, ato.div_class+" tg-set", ato.text, ato.id, ato.title, ato.style);
            ato.div_el = newEl;
            btn.setAttribute("btn_tnid", ato.btn_tnid);

            
            btn.addEventListener("click", async($e)=>{
                if (ato.div_el.classList.contains("active")==true) {
                    ato.div_el.classList.remove("active");
                    if (ato.cb!=null) ato.cb(ato, false);
                }
                else {
                    //should get all tg-set divs
                    //var els = ANNE.A.geAllSibs(ato.div_el, "tg-set");
                    for (var e = 0; e < el_toggle_divs.children.length; e++) {
                        el_toggle_divs.children[e].remove("active");
                    }
                    ato.div_el.classList.add("active");
            
                    if (ato.cb!=null) ato.cb(ato, true)
                };
            });
        }
        return $ats;
    }


    /**
     * simple toggle set of buttons
     */
    toggleBtnSet = function ($el, $nObjs, $classes, $callback, $defaultNid = null, $lbloverride = []) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($nObjs.length == 0) return;
        var elset = document.createElement("div");
        elset.classList = "toggle-btn-set";
        el.appendChild(elset);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var btnclass = $classes[n];
            var btn = document.createElement("button");
            btn.classList = "toggle-btn " + btnclass;
            if ($defaultNid && nObj.nid == $defaultNid) btn.classList.add("toggled-on");
            else btn.classList.add("toggled-off");
            btn.setAttribute("nid", nObj.nid);
            if ($lbloverride.length) btn.innerHTML = $lbloverride[n];
            else btn.innerHTML = nObj.text;
            elset.appendChild(btn);
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                //if this is off we are turning it on.
                //if it is on.. it stays on even when clicked again.
                if (target.classList.contains("toggled-off")) {
                    //turn off the curren toggled btn.
                    var btnset = ANNE.A.getFirstParentOf(target.parentNode, "toggle-btn-set", true);
                    if (btnset != null) {
                        var onbtn = btnset.querySelector(".toggled-on");
                        if (onbtn != null) {
                            onbtn.classList.remove("toggled-on");
                            onbtn.classList.add("toggled-off");
                        }
                    }
                    //turn on this one.
                    target.classList.remove("toggled-off");
                    target.classList.add("toggled-on");

                    //callback.
                    if ($callback != null && typeof $callback === "function") $callback(e);
                }
            });
        }
        return elset;
    }

    wireShowHideToggler=function($el_to_click, $el_showhide) {
        var el = this.__getEl($el_to_click);
        if (el == null) return;

		var elshowhide = null;
		if (el.classList.contains("hide")==true) {
			elshowhide = ANNE.A.setHTML($el_showhide, "show"); 
		}
		else {
			elshowhide = ANNE.A.setHTML($el_showhide, "hide"); 
		}

		if (elshowhide && elshowhide.getAttribute("showlink_id")==undefined) {
			elshowhide.setAttribute("showlink_id", el.id);
		}
		
        el.addEventListener("click", (e)=>{
            var newval = ANNE.A.toggle($el_showhide);
            //self fixing if things get whacky
            if (newval=="show") {
                if (el.innerHTML === 'show') el.innerHTML = 'hide';
            }
            else if(newval=="hide") {
                el.innerHTML = 'show';
            }
        });
    }

    toggleShowHideLink=function($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var el = document.getElementById('toggleLink');
        if (el.innerHTML === 'hide') el.innerHTML = 'show';
        else el.innerHTML = 'hide';
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $idToToggle 
     * @param {*} $addText 
     * @param {*} $defaultOpen 
     * @param {*} $ss 
     * @returns 
     */
    toggleMinMaxBtn = function ($el, $idToToggle, $addText = "", $defaultOpen = true, $ss = null, $addClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("div");
        btn.setAttribute("toggle_divid", $idToToggle);
        btn.classList = "ezMinMax hover";
        if ($addClass!="") btn.classList.add($addClass)
        if ($defaultOpen == false) {
            btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            ANNE.A.hide($idToToggle);
        }
        else {
            btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
            ANNE.A.fadeIn($idToToggle);
        }
        btn.style.cursor = "pointer";
        btn.style.display = "inline-block";
        el.append(btn);
        const ss = $ss;
        btn.addEventListener("click", async (e)=> {
            var target = e.target;
            target = ANNE.A.getFirstParentOf(target, "ezMinMax");
            var idToToggle = target.getAttribute("toggle_divid");
            if (idToToggle == undefined || idToToggle == null) return;
            var classz = ANNE.A.toggleWithFade(idToToggle);
            if (classz == "hide") {
                btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            }
            else if (classz == "show") {
                btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
                var ssEl = this.__getEl(ss);

                //var percentagePosition = (el.offsetTop / ssEl.scrollHeight) * 100;
                //ssEl.scrollTop = (ssEl.scrollHeight * percentagePosition) / 100;
                //console.log("show section");
                //var scrollHeight = ssEl.scrollHeight;
                //ssEl.scrollTop = scrollHeight;
            }
        });

        return btn;
    }


    __getEl = function ($el) {
        if (typeof $el === "string") {
            if ($el[0] == ".") return document.querySelector($el);

            //legacy references to jquery will have # in front of them.
            else if ($el[0] == "#") return document.getElementById($el.substring(1));

            return document.getElementById($el);
        }
        return $el;
    }

    delayThen = async ($dataparam, $delayms = 300, $callback = null) => {
        await UTILS.sleep($delayms);
        if ($callback != null) $callback($dataparam);
    }

    //we may add a delay to this..
    fadeIn = function ($el, $ms = 500, $ease = "power1.in", $displayProp = "") {
        const el = this.__getEl($el);
        if (el == null) return null;
		el.style.opacity=0;
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return;
        }
        gsap.to(el, {
            opacity:1,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = $displayProp;
                el.classList.add("show");
                el.classList.remove("hide");
            }
        });

        return el;
    }

    fadeInPop = function ($el, $addglowclass = "", $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = $displayProp;
        el.style.transition = "all 2000ms cubic-bezier(.39,.54,0,1.6);";
        if ($addglowclass != "") el.classList.add($addglowclass);
        return el;
    }


    fadeInDelayFadeOut = async ($el, $delayms, $inms = 500, $intt = "ease-in", $outms = 500, $outtt = "ease-out") => {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = "";
        el.style.transition = "all " + $innms + "ms " + $intt + ";";
        await UTILS.sleep($inms + $delayms);
        el.classList.add("hide");
        el.classList.remove("show");
        el.style.opacity = 0;
        el.style.display = "none";
        el.style.transition = "all " + $outms + "ms " + $outtt + ";";
        return el;
    }


    /**
     * This is blocking if it is awaited on!
     * Don't await it!
     */
    fadeOut = async ($el, $ms = 500, $delay = 0, $ease = "power1.out", $onComplete=null) => {
        const el = this.__getEl($el);
        if (el == null) return null;
        
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return el;
        }

        //max 2 sec delay to avoid accidents
        if ($delay > 0) await UTILS.sleep();

        gsap.to(el, {
            delay: Math.min(2, $delay/1000),
            opacity:0,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = "none";
                el.classList.add("hide");
                el.classList.remove("show");
                if ($onComplete!=null) $onComplete();
            }
        });

        return el;
    }


    /**
     * Tired of doing this.. so I will do it here.
     */
    formatSpanStrings = function ($el, $classes, $vars, $str, $bSHOW = true) {
        var el = this.__getEl($el);
        if ($classes.length == 0) return $str;
        if ($classes.length != $vars.length) {
            console.log("(formatSpanStrings) classes array and vars array are different lengths.  This may not do what you think it should do.");
        }
        for (var c = 0; c < $classes.length; c++) {
            var clz = $classes[c];
            var i = $str.indexOf("%");
            if (i >= 0) $str = $str.replace("%", "<span class='" + clz + "'>" + $vars[c] + "</span>");
        }
        el.innerHTML = $str;
        if ($bSHOW == true) ANNE.A.show(el);
    }

    getBottom = function ($dfc = ".smart-scroll-framer") {
        return parseInt(document.querySelector($dfc).getBoundingClientRect().bottom);
    }

    getElIndex = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return -1;
        return Array.prototype.indexOf.call(el.parentElement.children, el);
    }

    //because i forget names.
    getNearestParentOf = function($el, $stopClass, $bINCLUDESELF = true) {
        return this.getFirstParentOf($el, $stopClass, $bINCLUDESELF);
    }

    getFirstParentOf = function ($el, $stopClass, $bINCLUDESELF = true) {
        if ($stopClass == "") return null;
        var el = this.__getEl($el);
        if (el == null) return null;
        var safety = 0;
        if ($bINCLUDESELF == true) {
            if (el.classList.contains($stopClass) == true) return el;
        }
        while (el.parentNode != null) {
            if (safety > 20) return null;
            el = el.parentNode;
            if (el == null || el == undefined) return null;
            if (el.nodeName == "#document") return null;
            if (el.classList.contains($stopClass) == true) return el;
            safety++;
        }
        return null;
    }

    getFirstParentAttr = function ($el, $stopClass, $getAttribute) {
        if ($getAttribute == "") return null;
        var el = this.getFirstParentOf($el, $stopClass);
        if (el == null) return null;
        return el.getAttribute($getAttribute);
    }


    getHTML = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.innerHTML;
    }

    getValue = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.value;
    }

    /**
     * input (w/ inputid) marked with class="rqd" wired to element with mod-el=inputid of rqd class el
     * displays helper msg
     * DOM wires rqd element with a .modxx ptr to "mod-el" span element
     * uses css::before to display color/status in the mod-el span.
     */
    applyRQD = function ($el, $bVALID = true, $msg = "required") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("rqd") == true && el.modxx != undefined) {
            if ($bVALID) {
                el.modxx.style['color'] = "#00db00";
                el.modxx.innerHTML = 'ok';
            }
            else {
                el.modxx.style['color'] = "red";
                el.modxx.innerHTML = $msg;
            }
        }
        return el;
    }


    hasClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.classList.contains($class);
    }


    hide = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;

        if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "show";
			}
		}

        if (el.classList==undefined || el.classList==null) {
            alert("what in the world???");
            return;
        }
        if (el.classList.contains("hide") == false) el.classList.add("hide");
        if (el.classList.contains("show") == true) el.classList.remove("show");
        el.style.display = "none";
        return el;
    }
    hideParent = function ($el, $bHIDESELFTOO = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.hide(parent_el);
        if ($bHIDESELFTOO == true) this.hide(el);
    }

    hideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
        }
        return el;//return the last el
    }
    emptyAndHideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
            el.innerHTML="";
        }
        return el;//return the last el
    }

    showAll = function($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.show($arr[a]);
        }
        return el;
    }

    hideAll = function ($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.hide($arr[a]);
        }
        return el;
    }

    jsonSyntaxHighlight = function ($json) {
        if ($json == null) return "";
        if (typeof $json != 'string') {
            $json = JSON.stringify($json, undefined, 2);
        }
        $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }



    masterThumbImg = function ($el, $beNOBJ, $nObj, $imgClass = "", $bNOIMGIMG = false) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var img = document.createElement("img");
        if ($imgClass != "") img.classList = $imgClass;

        //get the thumb for this if we have it,
        //falls back on class pic if we don't have thumb for this.
        var url = MEM.__GET_DEFAULT_IMG_NEURON($beNOBJ, $nObj);
        if (url == "") {
            if ($bNOIMGIMG == true) img.setAttribute("src", "imgs/unknown.png");
        }
        else img.setAttribute("src", url);
        el.append(img);
        return img;
    }


    setElBorder = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;

        if ($pendingHanai != null) el.classList.add("hanai-border");
        else if ($nObj.ghost == true) el.classList.add("ghost-border");
        else if ($nObj.freeze == true) el.classList.add("freeze-border");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-border");
        else el.classList.add("other-owned-border");
    }

    setElBG = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($pendingHanai != null) el.classList.add("hanai-bg");
        else if ($nObj.ghost == true) el.classList.add("ghost-bg");
        else if ($nObj.freeze == true) el.classList.add("freeze-bg");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-bg");
        else el.classList.add("other-owned-bg");
    }

    getImgBorder = function ($nObj, $postfix="") {
        if ($postfix!="") {
            if ($postfix.substring(0,1)!="-") $postfix="-"+$postfix;
        }
        if ($nObj == null) return "";
        if ($nObj.ghost == true) return "ghost-img"+$postfix;
        else if ($nObj.owned_by == USER._userNObj.nid) return "self-owned-img"+$postfix;
        else if (ANNE.isANNE($nObj.owned_by)==true) return "anne-owned-img"+$postfix;
        else return "other-owned-img"+$postfix;
    }


    injectPartFail = function ($x) {
        console.log("newNeuronFail");
        console.log($x);
    }


    /**
     * Update SIF options from outside
     * @param {*} $el_datalist 
     * @param {*} $nObjs 
     * @param {*} $bEMPTY 
     */
    addSIF_opts_externalupdate=function($el_datalist, $nObjs, $bEMPTY=true) {
        var el_datalist = this.__getEl($el_datalist);
        if ($bEMPTY==true) ANNE.A.empty(el_datalist);
        for (var n=0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var option = document.createElement("option");
            option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
            option.setAttribute("temp_nid", nObj.temp_nid);
            option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
            el_datalist.append(option);
        }
    }
    

    /**
     * This is intended for lists and lookups that have potentially GINORMOUS result sets.
     * We don't want the entire universe of possibilities in a data object in the gui.
     * A neuron class with 100000 options would have 100k elements in a datalist.  This is silly.
     * 
     * Here we default to what the user has been doing in the app. (context of neurons)
     * 
     * When they click that local memory AND ANNODE memory is searched.
     * 
     * @param {Element or String} $el 
     * @param {Function} $search_cb 
     * @param {int} $seachCharMin 
     * @returns 
     */
    makeANNE_SIF=function($el, $search_locl_only_cb, $search_cb=null, $selected_cb=null, 
                                $defaultNOBJS=[], $searchCharMin=2, $bDODELETE=true, $findMoreClass="") {
        const el = this.__getEl($el);
        const local_cb = $search_locl_only_cb;
        const selected_cb = $selected_cb;

        if (el == null) return;
        if (el.type!="text" && el.nodeName!="INPUT"){
            console.log("to make an ANNE SIF we need to start with an input field!");
            return;
        }

        if(typeof $search_cb!=="function") {
            console.log("to make an ANNE SIF we need a search cb!");
            return;
        }

        el.classList.add("ANNE-SIF");

        const searchCharMin = $searchCharMin;

        el.setAttribute("autocomplete", "off");
        
        var bcr = el.getBoundingClientRect();
        var el_datalist = document.createElement("datalist");
        el_datalist.setAttribute("id", el.id+"__list");
        el.parentElement.append(el_datalist);
        el.setAttribute("list", el.id+"__list");       //tie input to data list.

        const btn = document.createElement("button");  
        btn.classList="btn-yellow-sq";
        if ($findMoreClass!="") btn.classList.add($findMoreClass);
        btn.setAttribute("id", el.id+"_sif");             
        btn.innerHTML = "find more";
        //btn.style.position="absolute";
        //btn.style['margin-left'] = (bcr.right-70)+"px";
        btn.setAttribute("list_id", el.id+"__list");
        el.parentNode.append(btn);
        if (bcr.height>0) {
            btn.style['max-height']=Math.floor(bcr.height)+2+"px";
            btn.style['height']=Math.floor(bcr.height)+2+"px";
        }
        else {
            btn.style['max-height']=32+"px";
            btn.style['height']=32+"px";
        }
        btn.style['max-width']='70px';
        btn.style['width']='70px';
        ANNE.A.hide(btn);
        
        const btn2 = document.createElement("button");  

        if ($bDODELETE) {
            btn2.classList="btn-red-sq";   
            btn2.setAttribute("id", el.id+"_sif_x");             
            btn2.innerHTML = "[x]";
            //btn2.style.position="absolute";
            //btn.style['margin-left'] = (bcr.right-40)+"px";
            btn2.setAttribute("list_id", el.id+"__list");
            el.parentNode.append(btn2);
            if (bcr.height>0) {
                btn2.style['max-height']=Math.floor(bcr.height)+2+"px";
                btn2.style['height']=Math.floor(bcr.height)+2+"px";
            }
            else {
                btn2.style['max-height']=32+"px";
                btn2.style['height']=32+"px";
            }
            btn2.style['max-width']='40px';
            btn2.style['width']='40px';

            //clear the selection.
            btn2.addEventListener("click", function(e){
                el.value="";
                ANNE.A.hide(btn);
                ANNE.A.hide(btn2);
                if (selected_cb!=null) selected_cb(-1);
                return;
            });

            ANNE.A.hide(btn2);
        }


        function addOpts($nObjs) {
            if ($nObjs==undefined || $nObjs==null) return;
            for (var n=0; n < $nObjs.length; n++) {
                var nObj = $nObjs[n];
                var option = document.createElement("option");
                option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
                option.setAttribute("temp_nid", nObj.temp_nid);
                option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
                el_datalist.append(option);
            }
        }
        //add in any default tnids (like local favs)
        if ($defaultNOBJS.length) addOpts($defaultNOBJS);

        el.setAttribute("search-roots", ""); 
        el.setAttribute("local-search-roots", ""); 

        function isInRoot($val, $which="search-roots") {
            var searchRootsArr = [];
            var search_roots = el.getAttribute($which);
            if (ANNE.UTILS.isset(search_roots)==true) {
                searchRootsArr = search_roots.split(",");
                for (var s=0; s < searchRootsArr.length; s++) {
                    if (searchRootsArr[s]=="") continue;
                    if ($val.indexOf(searchRootsArr[s])==0) {
                        return true;
                    }
                }
            }
            //jam the new search root val back in.  This optimizes lookups so we don't 
            //needlessly chase extended strings of strings we've already looked up!
            searchRootsArr.push($val.trim());
            el.setAttribute($which, searchRootsArr.join(","));
            return false;
        }

        var getActualVal = function() {
            //see if it matches anything we have in the datalist.
            //if so, it is selected -- so call the callback.
            //if there are multiples matching this text the user can use the dd to selecte which one they want.
            var selected_tnid = -1;
            var lowerz = el.value.toLowerCase();
            var actualVal = el.value;

            //find the selected datalist item by its unique name.
            for (var j = 0; j < el_datalist.options.length; j++) {
                if (el_datalist.options[j].value.toLowerCase()==lowerz) {
                    selected_tnid = parseInt(el_datalist.options[j].getAttribute("temp_nid") );
                    actualVal = el_datalist.options[j].getAttribute("text");
                    break;
                }
            }
            return {actualVal:actualVal, selected_tnid:selected_tnid};
        }

        
        //CALLED ON TEXT INPUT.
        var testLocal = async($e)=>{
            //willy wonka data lists only use the value prop as display.
            //if we want a disambig display in the option then that becomes the value.
            //use extra text attr in option for real text value.
            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            var selected_tnid = avObj.selected_tnid;

            //already searched for this and it is in the widget.
            //if (actualVal.length) {//} && isInRoot(actualVal, "local-search-roots")==false) {
            if (actualVal.length>0 && local_cb!=null) {
                ANNE.A.show("loading");
                var nObjs = await local_cb(actualVal);
                ANNE.A.hide("loading");
                if (ANNE.UTILS.isset(el_datalist)==false) {
                    el_datalist = document.getElementById(btn.getAttribute("list_id"));
                    if (ANNE.UTILS.isset(el_datalist)==false) return;
                }
                ANNE.A.empty(el_datalist);
                addOpts(nObjs);
            }
            //}

            if (selected_tnid>-1 && selected_cb!=null) {
                ANNE.A.hide(btn);
                ANNE.A.show(btn2);
                selected_cb(selected_tnid);
                return;
            }
            if (actualVal.length>searchCharMin) {
                ANNE.A.hide(btn2);
                ANNE.A.show(btn);
            }
            else {
                if (selected_tnid==-1 && selected_cb!=null) {
                    ANNE.A.hide(btn2);
                    ANNE.A.hide(btn);
                    selected_cb(-1);
                    return;
                }
            }
        }
    
        
        const returnKeyPress = async($e)=> {
            $e.preventDefault();
            if ($e.keyCode === 13) {
                if (btn.classList.contains("show")==true) {
                    var event = new MouseEvent('click', {
                        'view': window,
                        'bubbles': true,
                        'cancelable': true
                    })
                    btn.dispatchEvent(event);
                }
            }
        }

        //if more than the search min, open up the anne search btn.
        //if we haven't touched this search root yet, do a local ajax style search.
        el.addEventListener("input", $e=>testLocal($e));
        el.addEventListener("keyup", $e=>returnKeyPress($e));

        //find more button
        btn.addEventListener("click", async($e)=>{
            ANNE.A.show("loading");
            ANNE.A.hide(btn);

            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            
            //already searched for this and it is in the widget.
            if (actualVal.length>0 && isInRoot(actualVal, "search-roots")==true) return;

             //add in to local-search-roots too so data list feature takes over.
            if (actualVal.length>0) isInRoot(actualVal, "local-search-roots");  
        
            var nObjs = await $search_cb($e);
            if (ANNE.UTILS.isset(el_datalist)==false) {
                el_datalist = document.getElementById(btn.getAttribute("list_id"));
            }
            if (ANNE.UTILS.isset(el_datalist)==false) return;
            ANNE.A.empty(el_datalist);
            addOpts(nObjs);
            ANNE.A.hide("loading");
        });

        return btn;
    }


    setAsGhost=function($el, $class="ghost-text") {
        var el = this.__getEl($el);
        if (el == null) return;
        el.classList.add($class);
        var span=document.createElement("span");
        span.classList = "ghost-small-blue-ico";
        el.appendChild(span);
        el.placeholder = "Ghost neuron, must send TX to ANNE to make real.";
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeQuickDeleteBtn = function($el, $cb=null) {
        var el = this.__getEl($el);
        if (el == null) return;
        const btn = document.createElement("button");  
        btn.classList="btn-red-sq";                
        btn.innerHTML = "[x]";
        el.append(btn);

        if ($cb!=null && typeof $cb ==="function") {
            btn.addEventListener("click", e=>$cb(e));
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $injid 
     * @param {*} $paramRelon 
     * @param {*} $GUID 
     * @param {*} $page 
     * @param {*} $existingClumpNid 
     * @param {*} $lbl 
     * @param {*} $bNEWOK 
     * @returns 
     */
    makeANTFromParam = function ($el, $injid, $paramRelon, $GUID, $page = "selector_action", $existingClumpNid = "", $lbl = "", $bNEWOK = true) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($paramRelon == undefined || $paramRelon == null) return;
        var relNOBJ = MEMG.getNeuronByTempNid($paramRelon.r);
        var toNOBJ = MEMG.getNeuronByTempNid($paramRelon.t);
        var antag = document.createElement("an");
        antag.setAttribute("role", relNOBJ.text);
        antag.setAttribute("rolereln", relNOBJ.nid);
        antag.setAttribute("newok", +$bNEWOK);
        antag.setAttribute("role_class_nid", toNOBJ.nid);
        if ($lbl == "") {
            var parts2 = relNOBJ.text.split("_");
            parts2.shift();
            $lbl = parts2.join("_");
        }
        antag.setAttribute("lbl", $lbl);
        antag.setAttribute("injid", $injid);
        antag.setAttribute("pguid", $GUID);
        antag.setAttribute("clump_nid", $existingClumpNid);
        antag.setAttribute("page", $page);
        el.append(antag);
    }

    /**
     * Using clone will destroy the existing events. (that's what it is here for)
     */
    moveElement = function ($subj2bmoved, $el_destination, $bCLONE = true) {
        if ($subj2bmoved == null) return;
        var el_parent = $subj2bmoved.parentElement;
        if (el_parent == null) return;
        var el_destination = this.__getEl($el_destination);
        if (el_destination == null) return;
        el_parent.removeChild($subj2bmoved);
        if ($bCLONE == true) {
            return this.cloneAndAppend($subj2bmoved, el_destination, true);
        }
        else el_destination.appendChild($subj2bmoved);
        return $subj2bmoved;
    }

    /**
     * Take a DOM node from its parent and move it to the spot of first child (prepend)
     */
    moveToFirst = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.prepend(el);
    }
    moveToEnd = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.append(el);
    }


    /**
     */
    parseANT_sentence = function ($ant) {
        console.log("%c (parseANT_sentence)", TALK.bgpink)
    }

    /**
     * Parse the ANTag (<an/>) which can show up in a blob.
     */
    parseANTags = function ($el, $bCHASE = false, $vars = null, $extraClass = "") {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");

        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];

            //sentence type is for parsing sentence
            var type = ant.getAttribute("type");
            if (type != undefined) {
                console.log("%c ANT TAG TYPE " + type, TALK.bgpink)
                if (type == "sentence") {
                    this.parseANT_sentence(ant);
                    continue;
                }
            }

            var nid = ant.getAttribute("nid");
            if (nid != undefined && nid != "") {
                var nObj = MEMG.getNeuronByNid(nid);
                var txt = "";
                var dnrt = ant.getAttribute("dnrt");
                if (dnrt != null && dnrt == 1) {
                    txt = ant.innerHTML;
                }

                ant.innerHTML = "";

                if (nObj != null) {
                    EXT_BTNS.makeNeuronLink(ant, nid, $extraClass, "", false, txt);

                }
                else if ($bCHASE == true) {
                    console.log("we need to go get this value");
                }
            }

            // to do. create helpful injectable gui
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }

        }
    }


    /**
      * Parse the ANTag (<an/>) which can show up in a blob.
      * only do vars, not text/html tags.
      */
    parseANTagsVarsOnly = function ($el, $vars = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");
        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }
        }
    }

    stopPlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
        el.currentTime = 0;
    }
    pausePlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
    }
    stopSound = function($el) { 
        this.stopPlay($el);
    }
    startSound = function($el) { this.play($el); }
    playSound = function($el) { this.play($el); }
    play = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.play();
    }

    pointerEventsNone = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="none";
        ANNE.TALK.printf_console(TALK.RED, "pointerEventsNone");
        return el;
    }
    pointerEventsAll = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="all";
        ANNE.TALK.printf_console(TALK.GREEN, "pointerEventsAll");
        return el;
    }

    removeSelf = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        el.parentNode.removeChild(el);
        return el;
    }

    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     * @param {*} $prefName 
     * @param {*} $featureProp
     * @param {*} $fpLbl
     * @param {*} $fpClass
     * 
     * @returns 
     */
    neuronIdentityDisplay = async($nObj, $bEMULATE=false, $prefName="", $featureProp="did_annex", $fpLbl="ANNEX'd neurons", $fpClass="color-green")=> {
        var anneid = $nObj._anneid;
        if (anneid==undefined || anneid==null) {
            if ($bEMULATE==true) {
                anneid = AWARE.emulateANNEID($nObj);
            }
            else {
                anneid = await ANNE.API.getANNEID($nObj, false);
            }
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "anneid-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
        
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        if ($nObj.getParent()==null) {
            await ANNE.CHASE.queryMinBem($nObj);
        }
        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="anneid-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        var name = $prefName;
        if (name=="") name = $nObj.bestNameHTML();
        acn2.innerHTML = name;
        if (acn2.innerText.length>25) acn2.innerText = acn2.innerText.substring(0, 25);
        //if ($prefName[$prefName.length-1]!=")") acn2.innerText+=")";
        acn.appendChild(acn2);

        //anneid
        var fp = UTILS.commas(anneid[$featureProp]);
        var acs = document.createElement("div");
        acs.classList = "acs";
        acs.classList.add("acs-"+$nObj.nid);    //find the smart-reference.. however we update that.
        acs.innerHTML = "<span class='"+$fpClass+"'>"+fp+"</span>&nbsp;"+$fpLbl;
        acn.appendChild(acs);
        
        return el_acf;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {object} $match
     * @param {object} $arr
     * @returns 
     */
    neuronRevelatianMatchDisplay = async($nObj, $match, $lvlOffsets)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }

        //console.log($match);

        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
        el_acf.setAttribute("overlay", true);
        el_acf.setAttribute("tnid", $nObj.temp_nid);
        el_acf.setAttribute("id", "match_tnid_"+$nObj.temp_nid);
            
            var el_donut = document.createElement("div");
            el_donut.classList = "donutNONE";
            el_acf.appendChild(el_donut);
    
            var p1 = document.createElement("p");
            p1.classList="n-circ-annex";
            el_acf.appendChild(p1);
                var sp = document.createElement("span");
                sp.classList = "REV-ico";
                p1.appendChild(sp);

                var sp2 = document.createElement("span");
                sp2.style="color:color-green font-size:20px; font-weight:900";
                sp2.innerHTML="";
                p1.appendChild(sp2)

            var p2 = document.createElement("p");
            p2.classList = "n-circ-imgfix";
            el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

            var el_revcirc = document.createElement("div");
            el_revcirc.classList="rev-circ-name";
            el_acf.appendChild(el_revcirc);
            
                var acn2 = document.createElement("div");
                acn2.classList = "rev";
                var name = $nObj.bestNameHTML();
                acn2.innerHTML = name;
                el_revcirc.appendChild(acn2);

                var lbl = document.createElement("div");
                lbl.classList="revelatian-sublbl";
                
                if ($nObj.ntype==NTYPE_INST) lbl.innerHTML = beNOBJ.bestNameHTML() + "<span class='rev-tiny-ntype'>inst</span>";
                else if ($nObj.ntype==NTYPE_CLASS) lbl.innerHTML = beNOBJ.bestNameHTML()+ "<span class='rev-tiny-ntype'>class</span>";
                el_revcirc.appendChild(lbl);

                //var lvloffset0 = document.createElement("span");
                //if ($lvlOffsets[0]!=undefined) {
                   // lvloffset0.innerHTML = "<span class='rev-lvl0ffset0'>"+$lvlOffsets[0]+"</span>";
                //}
				//lbl.appendChild(lvloffset0);
                //var lvloffset1 = document.createElement("span");
                //if ($lvlOffsets[1]!=undefined) {
                   // lvloffset1.innerHTML = "<span class='rev-lvl0ffset1'>"+$lvlOffsets[1]+"</span>";
                //}
				//lbl.appendChild(lvloffset1);

                var pathways = document.createElement("div");
                pathways.setAttribute("id", "pathways_"+$nObj.temp_nid);
                pathways.classList="revelatian-pathways";
                el_revcirc.appendChild(pathways);

        // var match = {tnid:$matchPt.tnid, myl:1, paths:paths};
        // paths (arraymap) = {pathid:path, hits:1}
        return {el:el_acf, el_pathways:pathways};
    }


    /**
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronRevelatianDisplay = async($nObj)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
       
        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "REV-ico";
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-green font-size:20px; font-weight:900";
        sp2.innerHTML="";
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="rev-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "rev";
        var name = $nObj.bestNameHTML();
        if (name.length>25) name = name.substring(0, 25);
        acn2.innerHTML = name;
        acn.appendChild(acn2);

        var lbl = document.createElement("div");
        lbl.classList="revelatian-sublbl";
       
        lbl.innerHTML = "<span>parent&nbsp;<b>"+beNOBJ.bestNameHTML().substring(0, 25)+"</b></span>";
        acn.appendChild(lbl);
        return el_acf;
    }


    /**
     * make and populate an html element to display goodies for an annexable neuron
     * this is short-hand used for the user to click to get to a more in-depth view.
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronAplexDisplay = async($nObj, $bEMULATE=false, $prefName="", $colorClass="green", $icopost="2")=> {
        var apx = $nObj._aplex;
        if (apx==null) {
            if ($bEMULATE==true) {
                apx = new aplex($nObj, $nObj.nid, null);    //3rd param applex obj
                apx.emulateAplex();
            }
            else {
                apx = await ANNE.CHASE.get_aplex_data($nObj.nid, false, "neuronAplexDisplay");
            }
        }

        var el_acf = document.createElement("div");
        el_acf.classList = "awarez-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donut";
        el_acf.appendChild(el_donut);
        if (apx.posPer==null) apx.initPercents();
        var l1 = Math.round(apx.posPer*360);
        var l2 = Math.round(apx.negPer*360);
        var varstring = "--l1:0deg "+l1+"deg; --l2:"+l1+"deg "+(l1+l2)+"deg; --l3:"+(l1+l2)+"deg 360deg;";
        el_donut.style=varstring;

        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "annex-"+$colorClass+"-ico"+$icopost;
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-"+$colorClass+" font-size:20px; font-weight:900";
        sp2.classList = "annex-price-lvl";
        sp2.classList.add("apltnid-"+$nObj.temp_nid);
        sp2.innerHTML=apx.price_lvl;
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="awarez-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        if ($prefName!="") {
            acn2.innerHTML = $prefName;
        }
        else acn2.innerHTML = $nObj.text;
        acn.appendChild(acn2);

        if ($nObj.sponsored_by!="") {
            var spNOBJ = MEMG.getOrMakeIfNew($nObj.sponsored_by);
            var acs = document.createElement("div");
            acs.classList = "acs";
            acs.innerHTML = spNOBJ.bestName();
            acn.appendChild(acs);
        }

        return el_acf;
    }



    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        return this.removeSelf(parent);
    }

    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParentParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        var parentParent = parent.parentNode;
        if (parentParent == null) return this.removeSelf(parent);
        return this.removeSelf(parentParent);
    }


    removeAll=function($class) {
        const elements = document.querySelectorAll('.'+$class);
        elements.forEach(element => element.remove());
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $classname 
     * @returns 
     */
    removeClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.remove($classname);

        //remove a 2nd, if there.
        if (el.classList.contains($classname)==true) el.classlist.remove($classname);
        
        return el;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $childClass 
     * @returns 
     */
    removeByParentAndClass = function ($el, $childClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var els = el.querySelectorAll("." + $childClass);
        for (var e = 0; e < els.length; e++) {
            var elchild = els[e];
            el.removeChild(elchild);
        }
    }

    /**
     * remove the $removeClass from all element children inside of $el
     * that match the selector.  Adds it back to the targetEl.   
     * This is used to turn an active off, then on for the selected item.
     */
    removeAddClassOfAllIn = function ($el, $selector, $removeClass, $targetEl) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($selector[0] != "#" && $selector[0] != ".") $selector = "." + $selector; //presume class.
        el.querySelectorAll($selector).forEach(target => {
            target.classList.remove($removeClass);
        });
        var targetEl = this.__getEl($targetEl);
        if (targetEl == null) return;
        if (targetEl != null) targetEl.classList.add($removeClass);
    }

    setRadioVals = function ($pre, $numMax, $checkedVal) {
        for (var c = 0; c < $numMax; c++) {
            var el = this.__getEl($pre + "_" + c);
            if (el == null) return null;
            if ($checkedVal == c) el.checked = true;
            else el.checked = false;
        }
        return;
    }


    setBodyBackground = function ($url) {
        var newImgEl = document.createElement('img');
        newImgEl.addEventListener('load', function () {
            document.body.style['background-image'] = "url('" + $url + "')";
            ANNE.A.removeSelf(newImgEl);

        });
        newImgEl.setAttribute('src', $url); // Loads the image.
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaOut=function($el, $duration=1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.hide(el);
            el.style.alpha=0;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if (el.classList.contains("hide")==false) el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaScaleOut=function($el, $duration=1, $scale=0.1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            transform:'scale('+$scale+')',
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }
    

    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaIn=function($el, $duration=1, $ease="power1.in", $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.show(el);
            el.style.alpha=1;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                    el.classList.remove("hide"); //incase another animation
                }
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaScaleIn=function($el, $duration=1, $ease="power1.in", $scale=1, $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            transform:'scale('+$scale+')',
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                }
            }
        } );
    }


	showing = function($el) {
		var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == true) return true;
		if (el.classList.contains("hide") == true) return false; 
		if (el.style.display == "none") return false;
		return true;
	}
	
    /**
     * 
     * @param {*} $el 
     * @param {*} $displayProp 
     * @returns 
     */
    show = function ($el, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == false) el.classList.add("show");
        if (el.classList.contains("hide") == true) el.classList.remove("hide");
        if (el.classList.contains("faded-neuron") == true) el.classList.remove("faded-neuron");

		if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "hide";
			}
		}
		
        if ($displayProp == "") {
            var restore = el.getAttribute("display_restore");
            if (restore!=null && restore.length>0) {
				el.style.display = restore;
			}
            else if (el.style.display == "none") {
				el.style.display = "";
			}
        }
        else el.style.display = $displayProp;
        return el;
    }


    showParent = function ($el, $bSHOWSELFTOO = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.show(parent_el);
        if ($bSHOWSELFTOO == true) this.show(el);
        return el;
    }

    showMulti = function ($arrDivids, $displayProp = "") {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            if (el == null) continue;
            el.classList.add("show");
            el.classList.remove("hide");
            el.style.display = $displayProp;
            el.style.opacity = 1; //correct in case a fadeOut was used prior
        }
        return el;//return the last el
    }

    setAttribute = function ($el, $attribute, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.setAttribute($attribute, $value);
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setDeactiveWLbl = function ($el, $prepend = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($prepend == true) var dl = ANNE.A.ezElPrepend(el, "span", "dib deactivate-lbl", "");
        else var dl = ANNE.A.ezEl(el, "span", "dib deactivate-lbl", "");
        ANNE.A.ezEl(dl, "span", "dm", "");
        return dl;
    }

    setHTML = function ($el, $html, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = $html;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setJSON = function ($el, $json, $bSHOW = true, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($json == undefined || $json == null) return;
        var str = this.jsonSyntaxHighlight($json);
        el.innerHTML = str;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * Set key up date event.
     * All dates get handled by our doDateFormat handler
     */
    setKeyUpDate = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doDateFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doDateFormat);
    }

    setKeyUpTime = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doTimeFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doTimeFormat);
    }


    setStyle = function ($el, $jsoncss, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var keys = Object.keys($jsoncss);
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsoncss[prop];
            el.style[prop] = val;
        }
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setStyleVar = function ($el, $prop, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.setProperty($prop, $val);
    }


    setValue = function ($el, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($value==null) {
            el.selectedIndex=0;
        }
        else el.value = $value;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $localRef 
     */
    soundFileLink = function($el, $localRef, $bDOWAVEFORM=true) {
        var el = this.__getEl($el);
        if (el == null) return null;
    
        
        var el_btn = document.createElement("span");
        el_btn.innerHTML = "▶️";
		el_btn.style.cursor = "pointer";
		el_btn.classList = "hoverable";
        el.appendChild(el_btn);

        
        var audio_tag = document.createElement("audio");
        audio_tag.src = $localRef;
        el.append(audio_tag);

        el_btn.addEventListener('click', function() {
            audio_tag.play();
        });


        if ($bDOWAVEFORM==false) return audio_tag;
		
        var el_wf_canvas = document.createElement("canvas");
		el_wf_canvas.style["height"]="90px";
		el_wf_canvas.style["min-width"]="100px";
		
        el.append(el_wf_canvas);
        this.drawWaveform(audio_tag, el_wf_canvas);

        return audio_tag;
    }


    /**
     * 
     * @param {*} $audio_tag 
     * @param {*} $cavasForWaveform 
     */
    drawWaveform=function($audio_tag, $cavasForWaveform) {
        const audioElement = this.__getEl($audio_tag);
        if (audioElement==null) return;

        const canvas = this.__getEl($cavasForWaveform);
        if (canvas==null) return;

        const audioContext = new (window.AudioContext || window.webkitAudioContext)();
        const audioSrc = audioContext.createMediaElementSource(audioElement);
        const analyser = audioContext.createAnalyser();

        audioSrc.connect(analyser);
        analyser.connect(audioContext.destination);

        const bufferLength = analyser.frequencyBinCount;
        const dataArray = new Uint8Array(bufferLength);

        const canvasCtx = canvas.getContext('2d');

        function draw() {
        const WIDTH = canvas.width;
        const HEIGHT = canvas.height;

        analyser.getByteTimeDomainData(dataArray);

        canvasCtx.fillStyle = 'rgb(200, 200, 200)';
        canvasCtx.fillRect(0, 0, WIDTH, HEIGHT);

        canvasCtx.lineWidth = 2;
        canvasCtx.strokeStyle = 'rgb(0, 0, 0)';

        canvasCtx.beginPath();

        const sliceWidth = (WIDTH * 1.0) / bufferLength;
        let x = 0;

        for (let i = 0; i < bufferLength; i++) {
            const v = dataArray[i] / 128.0;
            const y = (v * HEIGHT) / 2;

            if (i === 0) {
            canvasCtx.moveTo(x, y);
            } else {
            canvasCtx.lineTo(x, y);
            }

            x += sliceWidth;
        }

        canvasCtx.lineTo(canvas.width, canvas.height / 2);
        canvasCtx.stroke();

        requestAnimationFrame(draw);
        }

        draw();
    }


    copyIco = function ($el, $val) {
        var el = this.__getEl($el);
        if (el) {
            ANNE.A.empty(el);
            this.addCopyIco(el, $val);
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $data 
     * @returns 
     */
    addCopyIco=function($el, $data){
        if (ANNE.UTILS==null) return;
        var spanpre=document.createElement("button");
        spanpre.classList="ico-copy";
        spanpre.title = "copy to clipboard";
        spanpre.setAttribute("data-copy", $data);
        var sp = document.createElement("i")
        sp.classList = 'bi bi-clipboard-plus';
        spanpre.append(sp);
        spanpre.addEventListener("click", (e)=>{
            ANNE.playSound("click");
            e.stopPropagation();
            e.preventDefault();
            ANNE.UTILS.copyToClipboard($data); 
        });
        var el = this.__getEl($el)
        if (el==null) return;
        el.append(spanpre);
    }


    setMaxHeights = function ($el, $maxheight = 0, $height = 0) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($maxheight > 0) el.style["max-height"] = $maxheight + "px";
        if ($height > 0) el.style["height"] = $height + "px";
    }


    //updates any class, ie, 'bal-[temp_nid]' tags in the DOM
    quickBalanceUpdate = function ($class, $html) {
        var bals = document.querySelectorAll(".bal-" + USER._userNObj.temp_nid);
        for (var b = 0; b < bals.length; b++) {
            bals[b].innerHTML = $html;
        }
    }

    //return the span
    span = function ($html, $class = "") {
        var span = document.createElement("span");
        if ($class != "") span.classList = $class;
        span.innerHTML = $html;
        return span;
    }

    /* strip all occurances of a class out from an el or its decendants */
    stripClass = function ($el, $class, $addClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var children = $el.querySelectorAll("." + $class);
        if (children == null || children.length == 0) return;
        for (var c = 0; c < children.length; c++) {
            children[c].classList.remove($class);
            if ($addClass != "") children[c].classList.add($addClass);
        }
    }

    toggleWithFade = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.fadeOut($el);
            return "hide";
        }
        else {
            this.fadeIn($el);
            return "show";
        }
    }

    toggle = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.hide($el);
            return "hide";
        }
        else {
            this.show($el);
            return "show";
        }
    }


    toggleClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class) == true) el.classList.remove($class);
        else el.classList.add($class);
    }

    /**
     * if class1 is turned on (after the fucntion) it is -1
     * if class2 is turned on (after the function) it is 1
     */
    toggleClass2 = function ($el, $class1, $class2) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class1) == true) {
            el.classList.remove($class1);
            el.classList.add($class2);
            return 1;
        }
        else if (el.classList.contains($class2) == true) {
            el.classList.remove($class2);
            el.classList.add($class1);
            return -1;
        }
        return 0;
    }

    triggerHover = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var event = new MouseEvent('mouseover', {
            'view': window,
            'bubbles': true,
            'cancelable': true
        });
        el.dispatchEvent(event);
    }


    setMaxH = function ($el, $maxH) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style['max-height'] = parseInt($maxH) + "px";
        el.style['height'] = parseInt($maxH) + "px"
    }

    toggleByClass = function ($el, $topSelector, $targetClassSelector, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent = el.closest("." + $topSelector);
        if (parent != null) {
            var toggleTarget = parent.querySelector("." + $targetClassSelector);
            if (toggleTarget != null) {
                if (toggleTarget.classList.contains("show")) this.hide(toggleTarget);
                else this.show(toggleTarget, $displayProp);
            }
        }
    }


    /**
     * Forgot about this one.. should try it again (lol)
     * 
     * @param {*} $el 
     * @param {*} $msg 
     * @param {*} $hovertext 
     * @param {*} $attr 
     * @param {*} $classList 
     * @returns 
     */
    tooltip = function ($el, $msg, $hovertext = "hover over me", $attr = "data-text", $classList = "tooltip") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $classList;
        span.setAttribute($attr, $msg);
        span.innerHTML = "&nbsp;&#9432;"+$hovertext;
        el.append(span);
        return span;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $event 
     * @param {*} $bHIDE 
     * @returns 
     */
    trigger = function ($el, $event, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el[$event] != undefined && typeof el[$event] === 'function') {
            el[$event]();
        }
        else {
            console.log("(anne_DOM.trigger) cannot trigger event " + $event + " on element " + el.id);
            return;
        }
        if ($bHIDE == true) this.hide(el);
    }


    updateAll = function($el, $html) {
        var els = document.querySelectorAll($el);
        if (els == null || els.length==0) return null;
        for (var e=0; e < els.length; e++) {
            e.innerHTML = $html;
        }
    }


    /**
     * 
     * @param {*} $input_el 
     * @param {*} $showbtn 
     * @param {*} $rNObj 
     * @param {*} $voNOBJ 
     * @param {*} $txt 
     * @param {*} $validc 
     * @param {*} $invalidc 
     * @returns 
     */
    validator = function ($input_el, $showbtn, $rNObj = null, $voNOBJ = null, $txt = "", $validc = "valid", $invalidc = "invalid") {
        if ($txt.length) {
            if ($rNObj != null && $rNObj.format_validator != null) {
                var valid_check_rez = $rNObj.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else if ($voNOBJ != null && $voNOBJ.format_validator != null) {
                var valid_check_rez = $voNOBJ.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else {
                ANNE.A.show($showbtn);
                $input_el.classList.remove($invalidc);
                $input_el.classList.add($validc);
                return true;
            }
        }
        else {
            ANNE.A.hide($showbtn);
            $input_el.classList.remove($validc);
            $input_el.classList.add($invalidc);
            return false;
        }

        return false;
    }


    /**
     * YES, this is dangerous, if we were loading external scripts.  We aren't.
     * We are loading scripts in our own pages.
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $pgui 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
    */
    load = async ($loadIntoEl, $pgui, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    loadIntoEl.innerHTML=resp;
                    var scripts = loadIntoEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            loadIntoEl.removeChild(scripts[s]);
                        }
                        loadIntoEl.append(newscript);
                        //eval(newscript);
                    }

                    //if this has a guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($pgui._guid && ($pgui.page.multiples == true || $pgui.page.instantiate == true)) {
                        var items = loadIntoEl.querySelectorAll('.dynamic');

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + $pgui._guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + $pgui._guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + $pgui._guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + $pgui._guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }

                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }

                    else {
                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }


    /**
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $spObj 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
     */
    load2 = async ($loadIntoEl, $spObj, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    var newEl = document.createElement("div");
                    if ($loadClass != "") newEl.classList = $loadClass;
                    else newEl.classList = "pichachu";
                    newEl.innerHTML = resp;
                    loadIntoEl.append(newEl);

                    //hmm. do we want to do this?  I suppose.
                    var scripts = newEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            newEl.removeChild(scripts[s]);
                        }
                        newEl.append(newscript);
                    }

                    //if this has a spObj._guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($spObj!=null && $spObj._guid!=null && $spObj._guid!="" && $spObj._guid.length>0) {
                        var items = newEl.querySelectorAll('.dynamic');
                        var guid=$spObj._guid;

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }
                        return resolve(newEl);
                    }

                    else {
                        return resolve(newEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }
    
}

var A = new anne_dom();
A.set_init(A);

/************** (from file ./js/base/anne_DOM.js)**/ 

class anne_utils extends anne_singleton {
    constructor(){
        super(false);
	}

    formatBitcoinPrice=function($price) {
        if (isNaN($price) || $price < 0) {
            throw new Error("Invalid input: Please provide a non-negative number.");
        }
        return $price.toFixed(8);
    }

    exists = function($val) {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        return true;
    }
    existsNonZero() {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        if ($val==0) return false;
    }
    ne($a,$b) {
        if (this.exists($a)==false) {
            if (this.exists($b)==true) return true;
            else return false;
        }
        if (this.exists($b)==false) return true;
        if ($a==$b) return false;
        return true;
     }

    quickAbbreviateName = function($text) {
        $text =  $text.trim();
        if ($text.length<10) return $text;
        const suffix = ["jr", "jr.", "sr", "sr.", "iii", "ii", "iv", "i", "v", "vi", "vii", "viii", "ix"];
        var parts = $text.split(" ");

        if (parts.length==2 && $text.length<17) return $text;

        var abbr = "";
        var first = "";
        var lastpart = parts[parts.length-1];
        var addsuffix = "";
        var pmax = parts.length;

        if (suffix.indexOf(lastpart.toLowerCase())!=-1) {
            addsuffix = lastpart;
            pmax = pmax-1;
        }

        for (var p=0; p < pmax; p++) {
            if (p==0) {
                abbr+=parts[p][0];
                first=parts[p];
            }
            else if (p < pmax-1){
                abbr+=parts[p][0]; //JD Rockfeller Jr
            }
            else if (p==(pmax-1)) abbr+=" "+parts[p];
            
        }
        return abbr;
    }

    stripAllNonAlphaChar=function($str, $charSlot) {
        var char = $str[$charSlot];
        if("abcdefghijklmnopqrstuvwzyz".indexOf(char)==-1) {
            $str = $str.substring(1);
            return this.stripAllNonAlphaChar($str, $charSlot);
        }
        return $str;
    }
    stripNonAlphaNum = function($str) {
        return this.stripNonIn($str, "abcdefghijklmnopqrstuvwzyz1234567890");
    }

    isMixedCase = function($str) {
        return /[a-z]/.test($str) && /[A-Z]/.test($str);
    }

    /**
     * note to self: comment this so i remember what it is.
     * @param {*} $str 
     * @param {*} $ifstr 
     * @returns 
     */
    extractNumberIf = async($str, $ifstr)=>{
        if ($str.indexOf($ifstr)!=-1) {
            if ($afterONLY==true) {
                var i = $str.indexOf($ifstr)+$ifstr.length;
                var newstr = $str.substr(i);
                return this.extractNumber($str);
            }
            else {
                return this.extractNumber($str);
            }
        }
        return null;
    }

    isSameYear = function ($nObj1, $nObj2) {
        var year1 = -1;
        var year2 = -1;
        
        if ($nObj1.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year1 = parseInt($nObj1.text);
        else {
            var year_relon = ANNE.MEM.getHasRelon($nOb1.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year1 = parseInt(yearNOBJ.text);
            }
        }

        if ($nObj2.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year2 = parseInt($nObj2.text);
        else {
            var year_relon = ANNE.MEMG.getHasRelon($nObj2.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year2 = parseInt(yearNOBJ.text);
            }
        }

        if(year1!=-1 && year2!=-1 && year1==year2) return true;
        return false;
        
    }


    extractNumber2 = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        parts.splice(p, 1);
                        return {nObj:numNOBJ, str:part, remaining:parts.join(" ")};
                    }
                    break;
                }
            }
        }
        return null;
    }


    extractNumber = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        return numNOBJ;
                    }
                    break;
                }
            }
        }
        return null;
    }

    extractNumberStr = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    return number+"";
                }
            }
        }
        return "";
    }


    replaceWeirdLetters = function($str) {
        const regex = /[À-ÖØ-öø-ÿ]/g;
        const AtoZ = {
          À: "A",
          Á: "A",
          Â: "A",
          Ã: "A",
          Ä: "A",
          Å: "A",
          Æ: "AE",
          Ç: "C",
          È: "E",
          É: "E",
          Ê: "E",
          Ë: "E",
          Ì: "I",
          Í: "I",
          Î: "I",
          Ï: "I",
          Ð: "D",
          Ñ: "N",
          Ò: "O",
          Ó: "O",
          Ô: "O",
          Õ: "O",
          Ö: "O",
          Ø: "O",
          Ù: "U",
          Ú: "U",
          Û: "U",
          Ü: "U",
          Ý: "Y",
          Þ: "TH",
          ß: "ss",
          à: "a",
          á: "a",
          â: "a",
          ã: "a",
          ä: "a",
          å: "a",
          æ: "ae",
          ç: "c",
          è: "e",
          é: "e",
          ê: "e",
          ë: "e",
          ì: "i",
          í: "i",
          î: "i",
          ï: "i",
          ð: "d",
          ñ: "n",
          ò: "o",
          ó: "o",
          ô: "o",
          õ: "o",
          ö: "o",
          ø: "o",
          ù: "u",
          ú: "u",
          û: "u",
          ü: "u",
          ý: "y",
          þ: "th",
          ÿ: "y",
        };
        var result = $str.replace(regex, (match) => AtoZ[match]);
        console.log(result); // Output: "AAA EEE III OOO UUU"
        return result;
    }


    containsTXID=function($rels, $txid) {
        for (var r=0; r <$rels.length; r++) {
            if ($rels[r].txid==$txid) return true;
        }
        return false;
    }
    

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    fixHack=function($str=null, $replace_underscores="") {
        if ($str==undefined || $str==null || typeof $str !="string") {
            console.log("don't call this, sir, and non-existant $str");
            return "";
        }

        //dont muck with date looking strings.
        var aParts = $str.split("-");
        if (aParts.length>=2) {
            var numbtest = parseInt($str.replaceAll("-", ""));
            if (numbtest > 0) {
                var parenObj = ANNE.UTILS.stripParens($str);
                console.log(parenObj);
                return parenObj;
            }
        }
        
        $str = $str.replaceAll("-2D","-");
        $str = $str.replaceAll("-28", "(");
        $str = $str.replaceAll("-29", ")");

        //too tired to chase it.  lazy kill
        //we already yank info from wikispooks in a hacky non rdf way..
        //so it comes with drama
        var apostrophe = $str.indexOf("-27");
        if (apostrophe!=-1 && apostrophe!=0 && apostrophe!=$str.length-1) {
            $str = $str.replaceAll("-27", "'");
        }
        
        var fix = $str.indexOf("-28unidentified-29");
        if (fix!=-1 && fix!=0) {
            $str = $str.replaceAll("-28unidentified-29", "");
        }

        $str = $str.replaceAll("-26", "&");
        $str = $str.replaceAll("27-", "");

        var parenObj = ANNE.UTILS.stripParensMulti($str);

        if ($replace_underscores!="") {
            parenObj.str = parenObj.str.replaceAll("_", $replace_underscores).trim();
        }
        else parenObj.str.trim();
        
        //console.log(parenObj);
        return parenObj;
    }

    
    blowOutAmpersands = function($arr) {
        var finalArray = [];
        for (var a=0; a < $arr.length; a++) {
            var val = $arr[a];
            if (val.indexOf("&")==-1) finalArray.push(val);
            else {
                var parts = val.split(" & ");
                for (var p=0; p < parts.length; p++) {
                    var val2 = parts[p].trim();
                    if (finalArray.indexOf(val2)==-1) finalArray.push( parts[p].trim() );
                }
            }
        }
        return finalArray;
    }


    //bob and suzy smith
    //s.b.  "bob smith", "suzy smith"
    extrapAmperStr = function($str) {
        var parts = $str.split(" & ");
        if (parts.length==1) part2 = $str.split(" and ");
        if (parts.length!=2) return null;
        
        var last_part = parts[ parts.length-1];
        var first_part = parts [0];
        if (last_part.split(" ").length > first_part.split(" ").length) {
            var last = last_part.split(" ");
            var last_name = last [ last.length-1];
            var first_person = first_part+" "+last_name;
            return [first_person, last_part];
        }
        return null;
    }

    /**
     * Strips a string of its paren and returns the new string with the stripped value in the context.
     * @param {*} $str 
     * @returns 
     */
    stripParens = function($str){
        const regex = /\s?\(([^)]+)\)/;
        var match = $str.match(regex);
        var newStr = $str.replace(/\s?\([^)]+\)/, '');
        var context = "";
        if (match && match.length>0) context = match[1];
        if (newStr==null || newStr=="" || newStr.length==0) newStr = $str;
        return {str:newStr.trim(), context:context};
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    prepNakedSearch = function($str) {
        var pOBJ = this.stripParens($str);
        pOBJ.inputStr = $str;
        const inputString = pOBJ.str;
        const regexPattern = /(['"])(.*?)\1/;

        let QUOTED = false;
        let extractedString = '';

        const match = inputString.match(regexPattern);
        if (match) {
            QUOTED = true;
            extractedString = match[2]; 
        }

        if (QUOTED==true) {
            pOBJ.bEXACT=true;
            pOBJ.str = extractedString;
        }
        else pOBJ.bEXACT=false;

        console.log("naked search:", pOBJ);
        return pOBJ;
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    stripParensMulti=function($str) {
        const regex = /\(([^)]+)\)/;
        const match = $str.match(regex);
        let newStr = $str.replace(/\([^)]+\)/, '');
        let context = [];
        if (match && match.length > 0) {
            context = match[1].split(',').map(item => item.trim());
        }
        if (newStr == null || newStr == "" || newStr.length == 0) {
            newStr = $str;
        }
        return {str: newStr.trim(), context: context};
    }


    /**
     * 
     * @param {*} $str 
     * @param {*} $stripifnotthis 
     * @returns 
     */
    stripNonIn=function($str, $stripifnotthis) {
        var newstr = $str.concat();
        for (var a=0; a < newstr.length; a++)  {
            var chara = newstr[a];
            if ($stripifnotthis.indexOf(chara)==-1) {
                newstr.slice(a, 1);
                a--;
            }
        }
        return newstr;
    }


    /**
     * For testing purposes, the firstseen datestamp marks
     * when the annode first saw a relon.
     * 
     * If the relon was in the mempool (time/block=-1) but the annode restarted..
     * and the relon never made it back to the mempool...
     * 
     * We have a faux relon in our localDB that the ANNE network doesn't actually know
     * about.  Our annode may know of it, but the the network does not.
     * 
     * We need to fix our indexDB.
     * 
     * We need to solve this problem in the annode as well.
     * Shutdown mid-mempool will leave neuron indexes created, neurons froze, etc..
     * which will reject the next attempt to create the same neuron indexes
     * (and will actually resolve the nid index conflict by creating a new one)
     * 
     * This will be a HUGE problem as that annode will be out of concensus.
     * @param {*} $sec 
     * @returns 
     */
    shouldBeMined = function($sec) {
        var se = this.timeElapsed($sec);
        if (se > 600) return true;
        return false;
    }

    timeElapsed = function($secs) {
        var currentTimestamp = Date.now();
        var timeElapsed = currentTimestamp - parseInt($secs);
        var secondsElapsed = Math.floor(timeElapsed / 1000);
        console.log("Time elapsed in seconds: " + secondsElapsed);
        return secondsElapsed;
    }


    secsElapsedToDHMS = function($secs) {
        const secmin = 60;
        const minshr = 60;
        const sechr = secmin*60;
        const secsday = sechr*24;
        var hrsday = 24;
        var minutes = Math.floor($secs/60);

        if (minutes > secmin) {
            var hours = Math.floor(minutes/60);
            if (hours > hrsday) {
                var days = Math.floor(hours / hrsday);
                var hrs_left = hours - (days*hrsday);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var dayswd = ANNE.TALK.pluralizeNounIf(days, "day", "s");
                var hrswd = ANNE.TALK.pluralizeNounIf(hrs_left, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
                return days+dayswd+" "+hrs_left+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
            else {
                hours = Math.floor(hours);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var hrswd = ANNE.TALK.pluralizeNounIf(hours, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s");
                return hours+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
        }
        else if (minutes > 0) {
            var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
            var secs_left = Math.round($secs % 60);
            var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
            return minutes +minswd+" "+secs_left+secsswd;
        }
        else {
            var secsswd = ANNE.TALK.pluralizeNounIf($secs, "sec", "s")
            return $secs+secsswd;
        }
    }

    forceInt = function($int) {
        if ($int==undefined || $int==null || $int=="") return 0;
        var int2 = parseInt($int);
        if (isNaN(int2)==true) return 0;
        return int2;
    }

    capFirstLetters = function($str) {
        var wds = $str.split(" ");
        for (var p=0; p < wds.length; p++) {
            wds[p] = wds[p][0].toUpperCase() + wds[p].substr(1);
        }
        return wds.join(" ");
    }

    concatUniqueNObjs=function($a1, $a2) {
        if ($a1.length==0) return $a2;
        if ($a2.length==0) return $a1;
        var a_temp_nid_list=[];
        for (var a=0; a < $a1.length; a++) {
            a_temp_nid_list.push($a1[a].temp_nid);
        }
        var a2_uniques = [];
        for (var a2=0; a2 < $a2.length; a2++) {
        var a2_temp_nid = $a2[a2].temp_nid;
            if (a_temp_nid_list.indexOf(a2_temp_nid)!=-1) continue;
            a2_uniques.push($a2[a2]);
        }
        return $a1.concat(a2_uniques);
    }


    copyToClipboard=function($str) {
        var el = document.createElement('textarea');
        el.value = $str;
        el.setAttribute('readonly', '');
        el.style = {position: 'absolute', left: '-9999px'};
        document.body.appendChild(el);
        el.select();
        document.execCommand('copy');
        document.body.removeChild(el);
    }

    concatUniqueRels = function ($arr1, $arr2) {
        var txids = [];
        var rels = [];
        for (var t=0; t < $arr1.length; t++) {
            var txid = $arr1[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr1[t]);
            }
        }
        for (var t=0; t < $arr2.length; t++) {
            var txid = $arr2[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr2[t]);
            }
        }
        return rels;
    }
    

    convertToCSV = function ($objArray, $bGENHEADER=false) {
        var array = typeof $objArray != 'object' ? JSON.parse($objArray) : $objArray;
        var str = '';

        if ($bGENHEADER==true) {
            var keys = Object.keys(array[0]);
            var header = new Object();
            for (var k=0; k < keys.length; k++) {
                header[keys[k]]=keys[k];
            }
            array.unshift(header);
        }

        for (var i = 0; i < array.length; i++) {
            var line = '';
            for (var index in array[i]) {
                if (line != '') line += ','

                line += array[i][index];
            }
            str += line + '\r\n';
        }
        return str;
    }

    l33t_checkForce = function($str) {
        var l33ts_num = [];
        var l33ts_alpha = [];
        
        //1234567890
        //1=l
        for (var s=0; s < $str.length; s++) {
            var char = $str[s]=$str[s];
            if (char=="0") {
                l33ts_num.push("0");
                l33ts_alpha.push("O");
            }
            else if (char=="1") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else if (char=="2") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else {
            } 
        }

    }


    /**
     * Alpha num.
     * single spaces ok.
     */
    cleanAnorgName = function($propName) {
        var str = $propName.replace(/ +(?= )/g,'');
        var cleanstr = str.replace(/[^a-z0-9 ]/gi,'');
        return cleanstr;
    }


    cleanPropName = function($propName) {
        var b = $propName.replace(/[^a-z0-9_]/gi,'');
        if (b.match(/^\d/)==true) {
            console.log("(utils.cleanPropname) result: "+"p_"+b);
            return "p_"+b;
        }
        return b;
    }

    roundThenCommas=function($x) {
        var rounded = Math.round($x);
        return this.commas($x);
    }

    //.toFixed(8);
    abbreviateNumber=function($num) {
        $num = new Number($num);
        if ($num >= 1e12) {
            return (Math.round($num / 1e11) / 10) + 't'; 
        } 
        else if ($num >= 1e9) {
            return (Math.round($num / 1e8) / 10) + 'b'; 
        } 
        else if ($num >= 1e6) {
            return (Math.round($num / 1e5) / 10) + 'm'; 
        } 
        else if ($num >= 1e3) {
            return (Math.round($num / 1e2) / 10) + 'k'; 
        } 
        else {
            return $num.toString(); 
        }
    }

    commas=function($x, $bDOSPACES=false) {
        if ($x === undefined || $x === null) {
            console.log("huh?");
            return;
        }

        let parts = $x.toString().split('.');
        let integerPart = parts[0];
        let decimalPart = parts.length > 1 ? '.' + parts[1] : '';
    
        if ($bDOSPACES === true) {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, " ");
        } 
        else {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        }
    
        return integerPart + decimalPart;
    }


    estimateTimeByBlocks = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        const currentDate = new Date();
        const futureDate = new Date(currentDate.getTime() + secsMore * 1000);
        const options = { month: 'long', day: 'numeric', year: 'numeric' };
        const formattedDate = futureDate.toLocaleDateString('en-US', options);
        console.log(formattedDate);
        return formattedDate;
    }   

    estimateTimeByBlocksLeft = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        return this.secsElapsedToDHMS(secsMore);
    }

    
    firstCap = function($str, $bALL=true) {
        if ($bALL==true) {
            var newstr = [];
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
            for (var p=0; p < parts.length; p++) {
                var str = parts[p][0].toUpperCase()+parts[p].substring(1);
                newstr.push(str);
            }
            return newstr.join(" ");
        }
        else {
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
        } 
    }


    //won't get fully spelled numbers like "one"
    isNum = function($n) {
        if (!isNaN(parseFloat($n)) && isFinite($n)==true) return true;
        const regex = /^\d+(st|nd|rd|th)?$/;
        return regex.test($n);
    }
    

    /**
     * 
     * from https://gist.github.com/tralves/9e5de2bd9f582007a52708d7d4209865
     * @param {*} bytes 
     * @returns 
     */
    frenlySize = function (bytes) {
        var thresh = 1024;
        if (Math.abs(bytes) < thresh) return bytes + ' B';
        var units = ['KB','MB','GB','TB','PB','EB','ZB','YB'];
        var u = -1;
        do {
            bytes /= thresh;
            ++u;
        } 
        while(Math.abs(bytes) >= thresh && u < units.length - 1);
        return bytes.toFixed(1)+' '+units[u];
    }



    /**
     * Calculate annex price lvl factor via nth fib.
     * 
     * could be done with matrix math but we won't be getting high lvl numbers here
     * and that optimization is just showing off.
     * this suffices.
     * 
     * DO NOT CHANGE THIS FUNCTION UNLESS YOU ALSO CHANGE THE FUNCTION ON THE ANNODE.
     * 
     * @param {*} $nth 
     * @returns 
     */
    nthFib=function($nth){
        if ($nth==1 || $nth==2) return 1;
        var seq = [1, 1];
        var nextval = 0;
        for (var n=2; n < $nth; n++) {
            nextval = seq[n-1] + seq[n-2];
            seq.push(nextval);
        }
        return nextval;
    }


    /**
     * Only 2 decimal places
     * @param {*} $annies 
     * @param {*} $gap 
     * @returns 
     */
    displayAnneValue = function($annies, $gap=" ") {
        var parts = ((parseInt($annies)/ANNE.ANNE_TO_ANNIES)+"").split(".");
        if (parts.length==2) {
            var whole = UTILS.commas(parts[0]);
            var deci = parts[1];
        }
        else if (parts.length==1) {
            if ($annies<ANNE.ANNE_TO_ANNIES) {
                var whole = 0;
                var deci = parts[0];
            }
            else {
                var whole = UTILS.commas(parts[0]);
                var deci = "0";
            }
        }
        if (deci==undefined) return whole;
        deci = deci+"";
        
        if (deci.length==2) deci = "."+$gap+deci;
        else if (deci.length==1) deci = "."+$gap+deci+"0";
        else if (deci.length==0) return whole;
        
        return whole;//+deci;
    }


    /**
     * 
     * @param {*} $val 
     * @param {*} $minDigits 
     * @returns 
     */
    formatDigits = function($val, $minDigits=4) {
         if ($val==null) {
            console.log("%cERROR val can't be null here", TALK.bgred);
            return "";
        }
        var final = $val+"";
       
        if (final.length<$minDigits) {
            var dif = $minDigits-final.length;
            for (var d=1; d <= dif; d++) { final = "0"+final; }
        }
        return final;
    }

    
    //For calculating edit distance
    editDistance = function (s1, s2) {
        s1 = s1.toLowerCase();
        s2 = s2.toLowerCase();
        var costs = new Array();
        for (var i = 0; i <= s1.length; i++) {
            var lastValue = i;
            for (var j = 0; j <= s2.length; j++) {
                if (i == 0)
                    costs[j] = j;
                else {
                if (j > 0) {
                    var newValue = costs[j - 1];
                    if (s1.charAt(i - 1) != s2.charAt(j - 1)) newValue = Math.min(Math.min(newValue, lastValue), costs[j]) + 1;
                    costs[j - 1] = lastValue;
                    lastValue = newValue;
                }
            }
        }
        if (i > 0)
            costs[s2.length] = lastValue;
        }
        return costs[s2.length];
    }


    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportJSONFileViaID=function($dataEl, $fileTitle, $el="", $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportJSONFile(jsonStringData, $fileTitle, $el, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $object 
     * @param {*} $fileTitle 
     */
    exportJSONFile=function($object, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.json' || 'export.json';
        var jsonstring = JSON.stringify($object);
        const bytes = new TextEncoder().encode(jsonstring);
        const json = new Blob([bytes], {
            type: "application/json;charset=utf-8"
        });

        var link = document.createElement("a");
        if (link.download !== undefined) { // feature detection
            // Browsers that support HTML5 download attribute
            var url = URL.createObjectURL(json);
            link.setAttribute("href", url);
            link.setAttribute("download", exportedFilenmae);
            link.style.visibility = 'hidden';
            var el = null;
            el = document.body;
            el.appendChild(link);
            link.click();
            el.removeChild(link);
        }
    }

    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFileViaID=function($dataEl, $fileTitle, $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportCSVFile(jsonStringData, $fileTitle, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $jsonStringOrObjectArray 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFile=function($jsonStringOrObjectArray, $fileTitle,  $bGENHEADER=true) {
        var csv = this.convertToCSV($jsonStringOrObjectArray, $bGENHEADER);
        this.exportCSVFile_inner(csv, $fileTitle);
    }

    /**
     * 
     * @param {*} $csv 
     * @param {*} $fileTitle 
     */
    exportCSVFile_inner=function($csv, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.csv' || 'export.csv';
        var blob = new Blob([$csv], { type: 'text/csv;charset=utf-8;' });
        if (navigator.msSaveBlob) { // IE 10+
            navigator.msSaveBlob(blob, exportedFilenmae);
        } 
        else {
            var link = document.createElement("a");
            if (link.download !== undefined) { // feature detection
                // Browsers that support HTML5 download attribute
                var url = URL.createObjectURL(blob);
                link.setAttribute("href", url);
                link.setAttribute("download", exportedFilenmae);
                link.style.visibility = 'hidden';
                var el = null;
                el = document.body;
                el.appendChild(link);
                link.click();
                el.removeChild(link);
            }
        }
    }

    posNegRand = function($fact) {
        return (Math.random()-.5)*$fact;
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValAndRemoveInObjnArr = function ($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return true;
            }
        }
        return false;
    }

    /**
     * 
     * @param {*} $rels 
     * @param {*} $skip_tnids 
     * @returns 
     */
    getLatest = function($rels, $skip_tnids=[]) {
        var lastValid = null;
        for (var sr=0; sr < $rels.length; sr++) {
            var rel = $rels[sr];
            if ($skip_tnids.length && $skip_tnids.indexOf(rel.r)!=-1) continue;
            if (lastValid==null) lastValid = rel;

            //if we have a bigger block height OR a 0 (indicating mempool not updated yet)
            else if ((lastValid.time) > 0 && (rel.time > 0 || rel.time==0)) {
                if (rel.time > lastValid.time) lastValid = rel;
            }

            //this has been firstseen later, meaning it is newer
            else if (lastValid.firstseen < rel.firstseen) {
                lastValid = rel;
            }
        }
        return lastValid;
    }


    /**
     * 
     * @param {*} $rels 
     * @returns 
     */
    getLastSeen = function($rels) {
        if ($rels==null) return null;
        if ($rels.length==0) return null;
        if ($rels.length==1) return $rels[0];
        var last_seen = 0;
        var lastseenRel = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (rel.firstseen!=undefined && rel.firstseen > last_seen) {
                lastseenRel = rel;
                last_seen = rel.firstseen;
            }
        }
        if (lastseenRel!=null) return lastseenRel;
        if ($rels.length) return $rels [ $rels.length-1 ];
    }

        
    /**
     * 
     * @param {*} $rels 
     * @param {*} $annie 
     * @returns 
     */
    getFirst = function($rels, $annie=0) {
        if ($annies>0) return this.getFirstWithGTEAnnies($rel, $annies);
        var first = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (first==null) first=rel;
            else if ((rel.time < first.time) || (rel.time==0 && first.time!=0)) first=rel;
            else if (rel.time==first.time) {
                if (rel.firstseen < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * attemps to handle multi payments if there was underpayment.
     * will use time/first seen from the last payment in the bunch to qualify
     * the payment as being the first seen (winner) of the FCFS 
     * 
     * @param {*} $rels 
     * @param {*} $annies 
     * @returns 
     */
    getFirstWithGTEAnnies = function($rels, $annies) {
        var first = null;
        var from_multi = new Object();

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var annies = rel.annies; 
            var rel_fs = rel.firstseen;
            var rel_time = rel.time;

            if (annies==undefined || annies==0) continue;
            if (annies < $annies) {
                var fm_this = from_multi[ rel.f ];
                if (fm_this==undefined) {
                    from_multi[ rel.f ] = {annies:annies, time:rel.time, firstseen:rel.firstseen, rel:rel};
                }
                else {
                    fm_this.annies += annies; 
                    if (rel.time > fm_this.time) {
                        fm_this.time = rel_time;
                        fm_this.firstseen = rel_fs;
                        fm_this.rel = rel;
                    }

                    if (fm_this.annies < $annies) continue; //no enough, still!
                }
            }

            if (first==null) first=rel;
            else if ((rel_time < first.time) || (rel_time==0 && first.time!=0)) first=rel;
            else if (rel_time==first.time) {
                if (rel_fs < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * 
     * @param {*} $txt 
     * @param {*} $bDECIMALOK 
     * @returns 
     */
    inputNumberOnly = function ($txt, $bDECIMALOK=false) {
        var newString = "";
        for (var c=0; c < $txt.length; c++) {
            var charCode = $txt.charCodeAt(c);
            //console.log(charCode);
            if ($bDECIMALOK==false) {
                if (charCode < 48 || charCode> 57) continue;
                newString = newString + $txt[c];
            }
            else if ($bDECIMALOK==true) {
                if (charCode < 48 || (charCode> 57 && charCode!=190)) continue;
                newString = newString + $txt[c]; 
            }  
        }
        return newString;
    }


    /**
     */
    addUniqueRel=function($arr, $rel) {
        UTILS.addUnique($arr, "txid", $rel);
    }


    /**
     * 
     * @param {Array} $objArr 
     * @param {Array} $destination_arr 
     * @param {String} $uniqueObjProp 
     */
    addUniqueObjArray=function($objArr, $destination_arr, $uniqueObjProp="txid") {
        for (var r=0; r < $objArr.length; r++) {
            var newObj = $objArr[r];
            UTILS.addUnique($destination_arr, $uniqueObjProp, newObj);
        }
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $tnids 
     */
    addUniqueTnids=function($arr, $tnids) {
        for (var t=0; t < $tnids.length; t++) {
            var tnid=$tnids[t];
            if ($arr.indexOf(tnid)==-1) $arr.push(tnid);
        }
    }



    /**
     * 
     * @param {Array} $arr 
     * @param {String} $uniqueProp 
     * @param {Object} $newRelon 
     * @returns 
     */
    addUnique=function($arr, $uniqueProp="txid", $newRelon) {
        if ($arr==null) return false;
        if ($newRelon==null) return false;

        if (this.findValInObjnArr($arr, $uniqueProp, $newRelon[$uniqueProp])==null) {
            $arr.push($newRelon);
            return true;
        }
        return false;
    }


    /**
     * 
     */
    deepCopy=function($obj) {
        if ($obj === null || typeof $obj !== 'object') {
            return $obj;
        }
    
        let copy = new $obj.constructor();
    
        for (let key in $obj) {
            if ($obj.hasOwnProperty(key)) {
                copy[key] = this.deepCopy($obj[key]);
            }
        }
        return copy;
    }
    
    
    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValInObjnArr = function ($arr, $prop, $val) {
        if ($prop!="") {
            for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return $arr[a];
            return null;
        }
        for (var a=0; a < $arr.length; a++) if ($arr[a]==$val) return $arr[a];
        return null;
    }


    /**
     * Does an item exist in the array with the existing prop value set?
     * @param {*} $arr 
     * @param {*} $prop 
     * @returns 
     */
    findValInObjnArr2 = function ($arr, $prop, $val) {
        for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return true;
        return false;
    }


    sortByProp = function($arr, $prop) {
        $arr.sort( 
            function($a, $b) {
                var val_a = $a[$prop];
                if (val_a==undefined) return 0;

                var val_b = $b[$prop];
                if (val_b==undefined) return 0;

                if (val_a < val_b) return -1;
                else if (val_a > val_b) return 1;
                return 0;
            }
        );
        return $arr;
    }


    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    removeObjInObjArr = function($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return;
            }
        }
    }


    //weak attempt to add .png to end of bad imgur url
    fixImgURL=function($imgUrl) {
        if ($imgUrl.indexOf("imgur")==-1) return $imgUrl;   //do nothing
        var aParts = $imgUrl.split(".");
        aParts.reverse();
        var ext = aParts[0];
        aParts.reverse();
        if (ext.length<3 || ext.length>4) $imgUrl = $imgUrl+".png";
        else if (ext!="png" && ext!="jpg") $imgUrl = $imgUrl+".png";
        return $imgUrl;
    }


    /**
     * try to format the bigger number.
     * if it has a deci and we can't short-hand it with a B or M or K,
     * then just format the number with commas.
     */
    formatNumber=function($num, $bROUND=false) {
        var num_parts = ($num+"").toString().split(".");
        var num = parseInt(num_parts[0]);
        if (num >0 && num_parts.length>1 && $bROUND==false) return this.formatNumberCommas($num);

        var b = num / 1000000000;
        if (b>=1) {
            if ($bROUND==true) {
                var b100=Math.floor(b*100)/100;
                num_parts[0]=((b100+"").split(".").join("."))+" B";
            }
            else num_parts[0]=b+"B";
        }
        else {
            var m = num / 1000000;
            if (m>=1) {
                if ($bROUND==true) {
                    var m100=Math.floor(m*100)/100;
                    num_parts[0]=((m100+"").split(".").join("."))+" M";
                }
                else num_parts[0]=m+"M";
            }
            else {
                var k = num / 1000;
                if (k>=1) {
                    if ($bROUND==true) {
                        var k100=Math.floor(k*100)/100;
                        num_parts[0]=((k100+"").split(".").join("."))+" K";
                    }
                    else num_parts[0]=k+" K";
                }
                else num_parts[0]=num_parts[0];
            }
        }
        if ($bROUND==true) return num_parts[0];
        else if (num_parts.length==1) return num_parts[0];
        return num_parts[0]+"."+num_parts[1];
    }


    formatNumberCommas=function($num) {
        var num_parts = ($num+"").toString().split(".");
        num_parts[0] = num_parts[0].replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        return num_parts.join(".");
    }

    convArr=function($arr) {
        var final_arr = [];
        var keys = Object.keys($arr);
        for (var k=0; k < keys.length; k++) {
            final_arr.push ( $arr [ keys[k]]);
        }
        return final_arr;
    }
    

    getSelectedOptionFromDD=function($dd_el) {
        var sel = document.getElementById($dd_el);
        return this.getSelectedOption(sel);
    }

    getSelectedOption = function($sel) {
        var opt;
        for ( var i = 0, len = $sel.options.length; i < len; i++ ) {
            opt = $sel.options[i];
            if ( opt.selected === true ) {
                break;
            }
        }
        return opt;
    }

    getUrlVars = function () {
        var vars = {};
        var parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function (m, key, value) {
            vars[key] = value;
        });
        return vars;
    }

    hex_to_ascii=function($str1) {
        var hex  = $str1.toString();
        var str = '';
        for (var n = 0; n < hex.length; n += 2) {
            str += String.fromCharCode(parseInt(hex.substr(n, 2), 16));
        }
        return str;
    }

    htmlEntities=function(str) {
        return String(str).replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/"/g, '&quot;').replace(/'/g, '&apos;');
    }

    
    isPrime($x){     
        var d = $x-1;
        while (d > 1){
          if (($x % d) == 0) return false;
          d--;
        }
        return true;
      }


    /**
     * 
     * @param {*} $map 
     * @param {*} $prop 
     * @returns 
     */
    __sortNumObjMapByProp = function($map, $prop="other") {
        var keys = Object.keys($map);
        keys.sort( function($a, $b) {
            var val_a = $map[$a][$prop];
            if (val_a==undefined) return 0;

            var val_b = $map[$b][$prop];
            if (val_b==undefined) return 0;

            if (val_a < val_b) return -1;
            else if (val_a > val_b) return 1;
            return 0;
        });
        
        var sortedMapObj = new Object();
        keys.forEach(key => {
            sortedMapObj[key] = $map[key];
          });
        return sortedMapObj;
    }


    __sortRelnKeyMap = function($relnMAP) {
        var keys = Object.keys($relnMAP);
        keys.sort( function($a, $b) {
            var rels_a = $relnMAP[$a];
            var text_a = MEMG.getNeuronByTempNid(rels_a[0].r).text.toLowerCase();
            var rels_b = $relnMAP[$b];
            var text_b = MEMG.getNeuronByTempNid(rels_b[0].r).text.toLowerCase();
            if (text_a < text_b) return -1;
            else if (text_a > text_b) return 1;
            return 0;
        });
        return keys;
    }

    endsWithStrip = function($str, $char=",") {
        if ($str[$str.length-1]==$char) {
            return $str.substring(0, $str.length-1);
        }
        return $str;
    }
    
    /**
     * 
     * @param {*} $val 
     * @returns 
     */
    isset = function($val) {
        if ($val==undefined || $val==null) return false;
        var _type = typeof $val;
        if (_type =="boolean") return $val;
        
        if (_type =="string") { 
            if($val.length==0) return false;
            return true;
        }
        if (_type=="number") {
            if($val==0) return false;
        }
        if (Array.isArray($val)==true && $val.length==0) return false;
        return true;
    }


    /**
     * move over unique props from one object to another
     * This is helpful when merging objects that are not in a custom class..
     * (basic objs or json being used)
     */
    moveUniquePropsToObj=function($obj_from, $obj_to) {
        if ($obj_from == null) return;
        if ($obj_to == null) return;
        var keys = Object.keys($obj_from);
        for (var k=0; k < keys.length; k++) {
            var prop = keys[k];
            if ($obj_to[prop]!=undefined) continue;
            var val = $obj_from[ prop ];
            $obj_to[prop]=val;
        }
    }


    printunixds=function($unixds) { 
        var d = new Date($unixds); //, timeZone:'UTC', timeZoneName:'short'
        const options = { year: 'numeric', month: '2-digit', day: '2-digit'};
        return d.toLocaleDateString('fr-CA', options)+" "+d.toLocaleTimeString();
    }

    //too tired to right now to implement the listsplitter.
    //this is for lazy sloppy jackass .csv files that include multiple
    //values in the column with a list.
    //we can split them into their own column and then give the user
    //the option to map it to an AKA.
    //for now, it gets nuked.
    processCSV_withHeader=async($csv, $delim, $listsplitter=";")=> {
        var allTextLines = $csv.split(/\r\n|\n/);
        var headers = [];
        var headerRow = allTextLines[0].split($delim);
        for (var h=0; h<headerRow.length; h++) {
            var clean = headerRow[h].replace(/['"]+/g, '');
            headers.push(clean.trim());
        }
        console.log(headers);
        var csvRowObjs = [];
        for (var n=1; n < allTextLines.length-1; n++) {
            var row = allTextLines[n].split($delim);
            var rowObj = [];
            for (var j=0; j<row.length; j++) {
                var clean2 = row[j].replace(/['"]+/g, '');
                var colName = headers[j];
                if (clean2.indexOf($listsplitter)!=-1) {
                    var aParts = clean2.split(";");
                    clean2 = aParts[0].trim();
                }
                if (headers!=null) rowObj[ colName ] = clean2;
                else rowObj.push(clean2.trim());
            }
            rowObj.__ROW_ACTIVE=true;
            csvRowObjs.push(rowObj);
        }
        return {'headers':headers, 'data': csvRowObjs};
    }

    processJsonKeyPairs=async($json, $delim)=> {
        var headers = [];
        var finaldata = [];
        for (var n=0; n < $json.length; n++) {
            var row = $json[n];
            //first loop snag all props from first row and make headers.
            if (n==0) {
                for (var prop in row) {
                    var header = prop;//.replace(/['"]+/g, '');
                    headers.push(header);
                }
            }
            var rowObj = [];
            //loop the headers, snag the objects
            for (var h=0; h<headers.length; h++) {
                var prop = headers[h];
                var value = row[prop];
                if (value=="null") value=="";
                else if (typeof value ==='string') value = value.replace(/['"]+/g, '');
                rowObj[ prop ] = value;
            }
            rowObj.__ROW_ACTIVE=true;
            finaldata.push(rowObj);
        }
        return {'headers':headers, 'data': finaldata};
    }
    
    pushUniqueProp=function($obj, $prop, $val) {
        if (Array.isArray($obj[$prop]) == true) { if ($obj[$prop].indexOf($val)==-1) $obj[$prop].push($val); }
        else {
            $obj[$prop]=[];
            $obj[$prop].push($val);
        }
    }

    fisherYatesShuffle=function($array) {
        for (let i = $array.length - 1; i > 0; i--) {
            const j = Math.floor(Math.random() * (i + 1));
            [$array[i], $array[j]] = [$array[j], $array[i]];
        }
        return $array;
    }

    range = function ($min,$max) {
        return Math.floor(Math.random()*($max-$min+1)+$min);
    }

    inclusive_range($min, $max) {
        return Math.floor(Math.random() * ($max - $min + 1)) + $min;
    }

    randTXID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+txid;
    }

    formatDeci = function($num, $benchmark) {
        var decis = this.decis($benchmark);
        return $num.toFixed(decis);
    }

    getRadioValue=function($radelid) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].checked) {
                return radios[i].value; 
            }
        }
        return null;
    }

    /**
     * 
     * @param {*} $base 
     * @param {*} $per - in the form of a whole number.  4 = 4%. 
     */
    addPercent=function($base, $per) {
        var decis =  this.decis($base);
        var base = $base;
        console.log(typeof base);
        if ((base instanceof Number) == false) base = new Number($base);
        base = base + (base * $per/100);
        return base.toFixed(decis+1);
    }

    setRadioValue=function($radelid, $value) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].value != $value) {
                radios[i].checked=false;
                continue;
            }
            radios[i].checked=true; 
            return;
        }
        return;
    }

    decis = function($num) {
        var numStr = $num;
        if (typeof $num == "Number") numStr = $num.toString();
        else numStr = $num+"";
        if (numStr.includes('.')) {
            return numStr.split('.')[1].length;
        }
        return 0;
    }


    /**
     * ghosts start with z.
     * @param {*} $len 
     * @returns 
     */
    randNID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+tx 
    }

    convertDMSLatToDec($degrees, $minutes, $seconds) {
        return $degrees + ($minutes / 60) + ($seconds / 3600);
    }
    convertDMSLongToDec($degrees, $minutes, $seconds, $direction) {
        var dd = $degrees + ($minutes / 60) + ($seconds / 3600);
        if ($direction === 'W') {
            return dd * -1;
        } 
        else if ($direction === 'E') {
            return dd;
        }
    }
    testGeo=function() {
        //giza great pyramid
        var lat = this.convertDMSLatToDD(40, 42, 31.0, "N"); // 40.70861°N
        var long = this.convertDMSLongToDD(74, 0, 32.0, "W"); // -74.00889°W
        console.log(lat+", "+long)
    }
    


    
    getRandHex = function($len) { return UTILS.randomHex($len);}
    randHex = function($len) { return UTILS.randomHex($len); }
    randomHex = function($len) {
        var maxlen = 8;
        var min = Math.pow(16,Math.min($len,maxlen)-1) 
        var max = Math.pow(16,Math.min($len,maxlen)) - 1;
        var n   = Math.floor( Math.random() * (max-min+1) ) + min;
        var r   = n.toString(16);
        while ( r.length < $len ) {
            r = r + this.randomHex( $len - maxlen );
        }
        return r;
    };

    randomSpherePoint=function($x0,$y0,$z0,$radius){
        var u = Math.random()+.01;
        var v = Math.random()+.01;
        var theta = 2 * Math.PI * u;
        var phi = Math.acos(2 * v - 1);
        var pt = new Object();
        pt.x = $x0 + ($radius * Math.sin(phi) * Math.cos(theta));
        pt.y = $y0 + ($radius * Math.sin(phi) * Math.sin(theta));
        pt.z = $z0 + ($radius * Math.cos(phi));
        return pt;
    }

    //nid or tnid array works here
    removeFromNidArray=function($arr, $nid) {
        var index = $arr.indexOf($nid)
        if (index!==-1)  $arr.splice(index,1);
    }

    removeFromTnidsArray=function($tnids_arr, $temp_nid) {
        var index = $tnids_arr.indexOf($temp_nid)
        if (index!==-1)  $tnids_arr.splice(index,1);
    }

    removeFromRelonsArray=function($relons_arr, $r) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.r==$r) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    removeRelonByTxid=function($relons_arr, $txid) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.txid==$txid) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    
    getHighestSimilarity = function($NOBJs, $lookupval) {
        var highestSim = 0;
        var highestNOBJ = null;
        for (var c=0; c <$NOBJs.length; c++) {
            var nObj = $NOBJs[c];
            var sim = ANNE.UTILS.similarity(nObj.text, $lookupval);
            if (sim>highestSim) {
                highestSim = sim;
                highestNOBJ = nObj;
            }
        }
        return {nObj:highestNOBJ, sim:highestSim};
    }


    //levenshtein distance, altered
    similarity = function(s1, s2, $factor=1) {
        var longer = s1;
        var shorter = s2;
        if (s1==undefined || s1==null) {
            alert("woh");
            return 0;
        }
        if (s2==undefined || s2==null) {
            alert("woh");
            return 0;
        }
        if (s1.length < s2.length) {
            longer = s2;
            shorter = s1;
        }
        var longerLength = longer.length;
        
        if (longerLength == 0) return 1.0; //s1 and s2 have no length? lolz.

        var lenFact = shorter.length/longer.length;
        if ($factor>1) {
            var final = Math.floor($factor * (longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength));
            return (final/$factor)*lenFact;
        }
        else return ((longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength))*$factor;
    }


    sortByNid=function($a, $b) {
        if ($a==undefined || $a==null || $b.nid==undefined || $b.nid==null) return 0;
        if ($a.nid < $b.nid) return -1;
        else if ($a.nid > $b.nid) return 1;
        return 0;
    }


    //targetting the earliest here..
    sortByBeTime=function($a, $b) {
        if ($a==undefined || $a.be_time==null || $b.be_time==undefined || $b.be_time==null) return 0;
        if ($a.be_time < $b.be_time) return 1;
        else if ($a.be_time > $b.be_time) return -1;
        return 0;
    }


    //sort clump neurons by their WHEN role.
    sortClumpsByWhen=function($events) {
        if ($events==undefined || $events==null || $events.length<=1) return;
        $events.sort( UTILS.sortByWhenRole );
        return;
    }
    


    //these are role_when arrays.  We will only use the first
    //in 99.99% of cases there will only be one of a given role per sentence.
    sortByWhenRole=function($a, $b, $bFWD=true) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.role_when==undefined || $a.role_when==null) {
            if ($b.role_when!=undefined && $b.role_when!=null) return 1; 
            else return 0;
        }
        else if ($b.role_when==undefined && $b.role_when==null) return -1; 

        var a_when = MEMG.getNeuronByTempNid($a.role_when[0].t);
        var b_when = MEMG.getNeuronByTempNid($b.role_when[0].t);
        
        if ($bFWD) return UTILS.compareDates(b_when, a_when);
        else return UTILS.compareDates(a_when, b_when);
    }

    tnidsUnion = function($a, $b) {
        var union = [];
        for (var a=0; a <$a.length; a++) {
            var atnid = $a[a];
            for (var b=0; b <$b.length; b++) {
                if ($b[b]==atnid) {
                    if (union.indexOf(atnid)==-1) {
                        union.push(atnid);
                        break;
                    }
                }
            }   
        }
        return union;
    }


    //we should have an ANNE epoch day number for all dates.
    //that would be a better (more efficient) date compare.

    compareDates = function($a, $b) {
        $a.processAsDateNeuron();
        $b.processAsDateNeuron();
        if ($a.year < $b.year) return -1;
        else if ($a.year > $b.year) return 1;
        else {
            if ($a.month < $b.month) return -1;
            else if ($a.month > $b.month) return 1;
            else {
                if ($a.day < $b.day) return -1;
                else if ($a.day > $b.day) return 1;
                else return 0;
            }
        }
        return 0;
    }

    isUSSuffix=function($test) {
        var suffix = ["JR", "SR", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X"];
        if ($test[0]==".") $test = $test.substring(1);
        $test = $test.toUpperCase();
        if (suffix.indexOf($test)!=-1) return true;
        return false;
    }

    removeOption = function($el, $val) {
        var el = document.getElementById($el);
        if (el==undefined || el==null) return;
        var options = document.querySelector("#"+$el).options;
        if (options==undefined || options==null) return;
        for (var o=0; o < options.length; o++) {
            if (options[o].value==$val) {
                options.remove(o);
                break;
            }
        }
    }


    roughSizeOfObject($object) {
        var objectList = [];
        var stack = [$object];
        var bytes = 0;
        while (stack.length) {
            var value = stack.pop();
            if (typeof value === 'boolean') bytes += 4;
            else if (typeof value === 'string') bytes += value.length * 2;
            else if (typeof value === 'number') bytes += 8;
            else if (typeof value === 'object' && objectList.indexOf( value ) === -1) {
                objectList.push( value );
                for( var i in value ) stack.push( value[ i ] );
            }
        }
        return bytes;
    }

    sleep=function($ms) {
        return new Promise(resolve => setTimeout(resolve, $ms));
    }

    //test this with DOM converstion...
    sortDD = function ($el) {
        var sel = docuement.getElementById($el);
        var selected = sel.value; // cache selected value, before reordering
        var opts_list = sel.querySelector('option');
        opts_list.sort(function (a, b) { return document.getElementById(a).text > document.getElementById(b).text ? 1 : -1; });
        sel.innerHTML="";
        sel.append(opts_list);  //append children back
        sel.value=selected; 
    }

    sortByFire=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire < $b.fire) return -1;
        if ($a.fire > $b.fire) return 1;
        return 0;
    }

    sortByFireThenIns=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire > $b.fire) return -1;
        if ($a.fire < $b.fire) return 1;
        if ($a.ins.length > $b.ins.length) return -1;
        if ($a.ins.length < $b.ins.length) return 1;
        return 0;
    }

    sortByDimension=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.type < $b.type) return -1;
        if ($a.type > $b.type) return 1;
        return 0;
    }

    sortByText=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text < $b.text) return -1;
        if ($a.text > $b.text) return 1;
        return 0;
    }

    sortTnidsByAlpha=function($tnids) {
        var nObjs = MEMG.getNeuronsFromTnids($tnids);
        nObjs.sort(this.sortByTextNoCase);
        return MEMG.getTnidList(nObjs);
    }

    sortByTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text==undefined || $b.text==undefined || $a.text==null || $b.text==null) {
            console.log("alsdfkjalsdfjlafjadkfj;alsdfj");
            return 0;
        }
        if ($a.text.toUpperCase() < $b.text.toUpperCase()) return -1;
        if ($a.text.toUpperCase() > $b.text.toUpperCase()) return 1;
        return 0;
    }


    sortRelonsByFromNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.f);
        if (aNOBJ==null || aNOBJ.text==undefined) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.f);
        if (bNOBJ==null || bNOBJ.text==undefined) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortRelonsByToNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.t);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.t);
        if (bNOBJ==null) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }


    sortGhostsLast=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.ghost==true) {
            if ($b.ghost==false) return 1;
            else return 0;
        }
        return 0;
    }

    sortByRelnTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        var aNOBJ = MEMG.getNeuronByTempNid($a.r);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.r);
        if (bNOBJ==null) return;

        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortByTime=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByTimeAsc=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return 1;
        if ($b.time && $a.emulated==false <=0) return -1;
        if ($a.time < $b.time) return -1;
        if ($a.time > $b.time) return 1;
        return 0;
    }
    sortByTimeASC=function($b,$a) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByFirstSeen=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.firstseen > $b.firstseen) return 1;
        if ($a.firstseen < $b.firstseen) return -1;
        return 0;
    }




    syntaxHighlight=function($json) {
        json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }

        

    /**
     * validate an ipv4 addressed
     * @param {*} $nObj 
     * @returns {Boolean} 
     */
     __ipv4Validation=function($nObj) {
        console.log("(__ipv4Validation) to do!");
        return true;
    }


    /**
     * validate an ipv6 address
     * @param {*} $nObj 
     * @returns {Boolean}
     */
    __ipv6Validation=function($nObj) {
        console.log("(__ipv6Validation) to do!");
        return true;
    }


    /**
     * Validate an email string conforms to known email strings.
     * 
     * @param {anne_neuron} - The neuron with .value OR string we want to test
     * @author {Mr.Scatman} 
     *
     */
    __emailValidation = function($nObj) {
        var mailformat = /(?!.*\.{2})^([a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+(\.[a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+)*|"((([ \t]*\r\n)?[ \t]+)?([\x01-\x08\x0b\x0c\x0e-\x1f\x7f\x21\x23-\x5b\x5d-\x7e\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|\\[\x01-\x09\x0b\x0c\x0d-\x7f\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))*(([ \t]*\r\n)?[ \t]+)?")@(([a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.)+([a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.?$/i
        
        //just a string (not an nObj);
        if (typeof $nObj === "string") {
            var test = mailformat.test($nObj.toLowerCase());
            //console.log(test);
            if(test==undefined || test==null) return "invalid format";
            else if (test==true) return true;
            else return false;
        }

        if ($nObj.value==undefined || $nObj.value==null || $nObj.value=="") {
            $nObj.value = $nObj.text.trim();    //default nObj .value is the .text if none was provided by a .set_value function.
        }
        if(mailformat.test($nObj.value.toLowerCase())==true) {
            $nObj.valid = true;
            return true;
        }
        else {
            $nObj.value = "";
            $nObj.valid = false;
            return false;
        }
    }

    //from devshed
    __urlValidation=function($str) {
        var pattern = new RegExp('^(https?:\\/\\/)?'+ // protocol
        '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|'+ // domain name
        '((\\d{1,3}\\.){3}\\d{1,3}))'+ // OR ip (v4) address
        '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*'+ // port and path
        '(\\?[;&a-z\\d%_.~+=-]*)?'+ // query string
        '(\\#[-a-z\\d_]*)?$','i'); // fragment locator
      return !!pattern.test($str);
    }

    
    /**
     * 
     * @param {*} $aMatches 
     * @param {*} $aMatches1 
     * @param {*} $prop 
     * @returns 
     */
    concatUnique=function($aMatches, $aMatches1, $prop="nid") {
        if($aMatches.length==0) return $aMatches1;
        if($aMatches1.length==0) return $aMatches;
        var arr = new arraymap();
        arr.setMapKey($prop);
        arr.addArray($aMatches);
        arr.addArray($aMatches1);
        return arr.arr;
    }


    uniqueConcat=function($a, $b, $prop="nid") {
        if ($a.length==0) return $b.concat();
        var a2 = $a.concat();
        for (var b=0; b < $b.length; b++) { 
            if ($a.indexOf($b[b])==-1) a2.push($b[b]);
        }
        return a2;
    }

    uniqueConcatRels = function($a, $b) {
        var a_rels_txids = [];
        var rels = [];
        for (var a=0; a < $a.length; a++) {
            a_rels_txids.push($a[a].txid);
            rels.push($a[a]);
        }
        for (var b=0; b < $b.length; b++) {
            if (a_rels_txids.indexOf($b[b].txid)!=-1) continue;
            a_rels_txids.push($b[b].txid);
            rels.push($b[b]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    poorMansAbbreviate=function($str) {
        $str = $str.trim();
        const words = $str.split(' ');
        
        if (words.length === 1) {
            return $str; 
        } 
        else {
            var abbr = '';    
            words.forEach((word, index) => {
                if (index === 0 || index === words.length - 1) {
                    abbr += word.charAt(0).toUpperCase() + '.';
                } 
                else {
                    abbr += word.charAt(0).toUpperCase();
                }
            });
            return abbr;
        }
    }

    
    /**
     * Check for a relon in a relon array.
     * If it is unique, add it to the array and return true.
     * if it is not unique, return false.
     * @param {array} $rels 
     * @param {Object} $rel 
     * @returns 
     */
    pushIfUniqueRel = function($rels, $rel) {
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if(rel.txid==$rel.txid) {
                return false;
            }
        }
        $rels.push($rel);
        return true;
    }

    seededRandom=function() {
        var x = Math.sin(Math.random()*1000) * 10000;
        return x - Math.floor(x);
    }

    getScatModSeedArrItem=function($seed, $arr) {
        var per = UTILS.getScatModSeedPer($seed);
        var max_usable = Math.round($arr.length * per);
        return $arr [ Math.floor(Math.random()*max_usable)];
    }
    

    getScatModSeedPer=function($seed) {
        $seed = Math.round($seed);
        var per = .5;
        if ($seed % 3 !=0) per +.1;
        if ($seed % 2 !=0) per +.15;
        if ($seed % 7 ==0) per +.25;
        return per;
    }
    
    isWholeNumber=function($number) {
        return $number % 1 === 0;
    }

    HEXtoRGB = function($HEX) {
        var isNumb = Number.isInteger($HEX);
        if (isNumb) var bigint = $HEX;
        else var bigint = parseInt($HEX.slice(1), 16);
        const r = (bigint >> 16) & 255;
        const g = (bigint >> 8) & 255;
        const b = bigint & 255;
        return { r, g, b };
    }

    RGBtoHEX = function($rgb) {
        const { r, g, b } = $rgb;
        const hex = `0x${((1 << 24) + (r << 16) + (g << 8) + b).toString(16).slice(1)}`;
        return hex;
    };

    shiftRGB = function($RGB, $incr) {
        const shiftedColor = {
          r: $RGB.r + $incr,
          g: $RGB.g + $incr,
          b: $RGB.b + $incr,
        };
        shiftedColor.r = Math.max(0, Math.min(255, shiftedColor.r));
        shiftedColor.g = Math.max(0, Math.min(255, shiftedColor.g));
        shiftedColor.b = Math.max(0, Math.min(255, shiftedColor.b));
        return shiftedColor;
    };
}


const UTILS = new anne_utils ();
UTILS.set_init(UTILS);

/************** (from file ./js/base/utils.js)**/ 

const LINGO_TYPE_ANNE = 0;
const LINGO_TYPE_GRAPH = 1;
const LINGO_TYPE_COMMON = 2;

const TENSE_PRESENT = 1;
const TENSE_PAST = 2;
const TENSE_PAST_COMPLETE = 3;


//This is not semantic langauge via clumps.  This is just a basic output
//class for debugging or quick english display of things.
class anne_talk extends anne_singleton {
	constructor(){
		super();
		
		//THIS IS A SIMPLE WAY TO PROGRAM GRAMMAR utilizing base roles as storage.
		//different languages can have different grammar patterns.
		//one could spend a very long time in here building a robust langauge handler..
		//it can be done for AI/AGI use cases.
		
		//WE CAN STORE THE ROLES PROPERLY.
		//WE CAN VERBALIZE THEM.
		//In time, (with a fuller vocab) we can fully parse and conceptualize into the ANNE semantics.

		//ANNE training is done by experience/use..  The very basics are put in at inception to have something 
		//to work with.  The idea is parts-of-speachless (we use a higher concept of part of anne)
		//Language is a map into a higher-order brain protocol.
		//The Language may change, and with it the ways to communicate, but the concepts are stored using the 
		//higher-order unified brain protocol.  The same brain can speak 4 different languages and doesn't rqeuire
		//4 unique concepts repeating what a "dog" is defined as.  There are 4 differeent lang word maps to the 
		//same concepts (if it is indeed the same concept) 

		//clump definition pattern.
		this.def = new Object();
		this.def.infini = {pat_part_name:"infini", roles:[{role:"role_action", p:"to"}]};	//no use ot do doInfinitive w/ rule as it is just a "to"+
		this.def.mode = {pat_part_name:"mode", roles:["role_mode"]};
		this.def.obj = {pat_part_name:"obj", roles:[{role:'role_affected_thing', mod:'role_affected_thing_mod', det_func:this.___doThingDet}, 
													{role:'role_resulting_new_thing', mod:'role_resulting_new_thing_mod', det_func:this.___doThingDet},
													{role:'role_theme_thing', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_instrument_mod',
													{role:'role_instrument', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_state', 'role_feelz', 'role_thought', 'role_fact']};

		this.def.manner = {pat_part_name:"manner", roles:[{role:"role_manner", modfunc:this.___doMannerEngShifter}]}
													
		this.def.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"to"}, 
				{role:'role_when', mod:'role_when_mod'},
				'role_start','role_end']
		};

		this.eng_pat_def = [
			this.def.infini,
			this.def.mode,
			this.def.obj,
			this.def.manner,
			this.def.wherewhen2
		]


		//action inst pattern
		this.actionInst = new Object();
		this.actionInst.subj = {pat_part_name:"subj", roles:['role_actor']};
		this.actionInst.mode = {pat_part_name:"mode", roles:['role_manner', 'role_mode']};
		this.actionInst.pred = {pat_part_name:"pred", roles:['role_action'], mod:'role_tense', modfunc:this.englishTensify}; //'role_reason'
		this.actionInst.obj = {pat_part_name:"obj", roles:['role_affected_thing', 'role_resulting_new_thing', 
										'role_state', 'role_feelz', 'role_thought', 'role_fact',
										'role_instrument_mod','role_instrument']};

		this.actionInst.theme_thing = {pat_part_name:"theme_thing", roles:['role_theme_thing_mod', 'role_theme_thing']};
		
		//to use role_source/path/goal in the same clump you will likely want 
		//more than one role_where_mod.  if so, tuff beans.  Create a nested neuron.
		this.actionInst.wherewhen = {
				pat_part_name:"wherewhen",
				roles:['role_source', 'role_path', 'role_goal',
				'role_where_mod', 'role_where', 'role_when_mod', 'role_when', 
				'role_start', 'role_end']
			};
		this.actionInst.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"at", det_func:this.___doWhereDet}, 
				{role:'role_when', mod:'role_when_mod', p:"on"}
				   ,'role_start', 'role_end']
		}
		this.actionInst.pre_connectors = {
			pat_part_name:"pre_connectors",
			recurse:true,
			roles:['role_cause', 'role_cdx', 'role_prev']};
		this.actionInst.post_connectors = {
			recurse:true,
			pat_part_name:"post_connectors",
			roles:['role_effect', 'role_result', 'role_next']};

		//this.actionInst.tense,
		this.eng_pat1 = [
			this.actionInst.subj, 
			this.actionInst.mode,
			this.actionInst.pred, 
			this.actionInst.theme_thing, 
			this.actionInst.obj, 
			this.actionInst.wherewhen2,
			this.actionInst.pre_connectors,
			this.actionInst.post_connectors
		];


		this.type_pre = ["sys", "be", "param", "has", "awareness", "exper", "clump", "sys msg", "brain", "role", "episodic"];

		this.vowel=["a", "e", "i", "o", "u"];  //and sometimes y? lolz.
		
		this.init();
	}


	init = function() {
		this.bgbigred = this.redbig;
		this.bgblack = 'background:black; color:white';
		this.bgblackbold = 'background:black; color:white; font-weight:bold';
		this.bgblackyellow = 'background:#171717; color: #fefe17';
		this.bgblue = 'background:blue; color: white';
		this.bgbluebig = 'background:blue; color: white; font-size:1.1rem';
		this.bgerror = 'background: red; color: white; font-weight:bold; font-size:1.1rem;';
		this.bgfire = 'background:#F7C24A; color: #171717';
		this.bggray = 'background:#ccc; color: #171717';
		this.bggreen = 'background: green; color: white';
		this.bggreenbig = 'background: green; color: white; font-size:1.1rem;';
		this.bgivoryblue = 'background:ivory; color: blue';
		this.bglightblue = 'background:lightblue; color: #171717';
		this.bglightbluebold = 'background:lightblue; color: #171717; font-weight:bold';
		this.bglightgreen = 'background:lightgreen; color: #171717';
		this.bglightred = 'background:lightgreen; color: #171717';
		this.bgneg = this.bgred;
		this.bgneu = this.bgyellow;
		this.bgorange = 'background: orange; color: #171717';
		this.bgpink = 'background:pink; color: #171717';
		this.bgcyan = 'background:cyan; color: #171717';
		this.bgpos = 'background: #00ff00; color: #171717';
		this.bgpurple = 'background:purple; color: white';
		this.bgred = 'background: red; color: white';
		this.bgredbig = 'background: red; color: white; font-size:1.1rem;';
		this.bgyellow = 'background:#fefe17; color: black';
		
		this.bgheading_yellow = 'background:black; color: #fefe17; font-size:1.1rem';
		this.bgheading_black = 'background:whitesmoke; color: black; font-size:1.1rem';
		this.bgheading = 'background:ivory; color: #171717; font-size:1.4; font-weight:bold';

		this.fromMsg = 'background:ivory; color: #171717; font-weight:bold';
		this.funct = 'background:pink; color: #171717; font-style:italic';
		this.ghosts = 'background:white; color: blue';
		this.msg = 'background: green; color: white';
		this.msgNote = 'background:ivory; color:green; font-style:italic';
		this.pooltitle = 'background:ivory; color: black; font-weigt:bold; font-size:1.2rem;';
		this.title = this.pooltitle;
		
		this.bold_black = 'background:ivory; color: #171717; font-weight:bold';
		this.black = 'background:ivory; color: #171717';
		this.green = 'background: ivory; color: green';
		this.orange = 'background: ivory; color: orange;';
		this.redbig = 'background: red; color: white';
		this.yellow = 'background: ivory; color: yellow';
		
		this.part_cnt = this.bgpink;
		this.part_depth = this.bgyellow;
		this.part_id = this.bglightblue;
		this.part_full = this.bggreen;
		this.part_hex = this.bglightblue;
		this.part_mod = this.bglightred;
		this.part_mod_ph = this.bgorange;
		this.part_part = this.bglightgreen;
		this.part_sig = this.bglightblue;
		
		this.action = 'background:red;color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.aux = 'background:salmon; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.connector = 'background:purple; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.manner = 'background:indianred; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.mode = 'background:firebrick; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
			
		this.tab0 ='';
		this.tab1 ='██';
		this.tab2 ='████';
		this.tab3 ='██████';
		this.tab4 ='████████';
		this.tab5 ='██████████';
		this.tab6 ='████████████';
		this.tab7 ='██████████████';
		this.tab8 ='████████████████';
		this.tab9 ='██████████████████';
		this.tab10='████████████████████';


		this.BLACK = 30;

		this.RED = 31;
		this.TRADE_RED = 311;
		
		this.GREEN = 32;
		this.TRADE_GREEN = 321;
		
		this.YELLOW = 33;
		this.YELLOW_i = 330;
		this.TRADE_YELLOW = 331;
		
		this.BLUE = 34;
		this.BLUE_i = 340;
		
		this.PINK = 35;
		this.CYAN = 36;
		this.WHITE = 37;

	}


	printf=function(...$args) {
		var lastColor = TALK.WHITE;
		var html = "";
		$args.forEach($arg => {
			if(typeof $arg==="string") html+="<span class='copr-"+lastColor+"'>"+$arg+"&nbsp;</span>";
			else {
				lastColor = parseInt($arg);
				if (lastColor==undefined) lastColor = TALK.WHITE;	
			}
		});
		return html;
	}

	/**
	 * 
	 * @param {*} $int 
	 * @returns 
	 */
	console_color_map = function($int) {
		$int = parseInt($int);
		if ($int==ANNE.TALK.BLACK) return this.bgblack;
		else if ($int==ANNE.TALK.RED) return this.bgred;
		else if ($int==ANNE.TALK.GREEN) return this.bggreen;
		else if ($int==ANNE.TALK.YELLOW) return this.bgyellow;
		else if ($int==ANNE.TALK.BLUE) return this.bgblue;
		else if ($int==ANNE.TALK.PINK) return this.bgpink;
		else if ($int==ANNE.TALK.CYAN) return this.bgcyan;
		else if ($int==ANNE.TALK.WHITE) return this.bgwhite;
		return this.bgblack;
	}
	
	printf_console=function(...$args) {
		var lastColor = this.bgblack;
		var strs = [];
		var colors = [];
		$args.forEach($arg => {
			if(typeof $arg==="string") {
				strs.push("%c"+$arg);
				colors.push(lastColor);
			}
			else {
				lastColor = this.console_color_map(parseInt($arg));
				if (lastColor==undefined) lastColor = this.bgwhite;	
			}
		});
		console.log(strs.join(""),...colors);
	}



	/**
	 * 
	 * @param {*} $vb 
	 * @param {*} $tense 
	 * @returns 
	 */
	 englishTensify = function($vb, $tense) {
		if ($tense=="past tense") {
			var rev = $vb.split("").reverse().join("");
			var c0 = rev[0];
			var c1 = "";
			if (rev.length>1) c1 = rev[1];
			if (c0=="e") return $vb+"d";
			if (c0=="y" && $text.length>1) return $vb.substring(0, $vb.length-1)+"ied";
			if (this.vowel.indexOf(c1)!=-1 && this.vowel.indexOf(c0)==-1) return $vb+c0+"ed";
			if (c0=="c") return $vb+"ked"; 
			return $vb+"ed";
		}
		else return $vb;
	}


	/**
	 * For a given role, what is the EC type we expect to fill this role?
	 * @param {String} $role_nid - nid long string 
	 * @returns 
	 */
	__getTopMatchRoleFromRole=function($role_nid) {
        if ($role_nid==MEM.role_action_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_theme_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_action.temp_nid) return MEM.ec_action.temp_nid;
        else if ($role_nid==MEM.role_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_actor.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_affected_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_posn_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_be_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_state.temp_nid) return MEM.ec_state.temp_nid;
        else if ($role_nid==MEM.role_when.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_where.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_manner.temp_nid) return MEM.ec_manner.temp_nid;
        else if ($role_nid==MEM.role_mode.temp_nid) return MEM.ec_mode.temp_nid;
        else if ($role_nid==MEM.role_connector.temp_nid) return MEM.ec_expect_more_of_prev.temp_nid;
        else if ($role_nid==MEM.role_aux_action.temp_nid) return MEM.ec_expect_action.temp_nid;
        else if ($role_nid==MEM.role_thing_mod.temp_nid) return EM.ec_expect_thing_mod.temp_nid;
        else if ($role_nid==MEM.role_question.temp_nid) return MEM.role_question.temp_nid;
        else {
            var nOBJ = MEMG.getNeuronByTempNid($temp_nid);
            console.log("%c unhandled (__getTopMatchRoleFromRole) %c "+nOBJ.text, TALK.bgred, TALK.green);
        }
    }

	/**
	 * 
	 * @param {*} $balanceObj 
	 */
	printBalanceObj=function($balanceObj) {
		console.log($balanceObj);
		console.log("%c annecoin "+ANNE.UTILS.commas($balanceObj.annecoin), TALK.bgblackyellow);
		console.log("%c annies "+ANNE.UTILS.commas($balanceObj.annies), TALK.bglightblue);

		if ($balanceObj.avail_now_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_now_annies/100), TALK.bglightblue);
		}
		if ($balanceObj.avail_soon_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_soon_annies/100), TALK.bglightblue);
		}
	}

	/**
	 * Shift an english word from the stored EC type to the desired EC type.
	 * 
	 * @param {anne_neuron} $roleNOBJ - the role we are shifting from. 
	 * @param {*} $valueNOBJ - the value of the role we are shiting from.
	 * @returns 
	 */
	__engShifter = function($roleNOBJ, $valueNOBJ) {
		console.log("(TALK.__engShifter)", $roleNOBJ, $valueNOBJ);

		//the top EC type of the role we are shifting from.
		var ec_top_desired_tnid = ANNE.TALK.__getTopMatchRoleFromRole($roleNOBJ.temp_nid);
		if (ec_top_desired_tnid==null) return;

		//the ec type we want to shift to.
		var ec_top_observed_tnid = $valueNOBJ.setECConceptualizingParent();
			
		//if the type is already right, return the text of the from.  No shift is needed.
		if (ec_top_desired_tnid==ec_top_observed_tnid) return $valueNOBJ.text;
		
		//we need to find the rule that shifts this concept from ec_top_observed_nid to the ec_top_desired_nid!
		for (var g=0; g < MEM.eng_grammar_rule.children.length; g++) {
			var ruleNOBJ = MEMG.getNeuronByTempNid( MEM.eng_grammar_rule.children[g]);
			var role_src = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_source.temp_nid);
			if (role_src==null) continue;
			if (role_src.t!=ec_top_observed_tnid) continue;	

			//we have a rule that starts from where our concept starts.
			//now we need to see if it converts to our desired
			var role_goal = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_goal.temp_nid);
			if (role_goal==null) continue;
			if (role_goal.t!=ec_top_desired_tnid) continue;

			//we found a rule!
			//we need to now look for the suffixes that use this rule.
			for (var i=0; i < ruleNOBJ.ins.length; i++) {
				var in_rel = ruleNOBJ.ins[i];
				if (in_rel.type!=TYPE_HAS) continue;
				if (in_rel.r!=MEM.eng_grammar_rule.temp_nid) continue;

				//we found a suffix that uses this rule.
				var suffixNOBJ = MEMG.getNeuronByTempNid(in_rel.f);
				if (suffixNOBJ==null) continue;
				return TALK.__applyOutEngSuffix(suffixNOBJ, $valueNOBJ);
			}
		}

		console.log("%c could not find rule.");
	}


	/**
	 * 
	 * @param {anne_neuron} $suffixNOBJ 
	 * @param {anne_neuron} $valueNOBJ 
	 * @returns 
	 */
	__applyOutEngSuffix=function($suffixNOBJ, $valueNOBJ) {
		console.log("(TALK.__applyOutEngSuffix)");
		var suffixOBJ  = ANNE.BRAIN.NLP.LC._suffixMap.getByText($suffixNOBJ.text);
		if (suffixOBJ==null || suffixOBJ.cdx_rule_out==undefined) {
			//console.log("%c NO KNOWN OUTBOUND ENGLISH RULE FOR SUFFIX "+$suffixNOBJ.text, TALK.bgredbig);
			return $valueNOBJ.text+$suffixNOBJ.text;
		}
		return $suffixNOBJ.cdx_rule_out($valueNOBJ.text);
	}



	/**
	 * Take a manner role or roles and convert them to be proper adverbs if they are not adverbs.
	 * @param {Array} $manner_roles 
	 * @returns 
	 */
	___doMannerEngShifter = function($manner_roles) {
		console.log("(TALK.___doMannerEngShifter)", $manner_roles);
		var manner_text = [];
		for (var w=0; w < $manner_roles.length; w++) {
			var rel_manner = $manner_roles[w];
			var toNOBJ = MEMG.getNeuronByTempNid(rel_manner.t);
			var manner = ANNE.TALK.__engShifter(MEM.role_manner, toNOBJ);
			manner_text.push(manner);
		}
		return TALK.listifyStrArr(manner_text);
	}

	___doThingDet = function($thing_roles) {
		console.log("(TALK.___doWhereDet)", $thing_roles);
		var thing_text = [];
		for (var w=0; w < $thing_roles.length; w++) {
			var rel_where = $thing_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				thing_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					thing_text.push("the "+toNObj.text);
				}
				else thing_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(thing_text);
	}

	___doWhereDet = function($where_roles) {
		//console.log("(TALK.___doWhereDet)", $where_roles);
		var where_text = [];
		for (var w=0; w < $where_roles.length; w++) {
			var rel_where = $where_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				where_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					where_text.push("the "+toNObj.text);
				}
				else where_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(where_text);
	}

	/**
	 * This print out the meaning (infinitive form definition) of an action neuron.
	 * Whatever ROLES are stored in the clump will define the clump.
	 * 
	 * We can also show connotations via TYPE_BRAIN and assocs but it will be 
	 * better to do that in the dossier file for this concept where we can display them
	 * more cleanly.
	 * 
	 * @param {anne_neuron} $clumpNOBJ 
	 * @param {Object} $gpat 
	 * @param {Boolean} $bDEBUG 
	 * @returns 
	 */
	verbalizeClassActionDef = function($clumpNOBJ, $gpat=null, $bDEBUG=true, $tenseOverride="", $lvl=0) {
		if  ($gpat==null) $gpat = this.eng_pat_def;
		if ($clumpNOBJ==null) return;
		var output = "";
		var parts = [];
		var order = [];

        if ($bDEBUG==true) {
			console.log("%c (verbalizeClassActionDef) lvl:"+$lvl+" for clump"+$clumpNOBJ.text, TALK.bggreenbig);
		}

		var part_of_anne = null;
		var role_pre_var = "";

		//if this is a class and not present tense, rewire to verbalize the parent action instead.
		if ($clumpNOBJ.ntype==NTYPE_CLASS && $clumpNOBJ.tense!="present tense" && $tenseOverride=="") {
			role_pre_var = "_meaning_";
		}
		
		console.log("role_pre_var: "+role_pre_var);

		//will have to fix roles here for EC TYPE shifts.. hmm.
		var doModNest=function($mods, $lvl=0, $det_func=null, $bDETFIRSTONLY=false) {
			console.log("(doModNest) lvl"+$lvl);
			var modTxt = "";
			var keys = Object.keys($mods)
			for (var k=0; k < keys.length; k++) {
				var rObj = $mods[ keys[k] ];
				if (rObj.mods!=undefined) {
					modTxt += doModNest(rObj.mods, $lvl+1, $det_func, true)+" ";
					$det_func = null;
				}
				if (rObj.nameless!=undefined) continue;
				var this_text =  "";
				if (rObj.r!=MEM.role_connector.temp_nid && $det_func!=null) {
					if ($bDETFIRSTONLY==true && k==0) this_text = $det_func([rObj]);
					else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				} 
				else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				modTxt+=this_text+" ";
			}
			return modTxt;
		}

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var pre_default = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre = "";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				var arr = $clumpNOBJ[role_pre_var+role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null) {
					mod_arr = $clumpNOBJ[role_pre_var+mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";
				order = order.concat(arr);


				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {

						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}

					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				//mod mods the WHOLE phrase if there are multiples.
				else if (role=="role_manner" && modfunc!=null) {
					to = modfunc(arr);
				}

				//{role:'role_where', mod:'role_where_mod', p:"to"},
				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}


				//new method of mods..
				else if (arr[0].mods!=undefined) {

					console.log("mods in role "+role, arr[0].mods);
					mod_arr = arr[0].mods;
					var modTxt = doModNest(mod_arr, 0, det_func);

					//the base role that has these mods (in eng pat we put the mod BEFORE)

					//note, this should not be an array, only one with our mods method.
					//will pinch it off here -- (will write a single verbalizeRole in a bit)
					to = modTxt+TALK.verbalizeRolesArr([ arr[0] ]);	

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						//var modNOBJ = MEMG.getNeuronByTempNid(mod_arr[0].t); //should also be role_text already..
						//console.log(modNOBJ);
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else {
					if (det_func!=null) to = det_func(arr);
					else to = TALK.verbalizeRolesArr(arr);
				}
				

				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		console.log(output);
		$clumpNOBJ.eval = output;
		if ($lvl!=0) return output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		var tense = $clumpNOBJ.tense;
		if ($tenseOverride!="") tense = $tenseOverride;

		return new anne_babytalk(output, order, part_of_anne, $clumpNOBJ.tense);
	}

	

    /**
     * Rudimentary verbalizer of _roles via a gpat (for lang or maodality specific orderings)
	 *
	 * @param {anne_neuron} $clumpNOBJ - the clump we want to print 
	 * @param {Object} $gpat - the language pattern we want to use
	 * @param {Array} $skipRoleNames - any roles we want to skip
     * @param {Boolean} $bDOSENTENCE - print in sentence mode
	 * @param {Boolean} $bDEBUG - degug the printing.
	 *
	 * @returns {anne_babytalk} - returns .output of sentence and .order of role tnids in the order they were printed from the gpat.
	 */
	 verbalizeInstActionClump=function($clumpNOBJ, $gpat=null, $skipRoleNames=[], $bDOSENTENCE=true, $bDEBUG=true){
        if ($gpat==null) $gpat = this.eng_pat1;
        var output = "";
		var parts = [];
		var order = [];
        if ($bDEBUG==true) console.log("%c print for clump"+$clumpNOBJ.text+" %cdosentence:"+$bDOSENTENCE, TALK.bggreenbig, TALK.bgblackyellow);
			
		var part_of_anne = null;	//EC type

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				var pre_default = "";
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre="";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				if ($skipRoleNames.length && $skipRoleNames.indexOf(role)!=-1) continue;

				var arr = $clumpNOBJ[role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null && mod!="") {
					mod_arr = $clumpNOBJ[mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";

				order = order.concat(arr);

				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {
						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}
					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);
				}

				else if (det_func!=null) {
					to = det_func(arr);
				}
		 
				else {
					to = TALK.verbalizeRolesArr(arr);
				}
				
				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		//make a sentence out of it?
		if ($bDOSENTENCE==true) {
			output = output.charAt(0).toUpperCase() + output.slice(1);
		}	
		console.log(output);
		$clumpNOBJ.eval = output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		return new anne_babytalk(output, order, part_of_anne);
    }


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	verbalizeRolesArr = function($arr) {
		if ($arr==undefined || $arr==null || $arr.length==0) return "[ERROR]";
		if ($arr.length==1) {
			var rObj = $arr[0];
			if (rObj.nameless!=undefined) return "";
			if (rObj.role_text!=undefined) return rObj.role_text;
			rObj.role_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			return rObj.role_text;
		}
		var str = "";
		var froms = [];
		var conj = "and";

		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			if (froms.length==0) froms.push(rObj.f);
			else if (froms.indexOf(rObj.f)!=-1) froms.push(rObj.f);
			else conj="or";
		}
		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			var role_text =  MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			str += role_text;
			rObj.role_text = role_text;
			if (r==$arr.length-2) str=str+" "+conj+" ";
			else if (r==$arr.length-1) ;
			else str=str+", ";
		}
		return str;
	}


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	listifyStrArr = function($arr) {
		if (ANNE.UTILS.isset($arr)==false) return "";
		if ($arr.length==1) return $arr[0];
		if ($arr.length==2) return $arr[0]+" and "+$arr[1];
		var finalstr = $arr[0];
		for (var a=1; a <$arr.length; a++) {
			if (a==$arr.length-1) finalstr = finalstr+", "+$arr[a];
			else finalstr = finalstr+" and "+$arr[a];
		}
		return finalstr;
	}


	
	/** 
	 * print simple relon tags.
	 * 
	 * This TALK class is an utter disaster.. changed a gazillion times.
	 * Use this to do basic printing.  We have lots of rework to do just
	 * for this basic talk.  Would be better to get full PAS understanding.
	 */ 
	 simpleRelonTags = function ($rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}
		if ($rObj.type==TYPE_AWARENESS) {
			var tt = TARGET_NID;
			if ($rObj.to_type==TO_TYPE_TXID) tt = TARGET_TXID;
			return this.verbalizeForAwareness($rObj, true);
		}

		var typeText = this.getTypeName($rObj.type, true, "rel-type");

		//willy wonka from name
		var fromNOBJ = null;
		var fromText = "";
		if ($rObj.from_nid!=null) {
			fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		}
		else if ($rObj.pending_from!=null) {
			fromNOBJ = MEMG.getNeuronByPendingNid($rObj.pending_from);
			if ($rObj.pending_from_debug!=undefined && $rObj.pending_from_debug.length>0) fromText = $rObj.pending_from_debug;//+"("+$rObj.pending_from+")";
			else fromText = $rObj.pending_from;
		}
		if (fromNOBJ==null) {
			console.log("NO FROM OBJ??");
			console.log($rObj);
			$rObj.error = true;
			return "[[error, bad relon]]";
		}
		else {
			if (fromNOBJ.emulated) {
				console.log("NO RELONS FROM EMULATED NEURONS -- they are not real!");
				return;
			}
			fromText = fromNOBJ.bestNameHTML();
		}
		var fromClass = this.getNeuronDisplayClass(fromNOBJ);

		//we want the name of the here even tho the from is actually
		//our nid (in the cmd relon)
		if ($rObj.debug_name!=undefined) {
			fromText = "(MAKE) "+$rObj.debug_name;
			fromClass = "neuron-ghost";
		}

		//reln text
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		if (relnNOBJ==null) {
		    console.log("problem with the reln. It must exist in local mem");
			$rObj.error = true;
		    return "[[ERROR in RELN "+$rObj.reln+"]]"
		}
		var relonText = relnNOBJ.text;
		var relnClass = "";
		if (relnNOBJ.nid==MEM.hanai_in.nid || relnNOBJ.nid==MEM.hanai_in_all.nid) relnClass = "neuron-hanai";
		else if (relnNOBJ.nid==MEM.hanai_out.nid)  relnClass = "neuron-hanai";
		else relnClass = this.getNeuronDisplayClass(relnNOBJ);
		

		if ($showfromNid==true) {
			fromText+="("+fromNOBJ.nid+")";
		}

		//to type
		var toText = this.verbalizeRelonToByToType($rObj, $bCOLORIZEBYTYPE);
		var ttypemsg = "";
		if ($ttype!=-1) ttypemsg = TXB.printType($ttype);
		if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" "+relonText+"&nbsp;<b>"+toText+"</b></span>";
		}
		else if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" MAKE LOCATION <b>&nbsp;"+toText+"</b>&nbsp;in&nbsp;"+fromText+"</span>";
		}

		//to stuff
		var toClass = "rel-to";
		var toNOBJ = null;
		if ($rObj.to_type==TO_TYPE_ANNE) {
			if ($rObj.pending_to_s!=undefined) {
				var parts = $rObj.pending_to_s.split("_");
				if (parts.length>1) {
					toNOBJ = MEMG.getNeuronByNid(parts[1].trim());
					if (toNOBJ!=null) {
						toClass = this.getNeuronDisplayClass(toNOBJ);
					}
				}
			}
			else { 
				toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
				if (toNOBJ!=null) {
					toClass = this.getNeuronDisplayClass(toNOBJ);
				}
			}
		}

		return "<span class='q-part "+fromClass+"' title='"+fromNOBJ.nid+"'>"+fromText+"</span>"+typeText+"<span class='q-part "+relnClass+"'>"+relonText+"</span><span class='q-part "+toClass+"'>"+toText+"</span><span class='q-part anne-tx-type'>*"+ttypemsg+"TX*</span>";
	}

	/**
	 * Helpful display to colorize things consistently thruout the app.
	 * This took a bit of time to arrive at -- for some time I tried to colorize
	 * class neurons red, instances blue, text green...
	 * but have now settled on ghost blue, anne yellow, text green, self red (maybe orange later), and hanai pink.
	 */
	getNeuronDisplayClass=function($nObj) {
		if ($nObj==null) {
			return "neuron-ghost";
		}
		
		if ($nObj.nid==MEM.USER._userNObj.nid) return "neuron-self";
		if ($nObj==undefined || $nObj==null) {
			return "neuron-ghost";
		}
		if ($nObj.ghost==true) return "neuron-ghost";
		else if (ANNE.isANNE($nObj.owned_by)==true) return "neuron-anne";
		else {
			if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
				return "neuron-self";
			}
			else {
				return "neuron-other"  //black for someone else's neuron (not ANNE's).
			}
		}
	}



	/**
	 * 
	 * @param {*} $realTXID 
	 * @param {*} $relon 
	 * @param {*} $sep 
	 * @param {*} $bCOLORIZE 
	 * @returns 
	 */
	genShortRelonTextByTXID=function($realTXID, $relon=null, $sep="&nbsp;»&nbsp;", $bCOLORIZE=false) {
		var relon = $relon;
		if ($relon==null) relon = MEMG.lookupRelonByTXID($realTXID);
		if (relon==null) return "";
		var fNObj = MEMG.getNeuronByNid(relon.from_nid);
		var rNObj = MEMG.getNeuronByNid(relon.reln);
		var tNOBJ = null;
		if (relon.t) tNOBJ = MEMG.getNeuronByNid(relon.to_s);
		var to = "";
		if (tNOBJ!=null) to = tNOBJ.bestNameHTML();
		else if ($relon.to_type==TO_TYPE_TXID) to="{url}";
		else if ($relon.to_type==TO_TYPE_BLOB) to="{blob}";
		else to = relon.to_s;
		var from = fNObj.text;
		if (from=="") from = fNObj.nid;
		if ($bCOLORIZE==true) {
			return "<b>"+from+"</b>"+$sep+"<em class='text-red'>"+rNObj.text+"</em>"+$sep+"<b class='text-green'>"+to+"</b>";
		} 
		return "<b>"+from+"</b>"+$sep+"<em>"+rNObj.text+"</em>"+$sep+"<b>"+to+"</b>";
	}


	/**
	 * Uses the Q to print a friendly gui.
	 */
	printSanitizedAlertCmds=function($updateElID, $aQ, $bPRINTBALANCE=true) {
		var msgDiv = ANNE.A.setHTML($updateElID, ""); //qrb_txs_list_div
		
		console.log($aQ);
		
		//append bar to qrb_progress div
		var progDiv = document.createElement("div");
		progDiv.classList = "bg-light tx-proggy";
		progDiv.setAttribute("id", "tx_totals");
		ANNE.A.append('qrb_progress', progDiv);

		var progressBar = document.createElement("div");
		progressBar.setAttribute("id", "tx_progress_bar");
		progressBar.classList = "proggy-bar positive";
		progressBar.style.display = "inline-block!important";

		ANNE.A.append(progDiv, progressBar, true);

		for (var c=0; c < $aQ.length; c++) {
			var cmdObj = $aQ[c];
			var txtlbl = "";

			if (cmdObj.cmds_special.length && cmdObj.cmds_special.indexOf(TXB.CMD_POPULATE_WITH_VARS_ONLY_NO_SEND)!=-1) {
				continue;//no need to display this in the gui.
			}
			
			if (cmdObj.relon!=undefined && cmdObj.relon.warning!=undefined) {
				ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q bg-warning", "tx", c+"x_warn",  cmdObj.relon.warning);
			}

			var msg = "";
			if (cmdObj.msg!="" && cmdObj.relon!=null && cmdObj.relon.debug_name==undefined) {
				msg = cmdObj.msg;
			}
			else if (cmdObj.relon!=undefined) {
				//$rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false
				msg = ANNE.TALK.simpleRelonTags(cmdObj.relon, cmdObj.TXTYPE, false);
			}

			
			if (cmdObj.relon!=undefined && cmdObj.relon!==null) {
				if (cmdObj.relon.type==TYPE_PARAM) {
					msg = msg +" <span class='tiny-text'>PARAM setter</span>";
				}
			}

			if (cmdObj.relon!=undefined && cmdObj.relon.sys_ref_txid!=undefined) {
				msg = msg +" <span class='tiny-text'>&nbsp;depends on "+cmdObj.relon.sys_ref_txid+"</span>";
			}

			if (cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_CHILD_TO_CMD)!=-1 
			|| cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_SELF_TO)!=-1) {
				msg = msg +"<span class='tiny-text'>&nbsp;ENCRYPTED AFTER NID IS ACQUIRED</span>";
			}

			var ico = "";
			ico = UTILS.formatNumber(cmdObj.totalAnnies/100)+ANNE.GUI_STUFF.getAnneImgHtml('16px');

			//friendly ico display in the q
			var ico3 = "";
			if (cmdObj.relon!=undefined) {
				ico3 = ANNE.TXB.getTxICOByType(cmdObj.TXTYPE, cmdObj.relon, "16");
			}

			var txtlbl = "<span class='q-relon-link'>tx #"+(c+1)+"</span>"+ico3;
			var total = "<span class='q-flexright' style='text-align:right'>"+ico+"</span>";

			var finalmsg = txtlbl+msg+total;

			ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q", "tx", c+"x",  finalmsg);
		}
	}


	/**
	*/
	printNids = function($nids, $title="") {
		if ($title!="") console.log("%c"+$title, ANNE.TALK.bgbigred);
		
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getNeuronByNid($nids[n]);
			console.log("n="+n+" "+nObj.text);
		}	
	}

	
	/**
	 *
	 */
	updateSantizedAlertCmd=function($updateElID, $cmdObj, $current_index, $total, $badgeEl="") {
		var txid = $cmdObj.broadcast_txid;
		var percentFloat = ($current_index+1)/$total;
		var bFINAL = false;
		var remain = $total-($current_index+1);
		
		if (remain>0) ANNE.A.setHTML($badgeEl, remain);
		else ANNE.A.setHTML($badgeEl, "");
		
		if ($current_index+1==$total) bFINAL=true;

		var percent = Math.ceil(percentFloat*100);
		ANNE.A.setStyle("tx_progress_bar", {width:percent+"%"}, true);
		var msgDiv = document.getElementById($updateElID);

		//try to track with the element getting updated.
		
		if (($cmdObj.skipped==undefined || $cmdObj.skipped==false) && txid!=undefined && txid.length>0) {
			var msg = "";
			if ($cmdObj.msg=="") msg = "success! ";
			else msg = $cmdObj.msg;
		
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-green-themeDARK", "bg-txid-alert-q");
			
			//update the txid #n with the txid ico.
			var txidcol = tx_alert.querySelector(".q-relon-link");
			if (txidcol!=null && txid!="") {
				ANNE.MEM_UDF.makeRelonViewLink(txidcol, txid, true, 12, true);
			}

			//if this was the last $cmdObj, print the complete transactions.  :)
			if (bFINAL==true) {
				var ezal = ANNE.A.ezAlert(msgDiv, "bg-green-themeDARK", "", "",  "<b>"+$total+"</b>&nbsp;Transactions Complete");
				ezal.classList.add("txs-complete");
				
			}
			return;
		}

		else if ($cmdObj.skipped!=undefined && $cmdObj.skipped==true) {
			var msg = "skipped::";
			if ($cmdObj.skipped_reason!=undefined) msg += $cmdObj.skipped_reason;
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-red", "bg-txid-alert-q");
		}
	}



	/*** printRelonByID ************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/ 
	printRelonByID=function($id) {
		var rObj = MEMG.lookupRelonByTXID($id);
		if (rObj) this.printRelon(rObj);
	}

	//here we just want the names.
	printNeurons = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log($nObjs[n].bestNameHTML());
		}
	}

	printNObjsList = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) { console.log($nObjs[n].bestName());}
	}

	/*** printNObjs ****************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/
	printNObjs = function ($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log("***********************************");
			var nObj = $nObjs[n];
			console.log("for neuron: "+nObj.nid);
			if (nObj.mine==true) console.log("mine=true!");
			else console.log("not mine");

			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printFromTempNidArray=function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log("for neuron: "+nObj.nid);
			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printOnlyRelons = function($TYPE, $aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		for (var r = 0; r < $aRelons.length; r++) {
			if ($aRelons[r].type!=$TYPE) continue;
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
	}

	/**
	 * Print an array of relons.
	 * 
	 * @param {Array} $aRelons 
	 * @param {Boolean} $plainText 
	 * @param {Boolean} $bDOTEMPNIDS 
	 * @returns 
	 */
	printRelons = function ($aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if (ANNE.UTILS.isset($aRelons)==false) {
			console.log("no relons to print");
			return;
		} 

		console.log($aRelons);

		for (var r = 0; r < $aRelons.length; r++) {
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
		return;
	}


	printTNIDS = function($tnids){
		var texts=[];
		for (var t=0; t < $tnids.length; t++) {
			var nOBJ = MEMG.getNeuronByTempNid($tnids[t]);
			console.log(nOBJ.text);
			texts.push(nOBJ.text);
		}
		console.log(texts.join(","));
		return texts.join(",");
	}



	/*** selfRelon ******************************************* xx/xx/2020 ***
	******************************************************** Mr.Scatman ****/ 
	selfRelon = function ($relon, $bYOU=true) {
		var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var toObj = MEMG.getNeuronByTempNid($relon.t);
		var totxt="";
		if (toObj!=null) totxt = toObj.text;
		else totxt = $relon.to_s;

		var your = "Your";
		if ($bYOU==false) your = "Their";
		var you = "You";
		if ($bYOU==false) you = "They";

		var msg = "<span class='text-primary'> "+you+"</span>";
		if ($relon.type==TYPE_AWARENESS) {
			
			var toTxt = "";
			if ($relon.to_type==TO_TYPE_ANNE) {
				toTxt = this.verbalizeForAwareness($relon);
			}
			else toTxt = $relon.to_s;
			return your + " "+relnTxt + " "+toTxt;
		}

		else if ($relon.type==TYPE_BE) {
			if ($relon.to_type==TO_TYPE_STRING) {
				var your = "<span class='text-primary'> "+your+"</span>";
				if ($relon.r==MEM.name.temp_nid) return your+" <span class='text-danger'>name </span>is <span class='text-success-lbl'>"+totxt+"</span>";
				else if ($relon.r==MEM.aka.temp_nid) return msg+" are <span class='text-danger'>also known as </span>is <span class='text-success-lbl'>"+totxt+"</span>";
			}
			msg = msg +" are ";
			var determiner = this.getDet(totxt);
			return msg +determiner+" <span class='text-danger'>"+totxt+"</span>";
		}
		else if ($relon.type==TYPE_SYS) {
			if (toObj!=null) {
				if ($relon.r==MEM.necrosis.temp_nid) {
					"<span class='text-danger-lbl'>"+relnNOBJ.text+" (delete) </span> <span class='text-success'>"+totxt+"</span>";;
				}
				else return relnNOBJ.text+" = "+totxt;
			}
			else return relnNOBJ.text+" = "+$relon.to_s; 
		}
		else if ($relon.type==TYPE_HAS) {
			msg = msg+" have ";	
			var determiner = this.getDet(relnNOBJ.text);
			var reln_txt = "<span class='text-danger'>"+relnNOBJ.text+"</span>";
			msg = msg+determiner+" "+reln_txt+" of ";

			var toStr = "";
			if (toObj!=null) toStr = totxt;
			else toStr = $relon.to_s;

			var to_txt = ""
			if (toObj!=null && toObj.bTERMINATING_TYPE==true) to_txt = "<span class='text-success'>"+toStr+"</span>";
			else to_txt = "<span class='text-danger'>"+toStr+"</span>";
			return msg + to_txt;
		}
	}


	printListByReln = function($outs, $r, $excludeTypes=[TYPE_PARAM], $sep=", ") {
		var relons = MEMG.getRelonsByTempR($outs, $r, $excludeTypes);
		if (relons.length==0) return "";
		return MEMG.getTextListByRelonTos(relons, $sep);
	}


	colorizeByPol=function($nObj) {
		if ($nObj.polarity==undefined) MEM.polarity.set_value($nObj);
		var fclass = "";
		if ($nObj.polarity==1) fclass = "text-success-lbl";
		else if ($nObj.polarity==-1) fclass = "text-danger-lbl";
		else if ($nObj.polarity==0) fclass = "text-warning-lbl";
		return "<span class='q-part "+fclass+"'>"+$nObj.text+"</span>";
	}


	printEzName = function($nObj) {
		var text = $nObj.text;
		var be = $nObj.be;
		if (text=="") {
			return "<em>(unnamed) "+be+"</em> <span class='raw-address'>"+$nObj.nid+"</span>";
		}

		return "<em>"+be+"</em> <span style='color:black; font-weight:bold'>"+text+"</span>&nbsp;(<span class='raw-address'>"+$nObj.nid+"</span>)";
	}



	/**
	 * Get a determiner for a given string.
	 * 
	 * Have to go to phonemes to get accurate with this.
	 * Not doing that now.  going the cheap and inaccurate road.
	 * 
	 * @author {Mr.Scatman}
	 */ 
	getDet=function ($str) {
		var str =$str.toLowerCase();
		var char = str.substring(0,1);
		var aeiou = "aeiou";
		if (aeiou.indexOf(char)!=-1) return "an";
		return "a";	
	}

	/**
	 * Another rudimentary det function in lieu of using NLP in this app.
	 * (that comes later with a more robust vocab and understanding of lang in the circuits..
	 * ...tho.. technically... we can use the NLP classes to parse now)
	 * @param {*} $nObj 
	 * @returns 
	 */
	getDet2=function ($nObj) {
		var str = $nObj.text.toLowerCase();

		if ($nObj.ntype==NTYPE_INST) {
			if ($nObj.bes.indexOf(MEM.person.temp_nid)!=-1 || $nObj.bes.indexOf(MEM.location.temp_nid)!=-1) return $nObj.text;
			return "the "+$nObj.text;
		}

		//an elephant, an apple, a peach, a car, a dog.
		else {
			var char = str.substring(0,1);
			var aeiou = "aeiou";
			if (aeiou.indexOf(char)!=-1) return "an "+$nObj.text;
			return "a "+$nObj.text;		
		}
	
	}
	


	/** for later.. */
	toSoundEncoding(word) {
		return word.toLowerCase().split('')
		.map(function(c) { return encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */ 
	soundsLike=function($w1, $w2) {
		return this.toSoundEncoding($w2) === this.toSoundEncoding($w2);
	}


	/** for later.. */ 
	toSoundEncoding=function($w) {
		return $w.toLowerCase().split('')
		.map(function(c) {return this.encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */
	encodeCharacter=function(c) {
		if ('kgq'.indexOf(c) >= 0) return 'k'; // velar stops
		if ('bpfv'.indexOf(c) >= 0) return 'p'; // labials
		if('td'.indexOf(c) >= 0) return 't'; // alveolars
		if ('szjc'.indexOf(c) >= 0) return 's'; // shushes
		if ('mn'.indexOf(c) >= 0) return 'm'; // nasals
		if ('lr'.indexOf(c) >= 0) return 'l'; // glides
		if ('aeiou'.indexOf(c) >= 0) return 'o'; // vowels
		if ('yhw'.indexOf(c) >= 0) return '';
		return '';
	}
	

	/**
	 * Quick n dirty ending printer for frenly day printing of dates.
	 */
	lazyDayIntEnding = function($int) {
		if ($int==1 || $int==21 || $int==31) return $int+"st";
		else if ($int==2 || $int==22) return $int+"nd";
		else if ($int==3 || $int==23) return $int+"rd";
		else return $int+"th";
	}


	/**
	 * poor man's printing -- no intelligence involved.	
	 * quite rudimentary -- but for now, it works.
	 */
	humanReadableRelon = function ($relon, $bNOFONT=false) {
		//redo this.. for now, just print relon.
		return this.printRelon($relon, true, false);
	}


	/**
	 */
	whoOwnsPhrase = function($who, $nObj=null) {
		if ($who=="anne") return "A.N.N.E. owns this neuron";
		else if ($who=="app") return "This app owns this neuron";
		else if ($who=="me") {
			if ($nObj.temp_nid==USER._userNObj.temp_nid) return "<span class='text-primary'>This IS you!</span>";
			else return "<span class='text-primary'>You owns this neuron!</span>";
		}
		else if ($who=="worldview") return "This is in your <span class='text-primary'>worldview</span>";
		
		return "**This neuron is not in your worldview**";
	}


	
	/**
	 */ 
	 verbalizeForAwareness = function ($relon, $bTAGIFY=true) {
		if ($relon.to_type!=TO_TYPE_ANNE) return " [[[don't use this w/ non neuron tos]]] ";
		var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
		var rNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var relnTxt = this.colorizeByPol(rNOBJ);

		//opinionz are just keywords..
		//if (rNOBJ.bes[0]==MEM.symbol.temp_nid) {
			//var determiner = this.getDet(relnTxt);
			//if ($bTAGIFY==false) return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+relnTxt;	
			//return "<span class='type-aware'>AWARE</span><span class='q-part rel-from'>I</span> <span class='q-part'>think </span><span class='q-part'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>is</span> "+relnTxt;	
		//}

		//else 
		if (rNOBJ.nid==MEM.rememberz.nid) {
			if ($bTAGIFY==false) return "rememberz "+toNOBJ.bestNameHTML();
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-reln'>rememberz</span><span class='q-part reln-to'>"+toNOBJ.bestNameHTML()+"</span>";
		}

		//the rest are pos/neg/neu feelings (buckets) or feelz (most refined expression of feeling)
		else {
			if ($bTAGIFY==false) return ""+toNOBJ.bestNameHTML()+" makes me FEELZ "+relnTxt;
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-to'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>makes</span><span class='q-part rel-from'>me</span>"+relnTxt;
		}
	}


	/**
	 */
	 printAwarenessForDisplay = function($relnNOBJ, $tnid) {
		var tnid = parseInt($tnid);
		var targetNOBJ = MEMG.getNeuronByTempNid(tnid);
		if (targetNOBJ==null) return "fix this";
		if (targetNOBJ.bes[0]==MEM.opinionz.temp_nid) {
			var determiner = this.getDet($relnNOBJ.text);
			return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+$relnNOBJ.text;	
		}
		else if ($relnNOBJ.nid==MEM.rememberz.nid) return "rememberz "+targetNOBJ.bestNameHTML();	
		return targetNOBJ.bestNameHTML()+" makes me FEELZ "+$relnNOBJ.bestNameHTML();
	}


	//tells us what something is from a text search.
	//this function is used as a priority when searching for things
	//we prioritize YOU first, then ANNE, then temp-worldview-pass (review this), then outside the network.
	verbalizePriorityFromTallySummary=function($tally, $textVal, $parentNObj=null, $prop="name") {
		var who = ""
		var has = "has";
		if ($tally.me) {
			who = "YOU";
			has = "have";
		}
		else if ($tally.anne) who = "A.N.N.E.";
		else if ($tally.app) who = "THIS '"+MEM_UDF.THIS_APP.text.toUpperCase()+"' APP";
		else if ($tally.worldview) who = "An adopted neuron in your network";
		else if ($tally.wv_pass) who = "A temporary-worldview-pass neuron"; 
		else if ($tally.other) who = "Someone outside of your network";
		who = "<span style='color:white'>"+who+"</span>"+"&nbsp;"+has;
		var search_msg = who+"&nbsp;already created a neuron with the same '"+$prop+"' <span class='text-success'>&nbsp;'"+$textVal+"'</span>";
		if ($parentNObj != null) search_msg=search_msg+"&nbsp;with parent class <span class='text-success'>&nbsp;'"+$parentNObj.text+"'</span>";
		msg = "As a friendly reminder;&nbsp;"+search_msg+"!  Are you sure you want to create another one?"; 
		return msg;
	}


	/**
	 */
	 typeTalk=function($type, $bTEXTONLY=false, $invertBG=false) {
		if ($bTEXTONLY) return this.type_pre[parseInt($type)];
		var invert = "";
		if ($invertBG==true) invert="-invert";
        var type_string="";
        if ($type==TYPE_SYS_MSG) type_string="<span class='type-sys"+invert+"'>sys msg</span>";
        else if ($type==TYPE_SYS) type_string="<span class='type-sys"+invert+"'>sys</span>";
        else if ($type==TYPE_BE) type_string="<span class='type-be"+invert+"'>be</span>";
        else if ($type==TYPE_PARAM) type_string="<span class='type-param"+invert+"'>param</span>";
        else if ($type==TYPE_HAS) type_string="<span class='type-has"+invert+"'>has</span>";
        else if ($type==TYPE_AWARENESS) type_string="<span class='type-aware"+invert+"'>aware</span>";
        else if ($type==TYPE_CLUMP) type_string="<span class='type-clump"+invert+"'>clump</span>";
		else if ($type==TYPE_ROLE) type_string="<span class='type-role"+invert+"'>role</span>";
		else if ($type==TYPE_BRAIN) type_string="<span class='type-brain"+invert+"'>brain</span>";
		else if ($type==TYPE_EPISODIC) type_string="<span class='type-episodic"+invert+"'>episodic</span>";
		else if ($type==TYPE_EXPERIENCE) type_string="<span class='type-experience"+invert+"'>experience</span>";
		else if ($type==TYPE_INTERNAL_PAY) type_string="<span class='type-pay"+invert+"'>internal payment</span>";
		else if ($type==TYPE_BROADCAST_PAY) type_string="<span class='type-pay"+invert+"'>broadcast payment</span>";
		else if ($type==TYPE_BRAIN120) type_string="<span class='type-brain120"+invert+"'>brain120 backlink</span>";
		else type_string = "unhandled ("+$type+")";
		return type_string;
    }


	/**
	 * 
	 * @param {*} $type 
	 * @param {*} $fromNOBJ 
	 * @param {*} $relnNOBJ 
	 * @param {*} $toType 
	 * @param {*} $toText 
	 */
	quickRelonLinePrint=function($type=null, $fromNOBJ=null, $relnNOBJ=null, $toType=null, $toText=null) {
		var parts = [];
		if ($type!=null) parts.push("["+ANNE.TALK.typeTalk($type)+"]");
		if ($fromNOBJ!=null) parts.push($fromNOBJ.bestNameHTML()+" ->");
		if ($relnNOBJ!=null) parts.push($relnNOBJ.bestNameHTML());
		if ($toType!=null) parts.push("["+TALK.toTypeTalk($toType)+"] ->");
		if ($toText!=null) parts.push($toText);
        return parts.join(" ");
	}

	
	/**
	 */
    toTypeTalk=function($to_type, $bI=false) {
        var to_type_string="";
        if($to_type==TO_TYPE_SEED) to_type_string="encrypted seed";
        else if($to_type==TO_TYPE_ANNE) to_type_string="address";
        else if($to_type==TO_TYPE_TXID) to_type_string="txid";
        else if($to_type==TO_TYPE_URL) to_type_string="url";
        else if($to_type==TO_TYPE_NUMBER) to_type_string="number";
        else if($to_type==TO_TYPE_DATE) to_type_string="date";
        else if($to_type==TO_TYPE_DATETIME) to_type_string="datetime";
        else if($to_type==TO_TYPE_STRING) to_type_string="string";
        else if($to_type==TO_TYPE_BLOB) to_type_string="blob";
        else if($to_type==TO_TYPE_TIME) to_type_string="time";
        else if($to_type==TO_TYPE_BOOLEAN) to_type_string="boolean";
        else if($to_type==TO_TYPE_ENCR_ANNE) to_type_string="encr addr";
        else if($to_type==TO_TYPE_ENCR_TXID) to_type_string="encr txid";
        else if($to_type==TO_TYPE_ENCR_URL) to_type_string="encr url";
        else if($to_type==TO_TYPE_ENCR_NUMBER) to_type_string="encr #";
        else if($to_type==TO_TYPE_ENCR_DATE) to_type_string="encr date";
        else if($to_type==TO_TYPE_ENCR_DATETIME) to_type_string="encr dtime";
        else if($to_type==TO_TYPE_ENCR_STRING) to_type_string="encr str";
        else if($to_type==TO_TYPE_ENCR_BLOB) to_type_string="encr blob";
        else if($to_type==TO_TYPE_ENCR_TIME) to_type_string="encr time";
        else if($to_type==TO_TYPE_ENCR_BOOLEAN) to_type_string="encr bool";
        

        if($bI==true) return "<i>"+to_type_string+"</i>";
        else return to_type_string;
    }


	/**
	 */
	use_d_pat=function($d_pat, $nObj, $bDOLINKS=true, $linkType=LINKS_NEURON_ANNET) {
		if ($d_pat==undefined || $d_pat==null || $d_pat.length==0) return;
		var msg = "";
		for (var d=0; d < $d_pat.length; d++) {
			var pat = $d_pat[d];
			var prop = pat.prop;
			var pre = pat.pre;
			var post = pat.post;
			var bALT = false;
			if (prop==undefined) continue;
			var val = $nObj[prop];
			if (val==undefined || val==null) {
				if (pat.alt_lbl!=undefined) {
					val = pat.alt_lbl;
					bALT = true;
				}
				else continue;
			}
			if ($bDOLINKS==true && bALT==false) {
				var valNOBJ = MEMG.getNeuronByTempNid(val);
				if (valNOBJ==null) {
					if (pat.cancel_prev!=undefined) return "";
					continue;
				}
				val = LINKS.makeNeuronLink(valNOBJ, $linkType);
				if (pat.show_aid!=undefined && pat.show_aid==true) {
					val = val +"&nbsp;("+valNOBJ.nid+")";
				}
			}

			if (pre!="") {
				if (msg!="") msg = msg+"&nbsp;"+pre;
				else msg = pre;
			}
			if (msg=="") msg = val;
			else msg = msg+"&nbsp;"+val;
			
			if (post!="") msg = msg+"&nbsp;"+post;
		}

		return msg.trim();
	}

	printTempNids = function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log(nObj.text, nObj);
		}
	}

	/**
	 * plaintext no longer does anything.
	 * this function is a DISASTER, fay.  -dlm
	 */
	printRelon = function ($rObj, $plainText=false, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}

		//console.log($rObj);
		var typeText = this.getTypeName($rObj.type);
		var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		if (fromNOBJ==null) {
			return;
		}

		var emulated = "";
		var fromText = fromNOBJ.text;
		if (fromText=="") fromText = fromNOBJ.nid;
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		var relonText = relnNOBJ.text;
		if ( fromNOBJ.emulated) emulated = "(emulated) ";
		var relonStr = emulated+typeText + " from:" + fromText+ "  reln:"+relonText;
		var toText = this.verbalizeRelonToByToType($rObj);
		relonStr += " to:"+toText;

		var toSPrintOnly = this.toSPrint($rObj);

		var freeze = fromNOBJ.freeze;
		if (freeze==1) console.log("%cfrom nid froze", TALK.bgivoryblue);

		var ghost = "";
		if ($rObj.ghost==true) ghost="(gh)";

		var sysreftxid=""
		if ($rObj.sys_ref_txid!=undefined) sysreftxid=" sysref:"+$rObj.sys_ref_txid;

		var annies=""
		if ($rObj.annies!=undefined) annies=" annies:"+$rObj.annies;
		
		var fhi="";
		if ($rObj.flathexint!=undefined) fhi=" fhi:"+$rObj.flathexint;
		
		if ($consoleColor!=-1) {
			console.log("%cg"+ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi, $consoleColor);
		}
		else {
			console.log(ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi);
		}

		if ($plainText==true) return emulated+"[" + typeText+"]"+fromText+relonText+toSPrintOnly;
		return relonStr;
	}


	toSPrint=function($rObj){
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, false);
		return toText;
	}


	/**
	 */
	verbalizeRelonToByToType = function ($rObj, $bCOLORIZEBYTYPE=true) {
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, $bCOLORIZEBYTYPE);
		return toText;
	}


	/**
	 */ 
	getToTextByToType = function ($to_type, $to_s, $colorize=true) {
		var text = $to_s;
		if ($to_type == TO_TYPE_ANNE) {
			var nObj = MEMG.getNeuronByNid($to_s);
			if (nObj) {
				if ($colorize==true) text = nObj.bestNameHTML();
				else text = nObj.bestName();
			}	
		}
		if ($colorize) return "<span class='to-type tt-"+$to_type+"'>"+text+"</span>";
		else return text;
	}


	/**
	 */
	getTypeName = function ($type, $bDOCLASSHTML=false, $addClass="") {
		if ($bDOCLASSHTML==false) {
			if ($type == TYPE_SYS || $type=="sys") return 'SYS';
			if ($type == TYPE_BE || $type=="be") return 'BE';
			if ($type == TYPE_PARAM || $type=="para") return 'PARAM';
			if ($type == TYPE_HAS || $type=="has") return 'HAS';
			if ($type == TYPE_AWARENESS || $type=="awareness") return 'AWARENESS';
			if ($type == TYPE_SYS_MSG || $type=="msg") return  'SYS MSG';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return  'EXPER';
			if ($type == TYPE_CLUMP || $type=="clump") return  'CLUMP';
			if ($type == TYPE_ROLE || $type=="role") return  'ROLE';
			if ($type == TYPE_BRAIN || $type=="brain") return  'BRAIN';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return  'INTERNAL PAYMENT';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return  'BROADCAST PAYMENT';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return  'BRAIN120';
			
			
		}
		else {
			if ($type == TYPE_SYS || $type=="sys") return '<span class="type-sys '+$addClass+'">SYS</span>';
			if ($type == TYPE_BE || $type=="be") return '<span class="type-be '+$addClass+'">BE</span>';
			if ($type == TYPE_PARAM || $type=="param") return '<span class="type-param '+$addClass+'">PARAM</span>';
			if ($type == TYPE_HAS || $type=="has") return '<span class="type-has '+$addClass+'">HAS</span>';
			if ($type == TYPE_AWARENESS || $type=="awareness") return '<span class="type-awareness '+$addClass+'">AWARE</span>';
			if ($type == TYPE_SYS_MSG || $type=="msg") return '<span class="type-sys-msg '+$addClass+'">SYS MSG</span>';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return '<span class="type-experience '+$addClass+'">EXPER</span>';
			if ($type == TYPE_CLUMP || $type=="clump") return '<span class="type-clump '+$addClass+'">CLUMP</span>';
			if ($type == TYPE_ROLE || $type=="role") return '<span class="type-role '+$addClass+'">ROLE</span>';
			if ($type == TYPE_BRAIN || $type=="brain") return '<span class="type-brain '+$addClass+'">BRAIN</span>';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return '<span class="type-brain '+$addClass+'">INTERNAL PAYMENT</span>';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return '<span class="type-brain '+$addClass+'">BROADCAST PAYMENT</span>';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return '<span class="type-brain '+$addClass+'">BRAIN120</span>';
			
		}

		//anything 100 and over is typebrain 0, 1, 2 etc
		if ($type>=20 && $type<120) return "BRAIN"+(parseInt($type));
		return 'unknown '+$type;
	}


	/**
	 */
	pluralizeNounIf = function($mixed, $baseWord, $plural, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $baseWord+$plural;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $baseWord+$plural;
		return $baseWord+$singularPostFix;
	}

	pluralizeNounIf2 = function($mixed, $baseWord, $pluralwd, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $pluralwd;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $pluralwd;
		return $baseWord+$singularPostFix;
	}

	/**
	 */
	pluralizeNP = function ($arr, $baseWord, $plural, $conj="or", $nestedProp="content") {
		if ($arr.length>1) return $baseWord+$plural+" "+this.listify($arr, $conj, $nestedProp);
		else  {
			if ($nestedProp!="") return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0][$nestedProp].text+"</span>";
			else return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0].text+"</span>";
		}
	}


	/**
	 */
	listify=function($arr, $conj, $nestedProp="content"){
		var list = "";
		for (var a=0; a < $arr.length; a++) {
			var sep = "";
			if ((a+1)==$arr.length) sep = "</span>&nbsp"+$conj+"&nbsp;<span class='text-success-lbl'>";
			else if (a>=1) sep = ",&nbsp;";
			else list = "<span class='text-success-lbl'>";
			
			if ($nestedProp) list += sep+""+$arr[a][$nestedProp].text+"";
			else list += sep+"<b>"+$arr[a].text+"</b>";
		}
		return list+"</span>";
	}


	/**
	 */
	prounounPick=function($arr, $pr, $prpl) {
		if ($arr.length>1) return $prpl;
		else return $pr;
	}


	//To make regular nouns plural, add ‑s to the end. ...
	//If the singular noun ends in ‑s, -ss, -sh, -ch, -x, or -z, add ‑es to the end to make it plural. ...
	//In some cases, singular nouns ending in -s or -z, require that you double the -s or -z prior to adding the -es for pluralization.
	pluralizeNoun=function($text) {
		var rev = $text.split("").reverse().join("");
		var c0 = rev[0];
		var c1 = "";
		if (rev.length>1) c1 = rev[1];
		if (c0=="s" || c0=="x" || c0=="z") {
			if (c0=="s" && c1=="s") return $text+"s";
			if (c0=="s") return $text+"ses";
			if (c0=="z") return $text+"zes";
			return $text+"es";
		}
		if (c0=="h" && (c1=="s" || c1=="c")) return $text+"es";
		return $text + "s";
	}


	/**
	 */
	beState = function($array) {
		if ($array.length==0) return "are";
		else if ($array.length==1) return "is";
		return "are";
	}

	/**
	 * make an english list phrase of "x, y, z and a" from an array of [x, y, z, a]
	 */
	listifyArray=function($arr) {
		if ($arr.length==0) return "";
		if ($arr.length==1) return $arr[0];
		var str = $arr[0];
		for (var a=1; a < $arr.length-1; a++) {str = str+", "+$arr[a];}
		str = str+ " and "+$arr[($arr.length-1)];
		return str;
	}


	/**
	 * Just some helpful dumb msgs popped in here for certain relons.
	 * We can display these in the relons page to inform the user a bit more
	 * about what this relons is/was for.
	 */
	basicExplain = async($rObj)=> {
		var rel_sponsor_msg="";

		if ($rObj.relon_sponsor!=undefined && $rObj.relon_sponsor!=null && $rObj.relon_sponsor!="") {
			rel_sponsor_msg = "  This was originally a collab";
		}

		if ($rObj.from_nid==MEM.___SELF___ && $rObj.reln==MEM.BLOCK_WINNER.nid) {
			return "This is a block reward winner.";
		}
		if ($rObj.type==TYPE_BROADCAST_PAY && $rObj.reln==MEM.annies.nid && $rObj.to_type==TO_TYPE_ANNE) {
			return "This was a broadcast payment relon.";
		}
		if ($rObj.ghost==true && $rObj.to_type>=TO_TYPE_ENCR_ANNE) {
			var fromNOBJ = MEMG.getNeuronByTempNid($rObj.f);
			if (fromNOBJ.ghost==true) return "This to value will be encrypted at run-time when broadcast.";
		}
		if ($rObj.r==MEM.anj_accepted.temp_nid) return "This is an accepted anjucator invite.";
		if ($rObj.r==MEM.anj_rejected.temp_nid) return "This is an declined anjucator invite.";
		if ($rObj.reln==MEM.anjucator.nid) {
			if ($rObj.sys_ref_txid!=undefined && $rObj.sys_ref_txid.length) {
				if ($rObj.to_type==TO_TYPE_ANNE) {
					var srtRel=MEMG.lookupByTXID($rObj.sys_ref_txid);
					if (srtRel!=null) {
						if (srtRel.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to add an anjucator to a neuron.";
						else if (srtRel.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to add an anjucator to a neuron.";
					}
				}
				else if ($rObj.to_type==TO_TYPE_TXID) {
					if ($rObj.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to be added as an anjucator to a neuron.";
					else if ($rObj.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to be added as an anjucator to a neuron.";
				}
			}
			else return "This is a request to add an anjucator to a neuron.";
		}
		if ($rObj.reln==MEM.winner_payout.nid) return "This is a winning payout for the boosties bucket draw."
		if ($rObj.reln==MEM.hanai_in.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is approval of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is approval of a HANAI request.";
		}
		if ($rObj.reln==MEM.hanai_in_reject.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is rejection of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is rejection of a HANAI request.";
		}
        if ($rObj.r==MEM.andy_collab_accepted.temp_nid) return "This is an accepted collab request.";
        if ($rObj.r==MEM.andy_collab_rejected.temp_nid) return "This is a rejected collab request.";
        if ($rObj.r==MEM.hanai_out.temp_nid) {
            var whoNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			return "This is a request to HANAI out a neuron into the <b>"+whoNOBJ.bestNameHTML()+"</b> worldview.";
        }
		if ($rObj.r==MEM.anndle_offer.temp_nid) {
			var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			var pending = "is a pending";
			if (ANNE.UTILS.isset($rObj.sys_ref_txid)==true) pending = "was an ";
			return "This "+pending+" ANNDLE offer that paid "+UTILS.formatNumber($rObj.annies)+" annies ("+UTILS.formatNumber($rObj.annies/100)+" ANNE coin) to control the '"+toNOBJ.text+"' ANNDLE.";
		}

        if ($rObj.type==TYPE_SYS_MSG) {
			if ($rObj.r==MEM.annexed_by.temp_nid) {

				var arel = null;
				var ANNEXED_THING = null;

				if ($rObj.sys_ref_txid!=null) arel = ANNE.MEMG.lookupByTXID($rObj.sys_ref_txid);
				if (arel!=null) {
					ANNEXED_THING = MEMG.getNeuronByTempNid(arel.t);
				}

				if ($rObj.f==USER._userNObj.nid) {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Your neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "Your neuron was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
				else {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "A neuron from user "+MEMG.getOrMakeIfNew($rObj.to_s)+" was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
			}

            //if ($rObj.r==MEM.aon.temp_nid) return "This is an onboarding relon with "+$rObj.annies+" annies given to the recipient.";
            return "This is a relon sent as a <b>message</b> to an ANNODE for instructions to take action.  It can also be generated by the annode on behalf of a keyless neuron and serve as a status update or receipt.";
        }

        else if ($rObj.type==TYPE_SYS) return "This is a <b>system</b> relon used to engage with ANNE protocol rules.";
        else if ($rObj.type==TYPE_BE) {
            if ($rObj.r==MEM.class.temp_nid || $rObj.r==MEM.inst.temp_nid) {
                return "This is a relon that specifies be-hierarchical parent information."+rel_sponsor_msg;
            }
            else if ($rObj.r==MEM.name.temp_nid) return "This is a relon that sets the neuron's name."+rel_sponsor_msg;
            else if ($rObj.r==MEM.aka.temp_nid) return "This is a relon that sets an <b>also known as</b> for this neuron."+rel_sponsor_msg;
            else if ($rObj.r==MEM.fka.temp_nid) return "This is a relon that sets a <b>formerly known as</b> for this neuron."+rel_sponsor_msg;
        	return "This is a relon that specifies additional be-defining information."+rel_sponsor_msg; 
        }
        else if ($rObj.type==TYPE_PARAM) {
            return "This is a <b>parameter</b> relon defining a relationship that instances can provide."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_HAS) {
            if (this.bCOLLAB==true) {
                if (this.bISPENDING==true || ($rObj.status_text!=undefined && $rObj.status_text=="p")) {
					return "This is a collab relon request.";
                }
                return "This is a collab relon request.";
            }
            return "This is a relon which specifies a parameter value a neuron <b>has</b>."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_AWARENESS) {
            return "This is a relon for run-time feelz and opinionz of an identity using ANNE.";
        }
        else if ($rObj.type==TYPE_BRAIN) {
            return "This is a relon used for wiring neurons together in the non-semantic <b>brain</b> dimension."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_CLUMP) return "This is a relon connecting to a more clump neuron."+rel_sponsor_msg;
        else if ($rObj.type==TYPE_ROLE) {
            return "This is a semantic role relon that exists in a clump neuron and defines a semantic role within the clump."+rel_sponsor_msg;
        }
		else if ($rObj.type==TYPE_INTERNAL_PAY) {
			if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block reward won via solo mining by "+toNOBJ.text;
			}
			else if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_SHARE_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block share reward won via share mining by "+toNOBJ.text;
			}
			else if ($rObj.r==ANNE.MEM.boosties.temp_nid) {
				return "This added boosties to this neuron.";
			}
            else return "This is an internal payment from a keyless neuron made by L1 protocol rules.";
        }
		else if ($rObj.type>=TYPE_BRAIN0 && $rObj.type<TYPE_BRAIN120) {
			return "This is a special type to indicate sorting order #"+$rObj.type;
		}
		else if($rObj.type==TYPE_BRAIN120) {
			return "This is a special BRAIN120 relon created by the ANNODE to document backlinks and provide helpful summaries.";
		}
		else if ($rObj.type==TYPE_BRAIN121) {
			return "myelination";
		}
	}


	logFunc=function($funct, $str="", $str2="", $str3="") {
		if ($str3!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2+" %c"+$str3, TALK.funct, TALK.bggreen, TALK.bgblue, TALK.bgyellow);
		else if ($str2!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2, TALK.funct, TALK.bggreen, TALK.bgblue);
		else if ($str!="") console.log("%c("+$funct+") %c"+$str, TALK.funct, TALK.bggreen);
	}
        
       
        

}


const TALK = new anne_talk ();
TALK.set_init(TALK);




/************** (from file ./js/base/TALK.js)**/ 


//
class generic_websocket {
    constructor($ws_uri, $bKEEP_ALIVE=true) {
        this.uri = $ws_uri;
        this.bKEEP_ALIVE = $bKEEP_ALIVE;
        this.ws = null;
        this.cb_data = null;
    }
    set_data_cb = function($cb) {
        this.cb_data = $cb;
    }
    connect = async()=> {
        return new Promise((resolve, reject) => {
            this.ws = new WebSocket(this.uri); 
            this.ws.onerror = function(err) {
                console.warn("on error");
                console.log("ERROR!", err );
            };
            this.ws.onopen = function() {
                resolve(true);
            };
            this.ws.onmessage = function ($evt) {
                if (this.cb_data!=null) this.cb_data($evt);
            };
            this.ws.onclose = function() {
                if (this.bKEEP_ALIVE==true) {
                    this.connect();
                }
            };
        });
    }

    sendJSON=function($json) {
        if (this.ws==null) return;
        var msg = JSON.stringify($json);
        this.ws.send(msg);
    }
}



var SUBSCRIBER = {}

SUBSCRIBER.bANNEX_ON = false;
SUBSCRIBER.bBOOSTIES_ON = false;
SUBSCRIBER.bBLOCKS_ON = false;
SUBSCRIBER.bANTOR_ON = false;

//the two sockets we use
SUBSCRIBER.bCHAINSOCKET_ON = false;
SUBSCRIBER.bALTDATASOCKET_ON = false;

SUBSCRIBER._cb_new_block = null;
SUBSCRIBER._cb_new_boosties = null;
SUBSCRIBER._cb_new_annex = null;
SUBSCRIBER._cb_new_transaction = null;
SUBSCRIBER._cb_new_mempool = null;
SUBSCRIBER._cb_antor = null;

SUBSCRIBER.chain_ws = null;
SUBSCRIBER.altdata_ws = null;

SUBSCRIBER.el_log = null;

//an assoc array of assoc array of callbacks by [type][subtpe] for alt data.
SUBSCRIBER._cb_alt_data = [];
SUBSCRIBER._cb_alt_data_cdxfilters = [];

SUBSCRIBER.el_cd = "socket_chaindata";
SUBSCRIBER.el_ad = "socket_altdata";

SUBSCRIBER._altdata_closed_cb = null;
SUBSCRIBER._chaindata_closed_cb = null;

/**
 * subscribe to websocket for new blocks. 
 */
SUBSCRIBER.subscribeToBlocks=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bBLOCKS_ON==false) {
            SUBSCRIBER.sendJSONToServer("onblock", null);
            SUBSCRIBER.bBLOCKS_ON = true;
        }
        resolve(true);
    });
}





SUBSCRIBER.subscribeToAntor=async($cb=null, $file_nid, $elid=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bANTOR_ON==false) {
            var params = {filter_nid:$file_nid, filter_nid_prop:"nid"};
            
            // the socket will pass back to us the el id when the antor is 
            // ready.  We can then map to the elid we want and display the goodies
            // directly from the socket return.  We can also store this in our antor file map.
            if ($elid!=null) params.elid = $elid;

            SUBSCRIBER.sendJSONToServer("onantor", params);
            SUBSCRIBER.bANTOR_ON = true;
        }
        resolve(true);
    });
}

/**
 * We are subscribing to mempool TXs that have us as the recipient.
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToOurMempool=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_mempool = $cb;
        SUBSCRIBER.sendJSONToServer("onmempool", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}

SUBSCRIBER.subscribeToOurMinedTXs=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_transaction = $cb;
        SUBSCRIBER.sendJSONToServer("ontransaction", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}


/**
 * subscribe to websocket for annexes (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToAnnex=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_annex = $cb;
        if (SUBSCRIBER.bANNEX_ON==false) {
            SUBSCRIBER.sendJSONToServer("onannex", null);
            SUBSCRIBER.bANNEX_ON = true;  
        }
        resolve(true);
    });
}


/**
 * subscribe to websocket for boosties (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToBoosties=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }  
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_boosties = $cb;
        if (SUBSCRIBER.bBOOSTIES_ON==false) {
            SUBSCRIBER.sendJSONToServer("onboosties", null);
            SUBSCRIBER.bBOOSTIES_ON = true;
        }
        resolve(true);
    });
}


/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initChainSocket = async()=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.chain_ws = new WebSocket(ANNE._annodesLS.last_socket+"chaindata"); 
        SUBSCRIBER.chain_ws.onerror = function(err) {
            console.warn("on error");
            console.log("ERROR!", err );
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
        };
        SUBSCRIBER.chain_ws.onopen = function() {
            //console.log("on open");
            SUBSCRIBER.bCHAINSOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to chaindata service!");
            }
            resolve(true);
        };
        
        SUBSCRIBER.chain_ws.onmessage = function (evt) {
            var json = JSON.parse(evt.data);
            var logmsghtml = SUBSCRIBER.formatLogData(json);
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog(logmsghtml);
            }
            var sub = json.subscription;
            if (sub==undefined || sub==null) {
                console.warn("Must have sub on return from socket");
                return;
            } 

            //ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on chain socket msg", ANNE.TALK.GREEN, sub);
            //console.log("chaindata socket", json);
            
            if (sub=="onblock") {
                if (SUBSCRIBER._cb_new_block!=null) {
                    SUBSCRIBER._cb_new_block(json);
                }
            }
            else if (sub=="onannex") {
                if (SUBSCRIBER._cb_new_annex!=null) {
                    SUBSCRIBER._cb_new_annex(json);
                }
            }
            else if (sub=="onboosties") {
                if (SUBSCRIBER._cb_new_boosties!=null) {
                    SUBSCRIBER._cb_new_boosties(json);
                }
            }
            else if (sub=="ontransaction") {
                console.log("new ontransaction");
                if (SUBSCRIBER._cb_new_transaction!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_transaction(json);
                }
            }

            //subscribes to specific mempool transactions (typically sends we receive.)
            else if (sub=="onmempool") {
                //console.log("new onmempool");
                if (SUBSCRIBER._cb_new_mempool!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_mempool(json);
                }
                else console.log("no callback set, do nothing.");
            }
        };

        SUBSCRIBER.chain_ws.onclose = function() {
            console.log("on close");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Disconnected from chaindata service!");
            }
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
            if (SUBSCRIBER._chaindata_closed_cb!=null) SUBSCRIBER._chaindata_closed_cb();
        };
        });
}




/**
 * if the altdatasocket is not on (never turned on, errored, or turend of)
 * then reconnect it before sending subscribe.
 * @param {*} $type 
 * @param {*} $subtype
 * @param {*} $extraJson 
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToAltData=async($type, $subtype, $extraJson, $cb=null)=> {  
    if (SUBSCRIBER.bALTDATASOCKET_ON==false) {
        await SUBSCRIBER._initAltDataSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) {
            if (SUBSCRIBER._cb_alt_data[$type]==undefined) {
                SUBSCRIBER._cb_alt_data[$type]=[];
            }
            if ($subtype!="") SUBSCRIBER._cb_alt_data[$type][$subtype]=$cb;
            else SUBSCRIBER._cb_alt_data[$type]["*"]=$cb;
        }
        SUBSCRIBER.sendJSONToServerAlt($type, $subtype, $extraJson);
        resolve(true);
    });
}


/**
 * 
 * @returns 
 */
SUBSCRIBER._initAltDataSocket = async()=> {
    if (SUBSCRIBER.bALTDATASOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.altdata_ws = new WebSocket(ANNE._annodesLS.last_socket+"altdata");
        SUBSCRIBER.altdata_ws.onopen = function() {
            console.log("alt data socket opened");
            SUBSCRIBER.bALTDATASOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to altdata service!");
            }
            resolve(true);
        };
        SUBSCRIBER.finalAltDataInit();
    });
}


/**
 * 
 * @param {*} $type 
 * @param {*} $subtype 
 * @param {*} $cdxfilter 
 */
SUBSCRIBER.setAltDataCallbackManual = function($type, $subtype, $cdxfilter) {
    if (SUBSCRIBER._cb_alt_data_cdxfilters[$type]==undefined) SUBSCRIBER._cb_alt_data_cdxfilters[$type]=new Object();
    SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype]=$cdxfilter;
}

SUBSCRIBER.removeAltDataCDXFilter = function($type,$subtype,$cb) {
    delete SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype];
}


/**
 * 
 */
SUBSCRIBER.finalAltDataInit=function() {
    SUBSCRIBER.altdata_ws.onmessage = function (evt) {
        var json = JSON.parse(evt.data);

        var logmsghtml = SUBSCRIBER.formatLogData(json);
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog(logmsghtml);
        }
        
        var adtype = json.adtype;
        if (adtype==undefined || adtype==null) {
            console.warn("Must have adtype on return from socket", json);
            return;
        } 

        var adsubtype = json.adsubtype;
        
        ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on altdata socket msg", ANNE.TALK.GREEN, adtype, ANNE.TALK.YELLOW, adsubtype);
        console.log("altdata socket", json);
            
        //filter cb (one-offs)
        if (SUBSCRIBER._cb_alt_data_cdxfilters[adtype]!=undefined 
            && adsubtype!=undefined 
            && SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype]!=undefined) {
            var cdxfilter = SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype];
            var filter = cdxfilter.filter;
            var cb = cdxfilter.cb;
            var results = json.data.results;
            for (var r=0; r < results.length; r++) {
                var obj = results[r];
                var propkey = filter.propkey;
                var propval = filter.propvalue;
                if (obj[propkey]==undefined) {
                    break;
                }
                if (obj[propkey]!=propval) {
                    break;
                }

                //take first one.  return all, then first match.
                cb(json, obj);
                return;
            }
        }
    
        if (adsubtype!=undefined && SUBSCRIBER._cb_alt_data[adtype]!=undefined) {
            if (SUBSCRIBER._cb_alt_data[adtype][adsubtype]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype][adsubtype](json);
            }
            else if (SUBSCRIBER._cb_alt_data[adtype]["*"]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype]["*"](json);
            }
        }
        else {
            console.log("*** no finalAltDataInit calback ***");
        }
    };

    SUBSCRIBER.altdata_ws.onclose = function() {
        console.log("on close");
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog("Disconnected from chaindata service!");
        }
        SUBSCRIBER.bALTDATASOCKET_ON=false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
        if (SUBSCRIBER._altdata_closed_cb!=null) SUBSCRIBER._altdata_closed_cb();

    };

    SUBSCRIBER.altdata_ws.onerror = function(err) {
        console.warn("on error");
        console.log("ERROR!", err )
        SUBSCRIBER.bALTDATASOCKET_ON = false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
    };
}







/**
 * 
 * @param {*} $subscribe 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServer=function($subscribe, $json) {
    //console.log("SendToServer subscribe:"+$subscribe);
    if ($json!=null) $json.subscribe = $subscribe;
    else $json = {subscribe:$subscribe};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.chain_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.chain_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.chain_ws.readyState);
    }
}


/**
 * 
 * @param {*} $adtype 
 * @param {*} $adsubtype 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServerAlt=function($adtype, $adsubtype, $json) {
    console.log("sendJSONToServerAlt adtype:"+$adtype, "adsubtype:"+$adsubtype);
    if ($json!=null) {
        $json.adtype = $adtype;
        $json.adsubtype = $adsubtype;
    }
    else $json = {adtype:$adtype, adsubtype:$adsubtype};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.altdata_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.altdata_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.altdata_ws.readyState);
    }
}



/**
 * 
 * @param {*} $logText 
 */
SUBSCRIBER.appendLog=function($logText) {
    if (SUBSCRIBER.el_log==null) return; 
    var log = document.getElementById(SUBSCRIBER.el_log);
    if (log==null) return;
    log.innerHTML = log.innerHTML + $logText + "<BR>>";
}


/**
 * 
 * @param {*} $d 
 * @returns 
 */
SUBSCRIBER.formatLogData = function($d) {
    const good_status = "<span class='good'>good</span>";
    const bad_status = "<span class='bad'>bad</span>";

    var html = "";
    if ($d.status=="good") html = good_status;
    else html = bad_status;
    html = html + "<span class='subscription'>"+$d.subscription +"</span>";
    if ($d.message!=undefined) html = html + "<span class='message'>"+$d.message+"</span>";
    if ($d.data!=undefined) {
        html = html + SUBSCRIBER.jsonSyntaxHighlight($d.data);
    }
    return html;
}


/**
 * 
 * @param {*} $json 
 * @returns 
 */
SUBSCRIBER.jsonSyntaxHighlight = function ($json) {
    if ($json == null) return "";
    if (typeof $json != 'string') {
        $json = JSON.stringify($json, undefined, 2);
    }
    $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
    return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
        var cls = 'number';
        if (/^"/.test(match)) {
            if (/:$/.test(match)) {
                cls = 'key';
            } else {
                cls = 'string';
            }
        } else if (/true|false/.test(match)) {
            cls = 'boolean';
        } else if (/null/.test(match)) {
            cls = 'null';
        }
        return '<span class="' + cls + '">' + match + '</span>';
    });
}

/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initGeneric = async($ws_uri, $KEEP_ALIVE=true)=> {
    var ws_generic = new generic_websocket($ws_uri, $bKEEP_ALIVE)
    ws_generic.connect();
}

/************** (from file ./js/network/subscriber.js)**/ 

var DB_local = {};
DB_local.db = null;
DB_local.dbName = "ANNODE_local";
DB_local.bSUPPORTED = false;
DB_local.version = 6;

DB_local.editName = function($POST) {
    if (DB_local.dbName.indexOf("_"+$POST)!=-1) return;
    DB_local.dbName = DB_local.dbName+"_"+$POST;
}

/**
 * Do we even use full pull history at all anymore in DB LOCAL????
 * @author {Mr.Scatman}
 */
 DB_local.___init=async()=> {
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback();
        return;
    }
    var done = await DB_local.openConnection(DB_local.dbName, DB_local.version);
    return;
}

DB_local.___initMMOnly=async()=>{
    if (DB_local.db!=null) return true;
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    return new Promise((resolve, reject) => {
        var request = indexedDB.open(DB_local.dbName, DB_local.version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;
            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }
            return resolve(false);  //no load needed
        };
    });
}


DB_local.getDBSize=async()=> {
    var size = await DB_local.getDBTblSize("collab_relons");
    size+= await DB_local.getDBTblSize("fullPullHistory");
    size+= await DB_local.getDBTblSize("ghost_relons");
    size+= await DB_local.getDBTblSize("traditional_txs");
    size+= await DB_local.getDBTblSize("external_requests");
    size+= await DB_local.getDBTblSize("adf_otc_swaps");
    size+= await DB_local.getDBTblSize("bot_trade_history");
    size+= await DB_local.getDBTblSize("otc_commish");
    return size;
}

DB_local.getDBTblSize=async($tbl, $method="readwrite")=> {
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                var arr = event.target.result;
                var json = JSON.stringify(arr);
                size = json.length;
                return resolve(size);
            };
        } 
    });
}

/**
 * We wait until we know what user we are dealing with to get these ones
 */
DB_local.__getGhostsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("ghost_relons");
    return aGhostRelons;
}


/**
 * Table for exensions to store a history of the requests they saw.
 * 
 * Use.  subtype, msg, exernaldata and datestamp.
 * @param {*} $onLoadCallback 
 * @returns 
 */
DB_local.__getExternalRequestsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("external_requests");
    return aGhostRelons;
}


/**
 */
DB_local.__getStoredTxs=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("traditional_txs");
}


DB_local.loadADFSwaps=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aSwaps = await DB_local.loadAll("adf_otc_swaps");
    var aSwapsFinal = [];
    for (var s=0; s < aSwaps.length; s++) {
        if (aSwaps[s].network==$network) aSwapsFinal.push(aSwaps[s]);
    }
    return aSwapsFinal;
}


/**
 * 
 * @param {*} $network 
 * @returns 
 */
DB_local.loadCommishHistory=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCommishHistory = await DB_local.loadAll("otc_commish");
    var aComisHistoryFinal = [];
    for (var s=0; s < aCommishHistory.length; s++) {
        if (aCommishHistory[s].network==$network) aComisHistoryFinal.push(aSwaps[s]);
    }
    return aComisHistoryFinal;
}



/**
 * 
 * @returns 
 */
DB_local.__getFullPullHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("fullPullHistory");
}


DB_local.__getBotTradeHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("botcoinbalance");
}

/**
 */
 DB_local.__getCollabRelons=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCollabs = await DB_local.loadAll("collab_relons");
    var ours = [];
    for (var c=0; c < aCollabs.length; c++) {
        if(aCollabs[c].from_nid==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].to_s==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].collaborator==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else {
            //do we have right to approve this?
        }
    }
    return ours;
}


/**
 * @author {Mr.Scatman}
 */
DB_local.___deleteEntireDB=async()=> {
    if (DB_local.bSUPPORTED==false) return;
    if (DB_local.db==null) return;
    return new Promise((resolve, reject) => {
        DB_local.db.close();
        var request = indexedDB.deleteDatabase(DB_local.dbName, 2);
        request.onblocked = function () {
            console.log("Delete blocked.");
            return resolve(false);
        }
        request.onerror = function () {
            console.log("Error deleting the DB");
            return resolve(false);
        };
        request.onsuccess = function () {
            console.log("Deleted OK.");
            return resolve(true);
        }   
    });             
}

/**
 * @author {Mr.Scatman}
 */
DB_local.openConnection=async($dbName, $version)=> {
    if (DB_local.db!=null) return true;
    return new Promise((resolve, reject) => {
        var request = indexedDB.open($dbName, $version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;

            if (DB_local.db.objectStoreNames.contains('ghost_relons')==false) {
                var ghostObjStore = DB_local.db.createObjectStore("ghost_relons", { keyPath:"txid" });
                ghostObjStore.createIndex("txid", "txid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('fullPullHistory')==false) {
                var fullPullHistoryStore = DB_local.db.createObjectStore("fullPullHistory", { keyPath:"pullid" });
                fullPullHistoryStore.createIndex("pullid", "pullid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('collab_relons')==false) {
                var collabRelonStore = DB_local.db.createObjectStore("collab_relons", { keyPath:"txid" });
                collabRelonStore.createIndex("txid", "txid", { unique: true });
            }
            
            if (DB_local.db.objectStoreNames.contains('traditional_txs')==false) {
                var ttStore = DB_local.db.createObjectStore("traditional_txs", { keyPath:["txid", "anne_app_user"] });
                ttStore.createIndex("usertxid", ["txid", "anne_app_user", "confirmed"], { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('external_requests')==false) {
                var extReqStore = DB_local.db.createObjectStore("external_requests", { keyPath:"req_id" });
                extReqStore.createIndex("req_id", "req_id", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('adf_otc_swaps')==false) {
                var swaps = DB_local.db.createObjectStore("adf_otc_swaps", { keyPath: ["swap_id", "network"] });
                swaps.createIndex("swap_id", "swap_id", { unique: false });
                swaps.createIndex("network", "network", { unique: false });
            }

            //ticker + currentTime = unique key
            if (DB_local.db.objectStoreNames.contains('botcoinbalance')==false) {
                var bottradehistory = DB_local.db.createObjectStore("botcoinbalance", {keyPath:["ticker", "currentTime"]});
                bottradehistory.createIndex("ticker", "ticker", { unique: false });
                bottradehistory.createIndex("currentTime", "currentTime", { unique: false });
            }

            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }

            return resolve(false);  //no load needed
        };
    });
}




DB_local.loadAll=async($tbl, $method="readwrite")=> {
    if (DB_local.db==null) return;
    if (DB_local.db.transaction==undefined) {
        return;
    }
    
    console.log("DB.loadAll "+$tbl);
    
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                return resolve(event.target.result);
            };
        } 
        else {
            var arr = [];

            // Fallback to the traditional cursor approach if getAll isn't supported.
            objectStore.openCursor().onsuccess = function(event) {
                var cursor = event.target.result;
                if (cursor) {
                    arr.push(cursor.value);
                    cursor.continue();
                } 
                else return resolve(arr);
            };
        }
    });
}


DB_local.___insert = async($objs, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    if ($objs.length==0) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {    
        $objs.forEach(function($obj) {
            var request = null;

            if ($tbl=="ghost_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                if ($obj.txid==undefined || $obj.txid==null || $obj.txid=="") return;

                //used in the gui but never broadcast.
                var TEMP_ONLY = false;
                if ($obj.TEMP_ONLY==true) {
                    TEMP_ONLY=true;
                }

                var sanitized_obj = {
                    from_nid: $obj.from_nid,
                    reln: $obj.reln,
                    rorigin: $obj.rorigin,
                    time:$obj.time,
                    to_s: $obj.to_s,
                    to_type: $obj.to_type,
                    txid: $obj.txid,
                    type: $obj.type,
                    ghost_order: $obj.ghost_order,
                    ghost_owner: $obj.ghost_owner,
                    owned_by:$obj.ghost_owner,
                    TEMP_ONLY: TEMP_ONLY
                }
                if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;

                request = objectStore.add(sanitized_obj);
            }
   
            //will need collab sanitized stuff too. (TO DO)
            else if ($tbl=="collab_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                request = objectStore.add($obj);
            }
            
            else {
                console.log("adding INDEXDB:"+$obj);
                request = objectStore.add($obj);
            }
        });

        transaction.onsuccess = function(event) {
            return resolve(true);
        };
        transaction.onerror = function(event) {
            return resolve(false);
        }
        return resolve(true);
    });
}


DB_local.___insert_single = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;

    //dont save this relon, EVER
    if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return objectStore.add($obj);    
}

DB_local.___updateAll = async($objs, $tbl="collab_relons", $method="readwrite")=> {
    for (var r=0; r < $objs.length; r++) {
        await DB_local.___update($objs[r], $tbl, $method);
    }
}

DB_local.___update = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    var sanitized_obj  = {};

    //for ghost relons we santitize the obj first.
    if ($tbl=="ghost_relons") {
        var TEMP_ONLY = false;
        if ($obj.TEMP_ONLY==true) {
            TEMP_ONLY=true;
        }
        sanitized_obj = {
            from_nid: $obj.from_nid,
            reln: $obj.reln,
            rorigin: $obj.rorigin,
            time:$obj.time,
            to_s: $obj.to_s,
            to_type: $obj.to_type,
            txid: $obj.txid,
            type: $obj.type,
            ghost_order: $obj.ghost_order,
            ghost_owner: $obj.ghost_owner,
            owned_by:$obj.ghost_owner,
            TEMP_ONLY:TEMP_ONLY
        }
        if ($obj.flathexint!=undefined) sanitized_obj=$obj.fathexint;
        if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;
    }

    //all other tbl objects are fine as-is.
    else sanitized_obj = $obj;

    return new Promise((resolve, reject) => {
        var request = objectStore.put(sanitized_obj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}


DB_local.___delete = async ($ids, $tbl="collab_relons", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($ids.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $ids.length; t++) {
        if ($ids[t]==undefined || (typeof $ids[t]!=='string' && typeof $ids[t]!=='number')) continue;
        await DB_local.__deleteFinal(objectStore, $ids[t]);
    }
}

DB_local.___deleteWithMultiKey = async ($multikey, $tbl="adf_otc_swaps", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($multikey.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $multikey.length; t++) {
        await DB_local.__deleteFinal(objectStore, $multikey[t]);
    }
}


DB_local.__deleteFinal = async($objectStore, $key)=>{
    return new Promise((resolve, reject) => {
        var request = $objectStore.delete($key);
        request.onsuccess = function(event) { 
            return resolve(true); 
        }
        request.onerror = function(event) { 
            return resolve(false); 
        } 
    });
}

/************** (from file ./js/db/db_local.js)**/